/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                             OFELI's Configuration File

  ==============================================================================*/

#ifndef __OFELI_CONFIG_H
#define __OFELI_CONFIG_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <complex>
#include <stdexcept>
#include "util/constants.h"
#include "util/macros.h"
#include "io/fparser/fparser.h"

/*! \file OFELI_Config.h
 *  \ingroup Util
 *  \brief File that contains some macros.
 *
 * All these macros can be modified before compiling the library.
 */

/*!
 * \def OFELI_VERSION
 * gives the current version of the library
 */
#define OFELI_VERSION                 "3.0.0"

/*!
 * \def OFELI_RELEASE_DATE
 * gives the date (month-year) of the current release
 */
#define OFELI_RELEASE_DATE            "2-2014"


#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define XGRAPH_                            0
#undef  INIT_PETSC
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/*! \def GRAPH_MEMORY
 *  \ingroup Mesh
 * \brief Memory necessary to store matrix graph.
 * \details This value is necessary only if nodes are to be renumbered.
 */
#define GRAPH_MEMORY                 1000000

/*! \def MAX_NB_EQUATIONS
 *  \ingroup Solver
 * \brief Maximum number of equations
 * \details Useful for coupled problems
 */
#define MAX_NB_EQUATIONS                   5

/*! \def MAX_NB_INPUT_FIELDS
 *  \ingroup Solver
 * \brief Maximum number of fields for an equation
 * \details Useful for coupled problems
 */
#define MAX_NB_INPUT_FIELDS                3

/*! \def MAX_NB_MESHES
 *  \ingroup Solver
 * \brief Maximum number of meshes
 * \details Useful for coupled problems
 */
#define MAX_NB_MESHES                     10

/*! \def MAX_NB_SUBMESHES
 *  \ingroup Mesh
 * \brief Maximum number of mesh partitions.
 * \details Useful if mesh is to be partitioned by class Partition
 */
#define MAX_NB_SUBMESHES                 500

/*! \def MAX_NB_ELEMENTS
 *  \ingroup Mesh
 * \brief Maximal Number of elements
 */
#define MAX_NB_ELEMENTS                10000

/*! \def MAX_NB_NODES
 *  \ingroup Mesh
 * \brief Maximal number of nodes
 */
#define MAX_NB_NODES                   10000

/*! \def MAX_NB_SIDES
 *  \ingroup Mesh
 * \brief Maximal number of sides in.
 */
#define MAX_NB_SIDES                   30000

/*! \def MAX_NB_EDGES
 *  \ingroup Mesh
 * \brief Maximal Number of edges.
 */
#define MAX_NB_EDGES                   30000

/*! \def MAX_NBDOF_NODE
 *  \ingroup Mesh
 * \brief Maximum number of DOF supported by each node
 */
#define MAX_NBDOF_NODE                     6

/*! \def MAX_NBDOF_SIDE
 *  \ingroup Mesh
 * \brief Maximum number of DOF supported by each side
 */
#define MAX_NBDOF_SIDE                     6

/*! \def MAX_NBDOF_EDGE
 *  \ingroup Mesh
 * \brief Maximum number of DOF supported by each edge
 */
#define MAX_NBDOF_EDGE                     2

/*! \def MAX_NB_ELEMENT_NODES
 *  \ingroup Mesh
 * \brief Maximum number of nodes by element
 */
#define MAX_NB_ELEMENT_NODES              20

/*! \def MAX_NB_ELEMENT_EDGES
 *  \ingroup Mesh
 * \brief Maximum number of edges by element
 */
#define MAX_NB_ELEMENT_EDGES              10

/*! \def MAX_NB_SIDE_NODES
 *  \ingroup Mesh
 *  \brief Maximum number of nodes by side
 */
#define MAX_NB_SIDE_NODES                  9

/*! \def MAX_NB_ELEMENT_SIDES
 *  \ingroup Mesh
 *  \brief Maximum number of sides by element
 */
#define MAX_NB_ELEMENT_SIDES               8

/*! \def MAX_NB_ELEMENT_DOF
 *  \ingroup Mesh
 *  \brief Maximum number of dof by element
 */
#define MAX_NB_ELEMENT_DOF   27

/*! \def MAX_NB_SIDE_DOF
 *  \ingroup Mesh
 *  \brief Maximum number of dof by side
 */
#define MAX_NB_SIDE_DOF   4

/*! \def MAX_NB_INT_PTS
 *  \ingroup Mesh
 *  \brief Maximum number of integration points in element
 */
#define MAX_NB_INT_PTS  20

/*! \def MAX_NB_VERTICES
 *  \ingroup Mesh
 *  \brief Maximum number of vertices
 */
#define MAX_NB_VERTICES                 2000

/*! \def MAX_NB_HOLES
 *  \ingroup Mesh
 *  \brief Maximum number of holes
 *  \details Used only if class Domain is invoked
 */
#define MAX_NB_HOLES                      10

/*! \def MAX_NB_LINES
 *  \ingroup Mesh
 *  \brief Maximum number of lines
 *  \details Used only if class Domain is invoked
 */
#define MAX_NB_LINES                     200

/*! \def MAX_NB_LINE_NODES
 *  \ingroup Mesh
 *  \brief Maximum number of nodes on each line
 *  \details Used only if class Domain is invoked
 */
#define MAX_NB_LINE_NODES               1000

/*! \def MAX_NB_CONT_LINES
 *  \ingroup Mesh
 *  \brief Maximum number of lines on each contour
 *  \details Used only if class Domain is invoked
 */
#define MAX_NB_CONT_LINES                 50

/*! \def MAX_NB_SUB_DOMAINS
 *  \ingroup Mesh
 *  \brief Maximum number of subdomains
 *  \details Used only if class Domain is invoked
 */
#define MAX_NB_SUB_DOMAINS                 5

/*! \def MAX_NB_MATERIALS
 *  \ingroup Mesh
 *  \brief Maximum number of materials
 */
#define MAX_NB_MATERIALS                  10

/*! \def MAX_NB_PAR
 *  \ingroup IO
 *  \brief Maximum number of parameters
 *  \details Used in class IPF
 */
#define MAX_NB_PAR                        50

/*! \def MAX_ARRAY_SIZE
 *  \ingroup IO
 *  \brief Maximum array size
 *  \details Used in class IPF
 */
#define MAX_ARRAY_SIZE                   100

/*! \def MAX_INPUT_STRING_LENGTH
 *  \ingroup IO
 *  \brief Maximum string length
 *  \details  Used in class IPF
 */
#define MAX_INPUT_STRING_LENGTH          100

/*! \def FILENAME_LENGTH
 *  \ingroup IO
 *  \brief Length of a string defining a file name
 */
#define FILENAME_LENGTH                  150

#ifndef DOXYGEN_SHOULD_SKIP_THIS
// Macro
#define XGRAPH_          0
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// A macro to define any number
#define ANY               123.45678901e05
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/*! \brief FieldName
 * \brief Enumerate variable that selects field
 */
enum FieldName {
   TEMPERATURE                  =   0,    ///< Temperature scalar field
   DISPLACEMENT                 =   1,    ///< Displacement vector field
   VELOCITY                     =   2,    ///< Velocity vector field
   PRESSURE                     =   3,    ///< Pressure scalar field
   ELECTRIC_FIELD               =   4,    ///< Electric vector field
   MAGNETIC_FIELD               =   5,    ///< Magnetic vector field
   POTENTIAL                    =   6,    ///< Potential vector field
   CONCENTRATION                =   7,    ///< Concentration field
   STRESS                       =   8,    ///< Stress vector field
   INTERNAL_ENERGY              =   9,    ///< Internal energy
   DENSITY                      =  10,    ///< Density
   MOMENTUM                     =  11,    ///< Momentum
};

/*! \enum FieldType
 * \brief Enumerate variable that selects field type
 */
 enum FieldType {
    NONE          =  0,     ///< No support information
    NODE_FIELD    =  1,     ///< DOFs are supported by nodes
    ELEMENT_FIELD =  2,     ///< DOFs are supported by elements
    SIDE_FIELD    =  3,     ///< DOFs are supported by sides (faces in 3-D, edges in 2-D)
    EDGE_FIELD    =  4      ///< DOFs are supported by edges
 };

/*! \enum DOFSupport
 * \brief Choose Support of degrees of freedom
 */
enum DOFSupport {
   NODE_DOF    =  1,            /*!< DOFs are supported by nodes */
   ELEMENT_DOF =  2,            /*!< DOFs are supported by elements */
   SIDE_DOF    =  3,            /*!< DOFs are supported by sides */
   EDGE_DOF    =  4             /*!< DOFs are supported by edges */
};


/*! \enum ElementShape
 * \brief Enumerate list for element shapes
 */
enum ElementShape {
   LINE,             ///< Elements are segment lines
   TRIANGLE,         ///< Elements are triangles
   QUADRILATERAL,    ///< Elements are quadrilaterals
   TETRAHEDRON,      ///< Elements are tetrahedra
   HEXAHEDRON,       ///< Elements are hexahedra (bricks)
   PENTAHEDRON       ///< Elements are pentahedra (prisms)
};

/*! \brief FileFormat
 * \brief Enumerate variable that selects external file formats
 */
enum ExternalFileFormat {
   OFELI_FF,    ///< OFELI file format
   GMSH,        ///< Gmsh file format
   GNUPLOT,     ///< Gnuplot file format
   MATLAB,      ///< Matlab m-file
   VTK,         ///< VTK file format
   TECPLOT,     ///< Tecplot file format
   EASYMESH,    ///< Easymesh file format
   GAMBIT,      ///< Gambit file format
   BAMG,        ///< Bamg file format
   NETGEN,      ///< Netgen file format
   TETGEN,      ///< Tetgen file format
   TRIANGLE_FF  ///< Triangle file format
};

/*! \enum ElementType
 * \brief Choose finite element type
 */
enum ElementType {
   LINE2    =  0,            /*!< Line element with 2 nodes (P1) */
   TRIANG3  =  1,            /*!< Triangular element with 3 nodes (P1) */
   QUAD4    =  2,            /*!< Quadrilateral element with 4 nodes (Q1) */
   TETRA4   =  3,            /*!< Tetrahedral element with 4 nodes (P1) */
   HEXA8    =  4,            /*!< Hexahedral element with 8 nodes (Q1) */
   PENTA6   =  5             /*!< Pentahedral element with 6 nodes (P1*Q1) */
};

#define PATH_MATERIAL_WIN  "c:\\Program Files\\ofeli\\material"
#ifdef WIN32
#define PATH_MATERIAL  "c:\\Program Files\\ofeli\\material"
#else
#include "datadir.h"
#endif


#if defined(_MSC_VER) && !defined(__MWERKS__)
#define _MSVCPP_ _MSC_VER
#endif

#if defined ( OFELI_COMPILED_WITH_DEBUGGING_ )
#define _DEBUG
#endif

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define OFELI_CONJ std::conj
#define OFELI_ABS  std::abs
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/*! \typedef lsize_t
 *  \ingroup Util
 * \brief This type stands for type \p unsigned \p long
 */
typedef  unsigned long         lsize_t;

/*! \typedef real_t
 *  \ingroup Util
 * \brief This type stands for \p double
 */
typedef  double                real_t;

/*! \typedef complex_t
 *  \ingroup Util
 * \brief This type stands for type \p std::complex<double>
 */
typedef  std::complex<double>  complex_t;

using std::min;
using std::max;

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#ifdef WIN32
#define PATH_SEP "\\";
#else
#define PATH_SEP "/";
#endif
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define MATERIAL_EXT      ".md"
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/**
 * @namespace OFELI
 * \brief Namespace OFELI groups all %OFELI library classes, functions and global variables
 */

namespace OFELI {

/**
 *  \ingroup Global
 *  \brief Time step counter
 *  \details This counter must be initialized by the user if the
 *  macro timeLoop is not used
 *  @remark May be used in conjunction with the macro TimeLoop.
 *  In this case, it has to be initialized before. Its default value is 1
 */
    extern unsigned int theStep;

/**
 *  \ingroup Global
 *  \brief Iteration counter
 *  \details This counter must be initialized by the user
 *  @remark May be used in conjunction with the macro IterationLoop.
 *  Its default value is 1
 */
    extern unsigned int theIteration;

/**
 *  \ingroup Global
 *  \brief Number of time steps
 *  @remark May be used in conjunction with the macro TimeLoop.
 */
    extern unsigned int NbTimeSteps;

/**
 *  \ingroup Global
 *  \brief Maximal number of iterations
 *  @remark May be used in conjunction with the macro IterationLoop.
 *  Its default value is 1000
 */
    extern unsigned int MaxNbIterations;

/**
 *  \ingroup Global
 *  \brief Parameter for verbosity of message outputting
 *  \details Its default value is 1
 */
    extern unsigned int Verbosity;

/**
 *  \ingroup Global
 *  \brief Time step label
 *  @remark May be used in conjunction with the macro TimeLoop.
 *  In this case, it has to be initialized before
 */
   extern double theTimeStep;

/**
 *  \ingroup Global
 *  \brief Time value
 *  @remark May be used in conjunction with the macro TimeLoop.
 *  Its default value is \c 0.0
 */
   extern double theTime;

/**
 *  \ingroup Global
 *  \brief Final time value
 *  @remark May be used in conjunction with the macro TimeLoop.
 *  In this case, it has to be initialized before
 */
   extern double theFinalTime;

/**
 *  \ingroup Global
 *  \brief Tolerance value for convergence
 *  @remark May be used within an iterative procedure.
 *  Its default value is \c 1.e-8
 */
   extern double theTolerance;

/**
 *  \ingroup Global
 *  \brief Value of discrepancy for an iterative procedure
 *  Its default value is \c 1.0
 */
   extern double theDiscrepancy;

/**
 *  \ingroup Global
 *  \brief Boolean variable to say if an iterative procedure has converged
 *  \details Its default value is \c false
 */
    extern bool Converged;

/**
 *  \ingroup Global
 *  Boolean to say if PETSc use was initialized.
 *  Useful only if PETSc is used
 */
    extern bool InitPetsc;

} /* namespace OFELI */

#endif
