/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

       Definition of abstract class 'AbsEqua' for Finite Element Equations

  ==============================================================================*/


#ifndef __ABS_EQUA_H
#define __ABS_EQUA_H

#include <float.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>

#ifdef USE_PETSC
#include "linear_algebra/petsc/PETScMatrix.h"
#endif

#include "OFELI_Config.h"
#include "mesh/Mesh.h"
#include "mesh/Element.h"
#include "mesh/Side.h"
#include "linear_algebra/DMatrix.h"
#include "linear_algebra/SpMatrix.h"
#include "linear_algebra/SkMatrix.h"
#include "linear_algebra/SkSMatrix.h"
#include "linear_algebra/TrMatrix.h"
#include "linear_algebra/LocalMatrix.h"
#include "linear_algebra/LocalVect.h"
#include "mesh/Material.h"
#include "io/UserData.h"
#include "solvers/LinearSolver.h"
#include "util/Gauss.h"
#include "io/Prescription.h"
#include "solvers/Eigen.h"

namespace OFELI {

/*! \defgroup Equation General Purpose Equations
 *  \brief Gathers equation related classes
 */

extern Material theMaterial;

/*! \file AbsEqua.h
 *  \brief Definition file for abstract class AbsEqua.
 */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define eMat   (_M)
#define sMat   (_sM)
#define eRHS   (_B)
#define sRHS   (_sB)
#define ePrev  (_P)
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/*! \enum PDE_Terms
 * Enumerate variable that selects various terms in partial differential equations
 */
enum PDE_Terms {
   MASS                =   1,     /*!< Consistent mass term                    */
   LUMPED_MASS         =   2,     /*!< Lumped mass term                        */
   CAPACITY            =   1,     /*!< Consistent capacity term                */
   LUMPED_CAPACITY     =   2,     /*!< Lumped capacity term                    */
   VISCOSITY           =   4,     /*!< Viscosity term                          */
   STIFFNESS           =   4,     /*!< Stiffness term                          */
   DIFFUSION           =   4,     /*!< Diffusion term                          */
   CONVECTION          =   8,     /*!< Convection term                         */
   DEVIATORIC          =   8,     /*!< Deviatoric term                         */
   DILATATION          =  16,     /*!< Dilatational term                       */
   ELECTRIC            =   8,     /*!< Electric term                           */
   MAGNETIC            =  16,     /*!< Magnetic term                           */
   LOAD                =  32,     /*!< Body load term                          */
   HEAT_SOURCE         =  32,     /*!< Body heat source term                   */
   BOUNDARY_TRACTION   =  64,     /*!< Boundary traction (pressure) term       */
   HEAT_FLUX           =  64,     /*!< Boundary heat flux term                 */
   CONTACT             = 128,     /*!< Signorini contact                       */
   BUOYANCY            =  16,     /*!< Buoyancy force term                     */
   LORENTZ_FORCE       =  32      /*!< Lorentz force term                      */
};
   

/*! \enum EqDataType
 * Enumerate variable that selects equation data type
 */
enum EqDataType {
   INITIAL_FIELD                =    1,    /*!< Initial condition                      */
   SOLUTION                     =    1,    /*!< Solution vector (same as Initial)      */
   INITIAL_AUX_1                =    2,    /*!< Initial auxiliary field                */
   INITIAL_AUX_2                =    3,    /*!< Initial auxiliary field                */
   INITIAL_AUX_3                =    4,    /*!< Initial auxiliary field                */
   INITIAL_AUX_4                =    5,    /*!< Initial auxiliary field                */
   BOUNDARY_CONDITION           =    6,    /*!< Boundary condition data                */
   BODY_FORCE                   =    7,    /*!< Body force data                        */
   SOURCE                       =    7,    /*!< Source data (same as Body force)       */
   POINT_FORCE                  =    8,    /*!< Localized (at point) force             */
   BOUNDARY_FORCE               =    9,    /*!< Boundary force data                    */
   FLUX                         =    9,    /*!< Flux data (same as Boundary force)     */
   TRACTION                     =    9,    /*!< Traction data (same as Boundary force) */
   AUX_INPUT_FIELD_1            =   10,    /*!< Auxiliary input field 1                */
   AUX_INPUT_FIELD_2            =   11,    /*!< Auxiliary input field 2                */
   AUX_INPUT_FIELD_3            =   12,    /*!< Auxiliary input field 3                */
   AUX_INPUT_FIELD_4            =   13,    /*!< Auxiliary input field 4                */
   DISPLACEMENT_FIELD           =   14,    /*!< A displacement field                   */
   VELOCITY_FIELD               =   15,    /*!< A velocity field                       */
   TEMPERATURE_FIELD            =   16     /*!< A temperature field                    */
};


/*! \enum ArrayType
 * Selects local or global option for array as argument.
 */
enum ArrayType {
   LOCAL_ARRAY  = 0,   /*!< For a local array labeled with local numbering */
   GLOBAL_ARRAY = 1    /*!< For a local array labeled with global numbering */
};


/*! \enum TimeScheme
 * Selects time integration scheme
 */
enum TimeScheme {
   STATIONARY      = 0,             /*!< No time scheme: stationary                */
   FORWARD_EULER   = 1,             /*!< Forward Euler scheme (Explicit)           */
   BACKWARD_EULER  = 2,             /*!< Backward Euler scheme (Implicit)          */
   CRANK_NICOLSON  = 3,             /*!< Crank-Nicolson scheme                     */
   HEUN            = 4,             /*!< Heun scheme                               */
   NEWMARK         = 5,             /*!< Newmark scheme                            */
   LEAP_FROG       = 6,             /*!< Leap Frog scheme                          */
   ADAMS_BASHFORTH = 7,             /*!< Adams-Bashforth scheme (2nd Order)        */
   AB2             = 7,             /*!< Adams-Bashforth scheme (2nd Order)        */
   RUNGE_KUTTA     = 8,             /*!< 4-th Order Runge-Kutta scheme (4th Order) */
   RK4             = 8              /*!< 4-th Order Runge-Kutta scheme             */
};

/*! \enum PDE
 * Choose partial differential equation to `
 */
enum PDE {
   LAPLACE,                      /*!< Laplace equation                        */
   DIFFUSION_CONVECTION,         /*!< Diffusion Convection equation           */
   THERMAL_PHASE_CHANGE,         /*!< Thermal phase change problem (Stefan)   */
   INCOMPRESSIBLE_NAVIER_STOKES, /*!< Incompressible Navier-Stokes equations  */
   LINEARIZED_ELASTICITY,        /*!< Linearized elasticity equations         */
   PLANAR_TRUSS,                 /*!< 2-D truss equation                      */
   SPATIAL_BEAM                  /*!< 3-D beam equations                      */
};

/*! \enum FEType
 * Choose Finite Element Type
 */
enum FEType {
   FE_2D_3N,                     /*!< 2-D elements, 3-Nodes (P1)              */
   FE_2D_6N,                     /*!< 2-D elements, 6-Nodes (P2)              */
   FE_2D_4N,                     /*!< 2-D elements, 4-Nodes (Q1)              */
   FE_3D_AXI_3N,                 /*!< 3-D Axisymmetric elements, 3-Nodes (P1) */
   FE_3D_4N,                     /*!< 3-D elements, 4-Nodes (P1)              */
   FE_3D_8N                      /*!< 3-D elements, 8-Nodes (Q1)              */
};

/*! \enum AnalysisType
 * Choose analysis type
 */
enum AnalysisType {
   STEADY_STATE,                 /*!< Steady state analysis                   */
   TRANSIENT,                    /*!< Transient analysis                      */
   OPTIMIZATION                  /*!< Optimization analysis                   */
};


/*! \class AbsEqua
 * \ingroup Equation
 * \brief Mother abstract class to describe equation.
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 */

template<class T_> class SkMatrix;
template<class T_> class SkSMatrix;
template<class T_> class SpMatrix;

template<class T_>
class AbsEqua
{

 public:

/// \brief Default constructor
    AbsEqua()
    {
       _set_matrix = false;
       _analysis = STEADY_STATE;
       _time_scheme = STATIONARY;
       _final_time = 0;
       _step = 0;
       _nb_fields = 1;
       _ls.setSolver(-1);
       _A = NULL;
       _b = _bc = _bf = _sf = _u = _v = NULL;
       _sol_given = _bc_given = _init_given = _bf_given = _sf_given = false;
       _eigen = false;
   }

/// \brief Constructor with mesh instance
    AbsEqua(class Mesh &mesh)
    {
       _theMesh = &mesh;
       _solver = -1;
       _set_matrix = false;
       _sol_given = _bc_given = _init_given = _bf_given = _sf_given = false;
       _eigen = false;
       _analysis = STEADY_STATE;
       _time_scheme = STATIONARY;
       _final_time = 0;
       _A = NULL;
    }

/// \brief Destructor
    virtual ~AbsEqua()
    {
       if (_A)
          delete _A;
#ifdef _OFELI_DEBUG
      std::clog << "An instance of class AbsEqua is destructed in\n";
      std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// \brief Define mesh and renumber DOFs after removing imposed ones
    void setMesh(class Mesh &m) { _theMesh = &m; _theMesh->removeImposedDOF(); }

/// \brief Return reference to Mesh instance
/// @return Reference to Mesh instance
    Mesh & getMesh() const { return *_theMesh; }

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** \brief Build equation.
 *  This member function is to be used if one wants to use a class that inherits
 *  from AbsEqua that handles the whole solution process (including assembly).
 */
    virtual void build() { }
    virtual int runOneTimeStep() { return 0; }
    virtual int run() { return 0; }
    virtual int runSteadyState() { return run(); }
    virtual int runTransient() { return 0; }
    virtual void getA0(Matrix<real_t> *A) { }
    virtual void getA1(Matrix<real_t> *A) { }
    virtual void getA2(Matrix<real_t> *A) { }
    virtual Vect<T_> *getbf(const Vect<T_> &f) { return NULL; }
    virtual Vect<T_> *getsf(const Vect<T_> &f) { return NULL; }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** \brief Set transient analysis settings
 *  \details Define a set of parameters for time integration
 *  @param [in] s Scheme
 *  @param [in] p1
 *  @param [in] p2
 *  @param [in] p3
 */
    void setTransient(int    s,
                      real_t p1=1,
                      real_t p2=1,
                      real_t p3=1)
    {
       _time_scheme = s;
       _time_parameter1 = p1;
       _time_parameter2 = p2;
       _time_parameter3 = p3;
       if (s==FORWARD_EULER)
          _theta = 0;
       else if (s==BACKWARD_EULER)
          _theta = 1;
       else if (s==CRANK_NICOLSON)
          _theta = 0.5;
   }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

// \brief Return number of (element or side) nodes.
//    virtual size_t getNbNodes() const;

// \brief Return number of (element or side) equations.
//    virtual size_t getNbEq() const;

/// \brief Return reference to linear solver instance
    LinearSolver<T_> &getLinearSolver() { return _ls; }

/** \brief Choose solver for the linear system
 *  @param [in] ls Solver of the linear system.
 *  To choose among the enumerated values: <tt>DIRECT_SOLVER</tt>, <tt>CG_SOLVER</tt>,
    <tt>GMRES_SOLVER</tt>
    <ul>
      <li> <tt>DIRECT_SOLVER</tt>, Use a facorization solver [default]
      <li> <tt>CG_SOLVER</tt>, Conjugate Gradient iterative solver
      <li> <tt>CGS_SOLVER</tt>, Squared Conjugate Gradient iterative solver
      <li> <tt>BICG_SOLVER</tt>, BiConjugate Gradient iterative solver
      <li> <tt>BICG_STAB_SOLVER</tt>, BiConjugate Gradient Stabilized iterative solver
      <li> <tt>GMRES_SOLVER</tt>, GMRES iterative solver
      <li> <tt>QMR_SOLVER</tt>, QMR iterative solver
    </ul>
 *  @param [in] pc Preconditioner to associate to the iterative solver.
 *  If the direct solver was chosen for the first argument this argument is not used.
 *  Otherwise choose among the enumerated values:
    <ul> 
      <li> <tt>IDENT_PREC</tt>, Identity preconditioner (no preconditioning [default])
      <li> <tt>DIAG_PREC</tt>, Diagonal preconditioner
      <li> <tt>ILU_PREC</tt>, Incomplete LU factorization preconditioner
    </ul>
 */
    void setSolver(int ls,
                   int pc=IDENT_PREC)
    {
       if (_set_matrix == false) {
          _matrix_type = SPARSE;
          _A = new SpMatrix<T_>;
          _A->setMesh(*_theMesh,0);
       }
       _ls.setSolver(ls,pc);
    }

/** \brief Compute eigenvalues and eigenvectors.
 *  \details Eigenvalues and vectors are computed using the Bathe's subspace
 *  iteration method.
 *  @param [in] nb_eigv Number of eigenvalues to compute
 *  @param [in] g Option to choose whether to solve a generalized
 *  eigenvalue problem (true) or a standard one (false).
 *  The generalized eigenvalue problem corresponds to the case where a
 *  consistent mass matrix (rather than a lumped one) is computed.
 *  Default value is false.
 */
    int solveEigenProblem(int  nb_eigv,
                          bool g=false)
    {
       int ret;
       _nb_eigv = nb_eigv;
       if (g) {
          SkSMatrix<real_t> K(*_theMesh), M(*_theMesh);
          buildEigen(K,M);
          _ev.setMatrix(K,M);
          ret = _ev.runSubSpace(_nb_eigv);
       }
       else {
          SkSMatrix<real_t> K(*_theMesh);
          Vect<real_t> M(_theMesh->getNbEq());
          buildEigen(K,M);
          _ev.setMatrix(K,M);
          ret = _ev.runSubSpace(_nb_eigv);
       }
       _eigen = true;
       return ret;
    }

/** \brief Return the n-th eigenvalue.
 *  \details This functions works only if the member function getEigen was called
 *  with an argument <tt>nb_eigv</tt> greater or equal to <tt>n</tt>.
 *  Otherwise it returns <tt>0</tt>.
 */
    real_t getEigenValue(int n) const
    {
       if (!_eigen || n>int(_nb_eigv))
          return 0;
       else
          return _ev.getEigenValue(n);
    }

/** \brief Store the eigenvector corresponding to a given eigenvalue
 *  @param [in] n Label of the eigenvalue
 *  @param [out] v Vect instance containing the corresponding eigenvector.
 *  This vector is resized.
 */
    void getEigenVector(int           n,
                        Vect<real_t>& v) const
    {
       v.setSize(_theMesh->getNbDOF());
       Vect<real_t> w(_theMesh->getNbEq());
       _ev.getEigenVector(n,w);
       v.insertBC(*_theMesh,w);
    }

/// \brief Return reference to eigenproblem solver
    class Eigen &getEigenSolver() { return _ev; }

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    virtual void buildEigen(SkSMatrix<real_t>& K,
                            SkSMatrix<real_t>& M) { }
    virtual void buildEigen(SkSMatrix<real_t>& K, 
                            Vect<real_t>&      M) { }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// \brief Set terms in equations
    virtual void setWithConvection(int f) { f = 0; }

    virtual void setTerms(PDE_Terms t) { _terms = t; }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** \brief Solve the linear system
 *  @param [in] A Pointer to matrix of the system (can be pointer to any matrix class)
 *  @param [in] b Vector containing right-hand side
 *  @param [in,out] x Vector containing initial guess of solution on input, actual solution on output
 */
    int SolveLinearSystem(Matrix<T_>* A,
                          Vect<T_>&   b,
                          Vect<T_>&   x)
    {
      _set_matrix = true;
      _ls.setMatrix(A);
      _ls.setRHS(b);
      _ls.setSolution(x);
      if (_ls.getSolver() < 0) {
         if (_matrix_type==SKYLINE)
            _ls.setSolver(DIRECT_SOLVER);
         else if (_matrix_type==(SPARSE|SYMMETRIC))
            _ls.setSolver(CG_SOLVER);
          else if (_matrix_type==SPARSE)
            _ls.setSolver(GMRES_SOLVER);
      }
      return _ls.solve();
    }

#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// \brief Set Analysis Type
    void setAnalysis(int a) { _analysis = a; }

/// \brief Set time iteration index
    void setTimeIndex(size_t step) { _step = step; }

/// \brief Set time step duration
    void setTimeStep(real_t t) { _time_step = t; }

/// \brief Return time step
    real_t getTimeStep() const { return _time_step; }

/// \brief Set current time
    void setTime(real_t t) { _time = t; }

/// \brief Set final time
    void setFinalTime(real_t t) { _final_time = t; }

/// \brief Set verbose parameter
    void setVerbose(int v) { _verbose = v; }

/// \brief Return number of fields
    size_t getNbFields() const { return _nb_fields; }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setTimeIntegration(int scheme) { _time_scheme = scheme; }
    int getTimeIntegration() const { return _time_scheme; }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */
    
#ifndef DOXYGEN_SHOULD_SKIP_THIS

/// \brief Return equation name
    string getEquationName() const { return _equation_name; }

/// \brief Return Finite Element Type
    string getFiniteElementType() const { return _finite_element; }

/** \brief Set equation input data
 *  @param [in] opt Parameter that selects data type for input. This parameter
 *  is to be chosen in the enumerated variable EqDataType
 *  @param [in] u Vect instance that contains input vector data
 *  List of data types contains <tt>INITIAL_FIELD</tt>, <tt>BOUNDARY_CONDITION_DATA</tt>, 
 *  <tt>SOURCE_DATA</tt> or <tt>FLUX</tt> with obvious meaning
 */
    virtual void setInput(EqDataType opt,
                          Vect<T_>&  u)
    {
       if (opt==INITIAL_FIELD) {
          _u = &u;
          _init_given = true;
       }
       else if (opt==SOLUTION) {
          _u = &u;
          _sol_given = true;
       }
       else if (opt==BOUNDARY_CONDITION) {
          _bc = &u;
          _bc_given = true;
       }
       else if (opt==SOURCE || opt==BODY_FORCE) {
          _bf = &u;
          _bf_given = true;
       }
       else if (opt==FLUX || opt==TRACTION) {
          _sf = &u;
          _sf_given = true;
       }
       else
          ;
    }

/// \brief Set prescription
/// @param [in] p Prescription instance
    void set(class Prescription& p) { _prescription = &p; }

/// \brief Choose tolerance value
    void setTolerance(real_t toler) { _toler = toler; }

/** \brief Get Equation matrix for the i-th order term. Case of matrix given by its pointer
 *  \details Space discretization of a partial differential equation results in
 *  a differential system of equations:\n
 *  A2 d2u/dt2 + A1 du/dt + A0 u = f\n
 *  We are thus limited to at most second order equations
 *  @param [in] i order of the term (between 0 and 2)
 *  @note This function is available if the matrix was initialized in the
 *  appropriate equation class
 */
      //    void getMatrix(Matrix<T_> *A, int i) const { A = _AA[i]; }

/** \brief Get Equation matrix for the i-th order term. Case of an SkMatrix instance
 *  \details Space discretization of a partial differential equation results in
 *  a differential system of equations:\n
 *  A2 d2u/dt2 + A1 du/dt + A0 u = f\n
 *  We are thus limited to at most second order equations
 *  @param [in] i order of the term (between 0 and 2)
 *  @note This function is available if the matrix was initialized in the
 *  appropriate equation class
 */
      //    void getMatrix(SkMatrix<T_> &A, int i) const { A = *(SkMatrix<T_> *)(_AA[i]); }
   
/** \brief Get Equation matrix for the i-th order term. Case of an SkSMatrix instance
 *  \details Space discretization of a partial differential equation results in
 *  a differential system of equations:\n
 *  A2 d2u/dt2 + A1 du/dt + A0 u = f\n
 *  We are thus limited to at most second order equations
 *  @param [in] i order of the term (between 0 and 2)
 *  @note This function is available if the matrix was initialized in the
 *  appropriate equation class
 */
//    void getMatrix(SkSMatrix<T_> &A, int i) const { A = *(SkSMatrix<T_> *)(_AA[i]); }

/** \brief Get Equation matrix for the i-th order term. Case of an SpMatrix instance
 *  \details Space discretization of a partial differential equation results in
 *  a differential system of equations:\n
 *  A2 d2u/dt2 + A1 du/dt + A0 u = f\n
 *  We are thus limited to at most second order equations
 *  @param [in] i order of the term (between 0 and 2)
 *  @note This function is available if the matrix was initialized in the
 *  appropriate equation class
 */
      //  void getMatrix(SpMatrix<T_> &A, int i) const { A = *(SpMatrix<T_> *)(_AA[i]); }
   
/** \brief Get Equation matrix for the i-th order term. Case of a DMatrix instance
 *  \details Space discretization of a partial differential equation results in
 *  a differential system of equations:\n
 *  A2 d2u/dt2 + A1 du/dt + A0 u = f\n
 *  We are thus limited to at most second order equations
 *  @param [in] i order of the term (between 0 and 2)
 *  @note This function is available if the matrix was initialized in the
 *  appropriate equation class
 */
      //    void getMatrix(DMatrix<T_> &A, int i) const { A = *(DMatrix<T_> *)(_AA[i]); }

#endif /* DOXYGEN_SHOULD_SKIP_THIS */
   
protected :
#ifndef DOXYGEN_SHOULD_SKIP_THIS
   Mesh              *_theMesh;
   int               _field_type, _time_scheme, _analysis, _verbose, _terms;
   int               _matrix_type, _solver, _max_it;
   size_t            _step, _nb_fields, _nb_eigv;
   real_t            _theta, _alpha, _beta, _time_step, _time, _final_time;
   real_t            _time_parameter1, _time_parameter2, _time_parameter3;
   class Eigen       _ev;
   Matrix<T_>        *_A, *_CM;
   Vect<T_>          *_bc, *_bf, *_sf, *_u, *_v, *_b, *_LM, _uu;
   bool              _eigen, _sol_given, _bc_given, _init_given, _bf_given, _sf_given; 
   bool              _constant_matrix, _constant_mesh, _set_matrix;
   int               _sol_type, _init_type, _bc_type, _bf_type, _sf_type;
   string            _equation_name, _finite_element;
   LinearSolver<T_>  _ls;
   real_t            _toler;
   Prescription      *_prescription;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

public:

#ifndef DOXYGEN_SHOULD_SKIP_THIS
   void setMatrix(SkSMatrix<T_> &A)
   {
      _A = &A;
      _matrix_type = SKYLINE|SYMMETRIC;
      _A->setMesh(*_theMesh);
   }

   void setMatrix(SkMatrix<T_> &A)
   {
      _A = &A;
      _matrix_type = SKYLINE;
      _A->setMesh(*_theMesh);
   }

   void setMatrix(SpMatrix<T_> &A)
   {
      _A = &A;
      _matrix_type = SPARSE;
      _A->setMesh(*_theMesh);
   }
   bool isConstantMatrix() const { return _constant_matrix; }
   bool isConstantMesh() const { return _constant_mesh; }
   void setConstantMatrix() { _constant_matrix = true; }
   void setConstantMesh() { _constant_mesh = true; }
   virtual void setTerms(int opt) { _terms = opt; }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */
};

} /* namespace OFELI */

#endif
