/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                     Template Class Matrix for banded matrices

  ==============================================================================*/


#ifndef _B_MATRIX_H
#define _B_MATRIX_H

#include "linear_algebra/Matrix.h"

namespace OFELI {

/*! \file BMatrix.h
 *  \brief Definition file for class BMatrix.
 */

template<class T_> class BMatrix;

/*! \class BMatrix
 *  \ingroup VectMat
 * \brief To handle band matrices.
 *
 * \details This class enables storing and manipulating band matrices.
 * The matrix can have different numbers of lower and upper co-diagonals
 *
 * \tparam <T_> Data type (double, float, complex<double>, ...)
 */


template<class T_>
class BMatrix : public Matrix<T_>
{

 public:

    using Matrix<T_>::operator();

/// \brief Default constructor.
/// \details Initialize a zero dimension band matrix
    BMatrix();

/** \brief Constructor that for a band matrix with given size and bandwidth.
 *  \details Assign 0 to all matrix entries.
 *  @param [in] size Number of rows and columns
 *  @param [in] ld Number of lower co-diagonals (must be > 0)
 *  @param [in] ud Number of upper co-diagonals (must be > 0)
 */
    BMatrix(size_t size,
            int    ld,
            int    ud);

/// \brief Copy Constructor
    BMatrix(const BMatrix& m);

/// \brief Destructor
    ~BMatrix();

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(class Mesh&  mesh,
                       size_t dof=0);
    void setMesh(class Mesh&  mesh,
                       size_t dof=0,
                       size_t type=0);
    void setMesh(      size_t dof, 
                 class Mesh&  mesh,
                       int    code=0);
    void setMesh(      size_t dof,
                       size_t nb_eq,
                 class Mesh&  mesh);
    void setDiag() { }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/** \brief Set size (number of rows) and storage of matrix
 *  @param [in] size Number of rows and columns
 *  @param [in] ld Number of lower co-diagonals (must be > 0)
 *  @param [in] ud Number of upper co-diagonals (must be > 0)
 */
    void setSize(size_t size,
                 int    ld,
                 int    ud);

/// \brief Multiply matrix by vector <tt>x</tt> and add result to <tt>y</tt>
    void MultAdd(const Vect<T_>& x,
                       Vect<T_>& y) const;

/// \brief Multiply matrix by vector <tt>a*x</tt> and add result to <tt>y</tt>
    void MultAdd(      T_        a, 
                 const Vect<T_>& x, 
                       Vect<T_>& y) const;

/// \brief Multiply matrix by vector <tt>x</tt> and save result in <tt>y</tt>
    void Mult(const Vect<T_>& x,
                    Vect<T_>& y) const;

/// \brief Multiply transpose of matrix by vector <tt>x</tt> and save result in <tt>y</tt>
    void TMult(const Vect<T_>& x,
                     Vect<T_>& y) const;
   
/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] x %Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(      T_           a,
              const BMatrix<T_>& x);
   
/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] x %Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(      T_          a,
              const Matrix<T_>* x);

/// \brief Add constant <tt>val</tt> to an entry <tt>(i,j)</tt> of the matrix.
    void set(      size_t  i,
                   size_t  j,
             const T_&     val);

/// \brief Add constant <tt>val</tt> value to an entry <tt>(i,j)</tt> of the matrix.
    void add(      size_t  i,
                   size_t  j,
             const T_&     val);

/** \brief Operator () (Constant version).
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ operator()(size_t i,
                  size_t j) const;

/** \brief Operator () (Non constant version).
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ & operator()(size_t i,
                    size_t j);

/// \brief Operator =.
/// \details Copy matrix <tt>m</tt> to current matrix instance.
    BMatrix<T_> & operator=(const BMatrix<T_>& m);

/// \brief Operator =.
/// \details Assign constant value <tt>x</tt> to all matrix entries.
    BMatrix<T_> & operator=(const T_& x);

/// \brief Operator *=.
/// \details Premultiply matrix entries by constant value <tt>x</tt>
    BMatrix<T_> & operator*=(const T_& x);

/// \brief Operator +=.
/// \details Add constant <tt>x</tt> to matrix entries.
    BMatrix<T_> & operator+=(const T_& x);

/** \brief Factorize matrix.
 *  \details Perform LU factorization of matrix. Entries of matrix are modified after
 *  use of this function
 *  @return <tt>0</tt> if factorization is successful, <tt>i</tt> means that the i-th pivot was null.
 */
    int Factor();

/** \brief Solve system with current matrix (forward and back substitution).
 *  @param [in] b Vect instance that contains right-hand side on input and solution on output.
 *  @return
 *  <ul>
 *    <li><tt>0</tt> if solution was normally performed
 *    <li><tt>n</tt> if the <tt>n</tt>-th pivot is null.
 *  </ul>
 * 
 *  @warning %Matrix is modified after this function.
 */
    int Solve(Vect<T_>& b);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// Return storage information
    size_t getColInd(size_t i) const;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// Return storage information
    size_t getRowPtr(size_t i) const;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/// \brief Return C-Array.
    T_* get() const { return _a; }

/// \brief Return entry <tt>(i,j)</tt> of matrix
    T_ get(size_t i, size_t j) const;

 private:
   size_t              _imin, _imax;
   int                 _ld, _ud;
   vector<vector<T_> > _a;

   using Matrix<T_>::_rmin;
   using Matrix<T_>::_cmin;
   using Matrix<T_>::_rmax;
   using Matrix<T_>::_cmax;
   using Matrix<T_>::_size;
   using Matrix<T_>::_nb_cols;
   using Matrix<T_>::_nb_rows;
   using Matrix<T_>::_length;
   using Matrix<T_>::_diag;
   using Matrix<T_>::_fact;
   using Matrix<T_>::_zero;
   using Matrix<T_>::_temp;
   using Matrix<T_>::_ch;
};

///////////////////////////////////////////////////////////////////////////////
//                         I M P L E M E N T A T I O N                       //
///////////////////////////////////////////////////////////////////////////////

template<class T_>
BMatrix<T_>::BMatrix()
{
   _size = _ld = _ud = 0;
   _fact = false;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class BMatrix is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
BMatrix<T_>::BMatrix(size_t size,
                     int    ld,
                     int    ud)
{
   try {
       if (size < 3 || ld < 0 || ud < 0 || ud + ld + 1 > int(size))
          THROW_RT("BMatrix(size_t,int,int): Illegal arguments.");
   }
   CATCH("BMatrix");
   setSize(size,ld,ud);
}

template<class T_>
BMatrix<T_>::BMatrix(const BMatrix& m)
{
   _size = m._size;
   _ld = m._ld;
   _ud = m._ud;
   _length = m._length;
   _a = m._a;
   _fact = m._fact;
}


template<class T_>
BMatrix<T_>::~BMatrix()
{
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class BMatrix is destructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void BMatrix<T_>::setMesh(class Mesh&  mesh,
                                size_t dof)
{
   try {
      THROW_RT("setMesh(Mesh,size_t): This member function is not valid for class BMatrix");
   }
   CATCH("BMatrix");
}


template<class T_>
void BMatrix<T_>::setMesh(class Mesh&  mesh,
                                size_t dof,
                                size_t type)
{
   try {
      THROW_RT("setMesh(Mesh,size_t,size_t): This member function is not valid for class BMatrix");
   }
   CATCH("BMatrix");
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void BMatrix<T_>::setMesh(      size_t dof,
                          class Mesh&  mesh,
                                int    code)
{
   try {
      THROW_RT("setMesh(size_t,Mesh,int): This member function is not valid for class BMatrix");
   }
   CATCH("BMatrix");
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void BMatrix<T_>::setMesh(      size_t dof,
                                size_t nb_eq,
                          class Mesh&  mesh)
{
   try {
      THROW_RT("setMesh(size_t,size_t,Mesh): This member function is not valid for class BMatrix");
   }
   CATCH("BMatrix");
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

template<class T_>
void BMatrix<T_>::setSize(size_t size,
                          int    ld,
                          int    ud)
{
   _nb_rows = _nb_cols = _size = size;
   _length = (ld+ud+1) * _nb_rows;
   _ld = ld; _ud = ud;
   _a.resize(_size);
   for (int i=0; i<int(_size); i++)
      _a[i].resize(_ld+_ud+1,static_cast<T_>(0));
   _ch.resize(_size);
   _ch[0] = 0;
   for (int i=1; i<int(_size); i++)
      _ch[i] = i+1;
   _fact = false;
}


template<class T_>
void BMatrix<T_>::MultAdd(const Vect<T_>& x,
                                Vect<T_>& y) const
{
  for (int i=0; i<int(_size); ++i)
     for (int j=max(i-_ld,0); j<min(i+_ud+_ld,int(_size)); ++j)
        y[i] += _a[i][j+_ld-i]*x[j];
}


template<class T_>
void BMatrix<T_>::MultAdd(      T_        a,
                          const Vect<T_>& x,
                                Vect<T_>& y) const
{
  for (int i=0; i<int(_size); ++i)
     for (int j=max(i-_ld,0); j<min(i+_ud+_ld,int(_size)); ++j)
        y[i] += a*_a[i][j+_ld-i]*x[j];
}


template<class T_>
void BMatrix<T_>::Mult(const Vect<T_>& x,
                             Vect<T_>& y) const
{
   y = static_cast<T_>(0);
   MultAdd(x,y);
}


template<class T_>
void BMatrix<T_>::TMult(const Vect<T_>& x,
                              Vect<T_>& y) const
{
}


template<class T_>
void BMatrix<T_>::Axpy(      T_           a,
                       const BMatrix<T_>& x)
{
}


template<class T_>
void BMatrix<T_>::Axpy(      T_          a,
                       const Matrix<T_>* x)
{
}


template<class T_>
void BMatrix<T_>::set(      size_t  i,
                            size_t  j,
                      const T_&     val)
{
   if (_ld+int(j)-int(i)>=0 && _ld+int(j-i)<=_ud+_ld)
      _a[i-1][_ld+j-i] = val;
}


template<class T_>
void BMatrix<T_>::add(      size_t  i,
                            size_t  j,
                      const T_&     val)
{
   if (_ld+int(j)-int(i)>=0 && _ld+int(j)-int(i)<=_ud+_ld)
      _a[i-1][_ld+j-i] += val;
}


template<class T_>
T_ BMatrix<T_>::operator()(size_t i,
                           size_t j) const
{
#ifdef _BOUNDS_DEBUG
   assert(i>0 || i<=_size);
#endif
   if (_ld+int(j)-int(i)>=0 && _ld+int(j)-int(i)<=_ud+_ld)
      return _a[i-1][_ld+j-i];
   else
      return _zero;
}


template<class T_>
T_ & BMatrix<T_>::operator()(size_t i,
                             size_t j)
{
#ifdef _BOUNDS_DEBUG
   assert(i>0 || i<=_size);
#endif
   try {
      if (_ld+int(j)-int(i)>=0 && _ld+int(j)-int(i)<=_ud+_ld)
         return _a[i-1][_ld+j-i];
      else
         THROW_RT("operator(): Illegal pair of indices "+itos(i)+","+itos(j)+").");
   }
   CATCH("BMatrix");
   return _zero;
}


template<class T_>
BMatrix<T_> & BMatrix<T_>::operator=(const BMatrix<T_>& m)
{
   _a = m._a;
   return *this;
}


template<class T_>
BMatrix<T_> & BMatrix<T_>::operator=(const T_& x)
{
   for (size_t i=0; i<_size; ++i)
      for (int j=0; j<_ld+_ud+1; ++j)
         _a[i][j] = x;
   return *this;
}


template<class T_>
BMatrix<T_> & BMatrix<T_>::operator*=(const T_& x)
{
   for (size_t i=0; i<_size; ++i)
      for (int j=0; j<_ld+_ud+1; ++j)
         _a[i][j] *= x;
   return *this;
}


template<class T_>
BMatrix<T_> & BMatrix<T_>::operator+=(const T_& x)
{
   for (size_t i=0; i<_size; ++i)
      for (int j=0; j<_ld+_ud+1; ++j)
         _a[i][j] += x;
   return *this;
}


template<class T_>
int BMatrix<T_>::Factor()
{
   for (int i=0; i<int(_size)-1; i++) {
      int kend=min(_ld+1,int(_size)-i), kjend=min(_ud+1,int(_size)-i);
      try {
         if (Abs(_a[i][_ld]) < OFELI_EPSMCH)
            THROW_RT("Factor(): The " + itos(i+1) + "-th pivot is null.");
      }
      CATCH("BMatrix");
      for (int k=1; k!=kend; k++) {
         _a[k+i][_ld-k] /= _a[i][_ld];
         for (int j=1; j!=kjend; j++)
            _a[k+i][j+_ld-k] -= _a[k+i][_ld-k] * _a[i][j+_ld];
      }
   }
   _fact = true;
   return 0;
}


template<class T_>
int BMatrix<T_>::Solve(Vect<T_>& b)
{
   try {
      if (_fact == false)
         THROW_RT("Solve(Vect<double>): %Matrix may not have been factorized.");
   }
   CATCH("BMatrix");
   try {
      if (_size < 3 || _ld < 0 || _ud < 0 || _ud + _ld + 1 > int(_size))
         THROW_RT("Solve(Vect<double>): Illegal arguments.");
   }
   CATCH("BMatrix");

   for (int i=0; i<int(_size)-1; i++) {
      int kend = min(_ld+1,int(_size)-i);
      for (int k=1; k!=kend; k++)
         b[k+i] -= _a[k+i][_ld-k] * b[i];
   }

   for (int i=_size-1; i>=0; i--) {
      int kend = min(_ud+1,int(_size)-i);
      for (int k=1; k<kend; k++)
         b[i] -= _a[i][k+_ld] * b[i+k];
      b[i] /= _a[i][_ld];
  }
  return 0;
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
size_t BMatrix<T_>::getColInd(size_t i) const
{
#ifdef _BOUNDS_DEBUG
   assert(i<=_length && i>0);
#endif
   return _ch[i-1];
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
size_t BMatrix<T_>::getRowPtr(size_t i) const
{
#ifdef _BOUNDS_DEBUG
   assert(i<=_nb_rows+1 && i>0);
#endif
   return _ch[i-1];
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
T_ BMatrix<T_>::get(size_t i,
                    size_t j) const
{
   if (_ld+int(j)-int(i)>=0 && _ld+int(j)-int(i)<=_ld+_ud)
      return _a[i-1][_ld+j-i];
   else
      return _zero;
}

///////////////////////////////////////////////////////////////////////////////
//                           ASSOCIATED  FUNCTIONS                           //
///////////////////////////////////////////////////////////////////////////////


/** \fn BMatrix<T_> operator*(T_ a, const BMatrix<T_> &A)
 *  \brief Operator * (Premultiplication of matrix by constant)
 *  \ingroup VectMat
 *  @return a*A
 */
template<class T_>
BMatrix<T_> operator*(      T_           a,
                      const BMatrix<T_>& A)
{
   BMatrix<T_> v(A);
   for (size_t i=0; i<A.getLength(); ++i)
      v[i] *= a;
   return v;
}


/** \fn ostream& operator<<(ostream& s, const BMatrix<T_> &a)
 *  \ingroup VectMat
 *  \brief Output matrix in output stream
 */
template<class T_>
ostream& operator<<(      ostream&     s,
                    const BMatrix<T_>& a)
{
   size_t rmin, rmax, cmin, cmax;
   a.getPrintView(rmin,rmax,cmin,cmax);
   s.setf(ios::right|ios::scientific);
   s << endl;
   for (size_t i=rmin; i<=rmax; i++) {
      s << "\nRow  " << setw(6) << i << endl;
      for (size_t j=cmin; j<=cmax; j++)
         s << "  " << setprecision(8) << std::setfill(' ') << setw(18) << a(i,j);
      s << endl;
   }
   return s;
}

} /* namespace OFELI */

#endif
