/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

               Definition of Template Class DMatrix for dense matrices

  ==============================================================================*/


#ifndef __DMATRIX_H
#define __DMATRIX_H

#include "linear_algebra/Matrix.h"


namespace OFELI {

/*! \file DMatrix.h
 *  \brief Definition file for class DMatrix.
 */

/*! \class DMatrix
 *  \ingroup VectMat
 * \brief To handle dense matrices.
 *
 * This class enables storing and manipulating general dense matrices. 
 * Matrices can be square or rectangle ones.
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 */


template<class T_>
class DMatrix : public Matrix<T_>
{
   using Matrix<T_>::_rmin;
   using Matrix<T_>::_cmin;
   using Matrix<T_>::_rmax;
   using Matrix<T_>::_cmax;
   using Matrix<T_>::_nb_rows;
   using Matrix<T_>::_nb_cols;
   using Matrix<T_>::_size;
   using Matrix<T_>::_length;
   using Matrix<T_>::_a;
   using Matrix<T_>::_diag;
   using Matrix<T_>::_ch;
   using Matrix<T_>::_zero;
   using Matrix<T_>::_theMesh;
   using Matrix<T_>::_fact;
   using Matrix<T_>::_is_diagonal;


 public:

    using Matrix<T_>::operator();

/*----------------------------  BASIC OPERATORS  ----------------------------*/

/// \brief Default constructor.
/// \details Initializes a zero-dimension matrix.
    DMatrix();

/// \brief Constructor for a matrix with <tt>nr</tt> rows and <tt>nr</tt> columns.
/// \details %Matrix entries are set to <tt>0</tt>.
    DMatrix(size_t nr);

/// \brief Constructor for a matrix with <tt>nr</tt> rows and <tt>nc</tt> columns.
/// \details Matrix entries are set to 0.
    DMatrix(size_t nr,
            size_t nc);

/** \brief Constructor that uses a Vect instance. The class uses the memory
 *  space occupied by this vector.
 *  @param [in] v Vector to copy
 */
    DMatrix(Vect<T_>& v);

/// \brief Copy Constructor.
/// @param [in] m %Matrix to copy
    DMatrix(const DMatrix<T_>& m);

/// \brief Destructor.
    ~DMatrix();

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(class Mesh&  mesh,
                       size_t dof=0);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(      size_t dof,
                 class Mesh&  mesh,
                       int    code=0);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(      size_t dof,
                       size_t nb_eq,
                 class Mesh&  mesh);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/// \brief Store diagonal entries in a separate internal vector
    void setDiag();

/// \brief Set size (number of rows) of matrix.
/// @param [in] size Number of rows and columns.
    void setSize(size_t size);

/** \brief Set size (number of rows and columns) of matrix.
 *  @param [in] nr Number of rows.
 *  @param [in] nc Number of columns.
 */
    void setSize(size_t nr,
                 size_t nc);

/// \brief Get <tt>j</tt>-th column vector.
    Vect<T_> getColumn(size_t j) const;

/// \brief Get <tt>i</tt>-th row vector.
    Vect<T_> getRow(size_t i) const;

/** \brief Assign a constant value to an entry of the matrix.
 *  @param [in] i row index of matrix
 *  @param [in] j column index of matrix
 *  @param [in] val Value to assign to <tt>a(i,j)</tt>.
 */
    void set(      size_t  i,
                   size_t  j,
             const T_&     val)
    { _a[_nb_cols*(i-1)+j-1] = val; }

/** \brief Copy a given vector to a prescribed row in the matrix.
 *  @param [in] i row index to be assigned
 *  @param [in] v Vector to copy
 */
    void setRow(      size_t    i,
                const Vect<T_>& v);

/** \brief Copy a given vector to a prescribed column in the matrix.
 *  @param [in] i column index to be assigned
 *  @param [in] v Vector to copy
 */
    void setColumn(      size_t    i,
                   const Vect<T_>& v);

/** \brief Multiply matrix by vector <tt>a*x</tt> and add result to <tt>y</tt>.
 *  @param [in] a constant to multiply by
 *  @param [in] x Vector to multiply by <tt>a</tt>
 *  @param [in,out] y on input, vector to add to. On output, result.
 */
    void MultAdd(      T_        a,
                 const Vect<T_>& x,
                       Vect<T_>& y) const;

/** \brief Multiply matrix by vector <tt>x</tt> and add result to <tt>y</tt>.
 *  @param [in] x Vector to add to <tt>y</tt>
 *  @param [in,out] y on input, vector to add to. On output, result.
 */
    void MultAdd(const Vect<T_>& x,
                       Vect<T_>& y) const;

/** \brief Multiply matrix by vector <tt>x</tt> and save result in <tt>y</tt>.
 *  @param [in] x Vector to add to <tt>y</tt>
 *  @param [out] y Result.
 */
    void Mult(const Vect<T_>& x,
                    Vect<T_>& y) const;

/** \brief Multiply transpose of matrix by vector <tt>x</tt> and add result in <tt>y</tt>.
 *  @param [in] x Vector to add to <tt>y</tt>
 *  @param [in,out] y on input, vector to add to. On output, result.
 */
    void TMult(const Vect<T_>& x,
                     Vect<T_>& y) const;

/** \brief Add constant <tt>val</tt> to entry <tt>(i,j)</tt> of the matrix.
 *  @param [in] i row index
 *  @param [in] j column index
 *  @param [in] val Constant to add
 */
    void add(      size_t  i,
                   size_t  j,
             const T_&     val);

/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] m %Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(      T_           a,
              const DMatrix<T_>& m);
   
/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] m %Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(       T_          a,
              const  Matrix<T_>* m);

/** \brief Operator () (Constant version).
 *  Return <tt>a(i,j)</tt>
 *  @param [in] i row index
 *  @param [in] j column index
 */
    T_ operator()(size_t i,
                  size_t j) const;

/** \brief Operator () (Non constant version).
 *  Return <tt>a(i,j)</tt>
 *  @param [in] i row index
 *  @param [in] j column index
 */
    T_ & operator()(size_t i,
                    size_t j);

/** \brief Factorize matrix (LU factorization).
 *  @return
 *  <ul>
 *     <li><tt>0</tt> if factorization was normally performed,
 *     <li><tt>n</tt> if the <tt>n</tt>-th pivot is null.
 *  </ul>
 */
    int Factor();

/** \brief Solve system with factorized matrix (forward and back substitution).
 *  @param [in,out] b Vect instance that contains right-hand side on input and solution on output.
 *  @return
 *  <ul>
 *     <li><tt>0</tt> if solution was normally performed,
 *     <li><tt>n</tt> if the <tt>n</tt>-th pivot is null.
 *  </ul>
 *  Solution is performed only if factorization has previouly been invoked.
 */
    int Solve(Vect<T_>& b);

/** \brief Solve system with factorized matrix (forward and back substitution).
 *  @param [in] b Vect instance that contains right-hand side
 *  @param [out] x Vect instance that contains solution
 *  @return
 *  <ul>
 *     <li><tt>0</tt> if solution was normally performed
 *     <li><tt>n</tt> if the <tt>n</tt>-th pivot is null
 *  </ul>
 *  Solution is performed only if factorization has previouly been invoked.
 */
    int Solve(const Vect<T_>& b,
                    Vect<T_>& x);

/// \brief Operator <tt>=</tt>
/// \details Copy matrix <tt>m</tt> to current matrix instance.
    DMatrix & operator=(DMatrix<T_>& m);

/// \brief Operator +=
/// \details Add matrix <tt>m</tt> to current matrix instance.
    DMatrix & operator+=(const DMatrix<T_>& m);

/// \brief Operator -=
/// \details Subtract matrix <tt>m</tt> from current matrix instance.
    DMatrix & operator-=(const DMatrix<T_>& m);

/// \brief Operator <tt>=</tt>
/// \details Assign matrix to identity times <tt>x</tt>
    DMatrix & operator=(const T_& x);

/// \brief Operator <tt>*=</tt>
/// \details Premultiply matrix entries by constant value <tt>x</tt>.
    DMatrix & operator*=(const T_& x);

/// \brief Operator <tt>+=</tt>
/// \details Add constant value <tt>x</tt> to matrix entries
    DMatrix & operator+=(const T_& x);

/// \brief Operator <tt>-=</tt>
/// \details Subtract constant value <tt>x</tt> from matrix entries.
    DMatrix & operator-=(const T_& x);

/** \brief Return storage information
 *  @param [in] i Index of stored entry in matrix
 *  @return Column index for this entry
 */
    size_t getColInd(size_t i) const;

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    size_t getRowPtr(size_t i) const;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/// \brief Return matrix as C-Array.
/// \details %Matrix is stored row by row.
    T_ *getArray() const { return _a; }

/// \brief Return entry <tt>(i,j)</tt> of matrix
    T_ get(size_t i,
           size_t j) const;

/** \brief Set a window for matrix printing
 *  @param [in] rmin first row index to view 
 *  @param [in] rmax last row index to view 
 *  @param [in] cmin first column index to view 
 *  @param [in] cmax last column index to view 
 */
    void setPrintView(size_t rmin,
                      size_t rmax,
                      size_t cmin,
                      size_t cmax);
};


///////////////////////////////////////////////////////////////////////////////
//                         I M P L E M E N T A T I O N                       //
///////////////////////////////////////////////////////////////////////////////


template<class T_>
DMatrix<T_>::DMatrix()
{
   _length = _size = _nb_rows = _nb_cols = 0;
   _fact = false;
   _is_diagonal = false;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class DMatrix is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
DMatrix<T_>::DMatrix(size_t nr)
{
   setSize(nr);
   _is_diagonal = false;
   _rmin = 1, _rmax = _nb_rows;
   _cmin = 1, _cmax = _nb_cols;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class DMatrix is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
DMatrix<T_>::DMatrix(size_t nr,
                     size_t nc)
{
   setSize(nr,nc);
   _is_diagonal = false;
   _rmin = 1, _rmax = _nb_rows;
   _cmin = 1, _cmax = _nb_cols;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class DMatrix is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
DMatrix<T_>::DMatrix(Vect<T_>& v)
{
   _is_diagonal = true;
   _length = v.size();
   _nb_rows = _nb_cols = _size = sqrt(real_t(_length));
   _rmin = 1, _rmax = _nb_rows;
   _cmin = 1, _cmax = _nb_cols;
   _a = v;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class DMatrix is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
DMatrix<T_>::DMatrix(const DMatrix<T_>& m)
{
   setSize(m._nb_rows,m._nb_cols);
   _a = m._a;
   _diag = m._diag;
   _fact = m._fact;
   _is_diagonal = m._is_diagonal;
   _rmin = m._rmin, _rmax = m._rmax;
   _cmin = m._cmin, _cmax = m._cmax;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class DMatrix is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
DMatrix<T_>::~DMatrix()
{
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class DMatrix is destructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void DMatrix<T_>::setMesh(class Mesh&  mesh,
                                size_t dof)
{
   Matrix<T_>::init_set_mesh(mesh,dof);
   _length = _size*_size;
   _diag.resize(_size);
   _a.resize(_length);
   _zero = static_cast<T_>(0);
   _fact = false;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void DMatrix<T_>::setMesh(      size_t dof,
                          class Mesh&  mesh,
                                int    code)
{
// This is just to avoid warning on unused variable
   dof  = 0;
   code = 0;
   _theMesh = &mesh;
   if (mesh.getDim()==0) { }
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void DMatrix<T_>::setMesh(      size_t dof,
                                size_t nb_eq,
                          class Mesh&  mesh)
{
// This is just to avoid warning on unused variable
   dof = 0;
   nb_eq = 0;
   if (mesh.getDim()==0) { }
   _theMesh = &mesh;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
void DMatrix<T_>::setDiag()
{
   for (size_t i=0; i<_size; i++)
      _diag.set(i+1,_a[_nb_cols*i+i]);
}


template<class T_>
void DMatrix<T_>::setSize(size_t size)
{
   _nb_rows = _nb_cols = _size = size;
   _length = _nb_rows*_nb_cols;
   _a.resize(_length);
   _diag.resize(_size);
   _rmin = 1;
   _rmax = _nb_rows;
   _cmin = 1;
   _cmax = _nb_cols;
}


template<class T_>
void DMatrix<T_>::setSize(size_t nr,
                          size_t nc)
{
   _nb_rows = nr;
   _nb_cols = nc;
   _size = 0;
   if (_nb_rows==_nb_cols)
      _size = _nb_rows;
   _length = lsize_t(_nb_rows*_nb_cols);
   _a.resize(_length);
   _rmin = 1, _rmax = _nb_rows;
   _cmin = 1, _cmax = _nb_cols;
}


template<class T_>
Vect<T_> DMatrix<T_>::getColumn(size_t j) const
{
   Vect<T_> v(_nb_rows);
   for (size_t i=0; i<_nb_rows; i++)
      v.set(i+1,_a[_nb_cols*i+j-1]);
   return v;
}


template<class T_>
void DMatrix<T_>::setColumn(       size_t    j,
                             const Vect<T_>& v)
{
   for (size_t i=0; i<_nb_rows; i++)
      _a[_nb_cols*i+j-1] = v[i];
}


template<class T_>
Vect<T_> DMatrix<T_>::getRow(size_t i) const
{
   Vect<T_> v(_nb_cols);
   for (size_t j=0; j<_nb_cols; j++)
      v.set(j+1,_a[_nb_cols*(i-1)+j]);
   return v;
}


template<class T_>
void DMatrix<T_>::setRow(      size_t    i,
                         const Vect<T_>& v)
{
   for (size_t j=0; j<_nb_cols; j++)
      _a[_nb_cols*(i-1)+j] = v[j];
}


template<class T_>
void DMatrix<T_>::MultAdd(      T_        a,
                          const Vect<T_>& x,
                                Vect<T_>& y) const
{
  for (size_t i=0; i<_nb_rows; i++)
     for (size_t j=0; j<_nb_cols; j++)
        y.add(i+1,a*_a[_nb_cols*i+j]*x[j]);
}


template<class T_>
void DMatrix<T_>::MultAdd(const Vect<T_>& x,
                                Vect<T_>& y) const
{
   T_ s;
   for (size_t i=0; i<_nb_rows; i++) {
      s = 0;
      for (size_t j=0; j<_nb_cols; j++)
         s += _a[_nb_cols*i+j] * x[j];
      y.set(i+1,s);
   }
}


template<class T_>
void DMatrix<T_>::Mult(const Vect<T_>& x,
                             Vect<T_>& y) const
{
   y = T_(0);
   MultAdd(x,y);
}


template<class T_>
void DMatrix<T_>::TMult(const Vect<T_>& x,
                              Vect<T_>& y) const
{
   for (size_t i=0; i<_nb_rows; i++)
      for (size_t j=0; j<_nb_cols; j++)
         y.add(i+1,_a[_nb_cols*(j-1)+i-1]*x[j]);
}


template<class T_>
void DMatrix<T_>::add(      size_t  i,
                            size_t  j,
                      const T_&     val)
{
   _a[_nb_cols*(i-1)+j-1] += val;
}


template<class T_>
T_ DMatrix<T_>::operator()(size_t i,
                           size_t j) const
{
#ifdef _BOUNDS
   assert(i>0);
   assert(i<=_nb_rows);
   assert(j>0);
   assert(j<=_nb_cols);
#endif
   return _a[_nb_cols*(i-1)+j-1];
}


template<class T_>
T_ & DMatrix<T_>::operator()(size_t i,
                             size_t j)
{
#ifdef _BOUNDS
   assert(i>0);
   assert(i<=_nb_rows);
   assert(j>0);
   assert(j<=_nb_cols);
#endif
   return _a[_nb_cols*(i-1)+j-1];
}


template<class T_>
int DMatrix<T_>::Factor()
{
   register size_t i, j, k;
   try {
      if (_size==0)
         THROW_RT("Factor(): Can't factorize a rectangle matrix.");
   }
   CATCH("DMatrix");
   for (i=1; i<_size; i++) {
      for (j=1; j<=i; j++) {
         try {
            if (Abs(_a[_nb_rows*(j-1)+j-1]) < OFELI_EPSMCH)
               THROW_RT("Factor(): The " + itos(i) + "-th pivot is null.");
         }
         CATCH("DMatrix");
         _a[_nb_rows*i+j-1] /= _a[_nb_rows*(j-1)+j-1];
         for (k=0; k<j; k++)
            _a[_nb_rows*i+j] -= _a[_nb_rows*i+k]*_a[_nb_rows*k+j];
      }
      for (j=i+1; j<_size; j++)
         for (k=0; k<i; k++)
            _a[_nb_rows*i+j] -= _a[_nb_rows*i+k]*_a[_nb_rows*k+j];
   }
   _fact = true;
   return 0;
}


template<class T_>
int DMatrix<T_>::Solve(Vect<T_>& b)
{
   try {
      if (!_fact)
         THROW_RT("Solve(b): Matrix may have not been factorized."); 
   }
   CATCH("DMatrix");
   try {
      if (_size==0)
         THROW_RT("Solve(Vect<doube>): Can't solve a system with a rectangle matrix.");
   }
   CATCH("DMatrix");
   for (size_t i=0; i<_size; i++) {
      T_ s = 0;
      for (size_t j=0; j<i; j++)
         s += _a[_nb_cols*i+j] * b[j];
      b[i] -= s;
   }
   for (int ii=int(_size)-1; ii>-1; ii--) {
      try {
         if (Abs(_a[_nb_cols*ii+ii]) < OFELI_EPSMCH)
            THROW_RT("Solve(Vect<real_t>): The " + itos(ii+1) + "-th pivot is null.");
      }
      CATCH("DMatrix");
      b[ii] /= _a[_nb_cols*ii+ii];
      for (size_t j=0; j<size_t(ii); j++)
         b[j] -= b[ii] * _a[_nb_cols*j+ii];
   }
   return 0;
}


template<class T_>
int DMatrix<T_>::Solve(const Vect<T_>& b,
                             Vect<T_>& x)
{
   x = b;
   return Solve(x);
}


template<class T_>
DMatrix<T_>& DMatrix<T_>::operator=(DMatrix<T_>& m)
{
   _a = m._a;
   return *this;
}


template<class T_>
DMatrix<T_>& DMatrix<T_>::operator+=(const DMatrix<T_>& m)
{
   for (size_t i=1; i<=_length; i++)
      _a.add(i,m._a[i-1]);
   return *this;
}


template<class T_>
DMatrix<T_>& DMatrix<T_>::operator-=(const DMatrix<T_>& m)
{
   for (size_t i=0; i<_length; i++)
      _a.add(i,-m._a[i-1]);
   return *this;
}


template<class T_>
DMatrix<T_>& DMatrix<T_>::operator=(const T_& x)
{
   try {
      if (_nb_rows!=_nb_cols)
         THROW_RT("operator=(T_): Operator is valid for square matrices only.");
   }
   CATCH("DMatrix");
   _fact = false;
   Clear(_a);
   for (size_t i=1; i<=_size; i++) {
      set(i,i,x);
      _diag[i-1] = x;
   }
   return *this;
}


template<class T_>
DMatrix<T_> & DMatrix<T_>::operator*=(const T_& x)
{
   for (size_t i=1; i<=_length; i++)
      _a.set(i,_a[i-1]*x);
   return *this;
}


template<class T_>
DMatrix<T_> & DMatrix<T_>::operator+=(const T_& x)
{
   for (size_t i=1; i<=_length; i++)
      add(i,x);
   return *this;
}


template<class T_>
DMatrix<T_> & DMatrix<T_>::operator-=(const T_& x)
{
   for (size_t i=1; i<=_length; i++)
      add(i,-x);
   return *this;
}


template<class T_>
size_t DMatrix<T_>::getColInd(size_t i) const
{
#ifdef _BOUNDS_DEBUG
   assert(i<=_length && i>0);
#endif
   return _ch[i-1];
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
size_t DMatrix<T_>::getRowPtr(size_t i) const
{
#ifdef _BOUNDS_DEBUG
   assert(i<=_nb_rows+1 && i>0);
#endif
   return _ch[i-1];
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
T_ DMatrix<T_>::get(size_t i,
                    size_t j) const 
{
   return _a[_nb_cols*(i-1)+j-1]; 
}


template<class T_>
void DMatrix<T_>::setPrintView(size_t rmin,
                               size_t rmax,
                               size_t cmin,
                               size_t cmax)
{
   _rmin = rmin;
   _rmax = rmax;
   if (_rmin==0)
      _rmin = 1;
   if (_rmax>_nb_rows)
      _rmax = _nb_rows;
   _cmin = cmin;
   _cmax = cmax;
   if (_cmin==0)
      _cmin = 1;
   if (_cmax>_nb_cols)
      _cmax = _nb_cols;
}


template<class T_>
void DMatrix<T_>::Axpy(      T_           a,
                       const DMatrix<T_>& m)
{
   Axpy(a,m._a,_a);
}


template<class T_>
void DMatrix<T_>::Axpy(      T_          a,
                       const Matrix<T_>* m)
{
   for (size_t i=0; i<_length; i++)
      _a[i] += a * m->_a[i];
}


///////////////////////////////////////////////////////////////////////////////
//                           ASSOCIATED  FUNCTIONS                           //
///////////////////////////////////////////////////////////////////////////////


/// \fn ostream& operator<<(ostream& s, const DMatrix<T_>& a)
/// \ingroup VectMat
/// \brief Output matrix in output stream
template<class T_>
ostream& operator<<(      ostream&     s,
                    const DMatrix<T_>& a)
{
   size_t rmin, rmax, cmin, cmax;
   a.getPrintView(rmin,rmax,cmin,cmax);
   s.setf(ios::right|ios::scientific);
   s << endl;
   for (size_t i=rmin-1; i<rmax; i++) {
      s << "\nRow  " << setw(6) << i+1 << endl;
      for (size_t j=cmin-1; j<cmax; j++)
         s << "  " << setprecision(8) << std::setfill(' ')
           << setw(18) << a(i+1,j+1);
      s << endl;
   }
   return s;
}

} /* namespace OFELI */

#endif
