/*==============================================================================

                                    O  F  E  L  I

                            Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                  Definition of class 'SpMatrix' for Sparse matrix

  ==============================================================================*/


#ifndef __SPMATRIX_H
#define __SPMATRIX_H


#include "linear_algebra/Matrix.h"
#include "solvers/CG.h"
#include "solvers/BiCG.h"
#include "solvers/CGS.h"
#include "solvers/BiCG.h"
#include "solvers/BiCGStab.h"
#include "solvers/GMRes.h"
#include "solvers/QMR.h"

namespace OFELI {

/*! \file SpMatrix.h
 *  \brief Definition file for class SpMatrix.
 */

template<class T_> class Prec;
template<class T_> class Vect;


/*! \class SpMatrix
 *  \ingroup VectMat
 *  \brief To handle matrices in sparse storage format.
 *
 * This template class enables to store and manipulate a sparse matrix,
 * i.e. only nonzero terms are stored. Internally, the matrix is stored as a
 * vector instance and uses for the definition of its graph a <tt>Vect<size_t></tt>
 * instance row_ptr and a Vect<size_t> instance <tt>col_ind</tt> that contains
 * respectively addresses of first element of each row and column indices.
 *
 * To illustrate this, consider the matrix
 * \verbatim
            1   2   0
            3   4   0
            0   5   0
   \endverbatim
 *
 * Such a matrix is stored in the vector<real_t> instance {1,2,3,4,5}.
 * The vectors <tt>row_ptr</tt> and <tt>col_ind</tt> are respectively:
 * <tt>{1,3,5,6}</tt>, <tt>{1,2,1,2,2}</tt>
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 */

template<class T_> class SpMatrix : public Matrix<T_>
{
    using Matrix<T_>::_rmin;
    using Matrix<T_>::_cmin;
    using Matrix<T_>::_rmax;
    using Matrix<T_>::_cmax;
    using Matrix<T_>::_nb_rows;
    using Matrix<T_>::_nb_cols;
    using Matrix<T_>::_size;
    using Matrix<T_>::_length;
    using Matrix<T_>::_zero;
    using Matrix<T_>::_temp;
    using Matrix<T_>::_a;
    using Matrix<T_>::_diag;
    using Matrix<T_>::_row_ptr;
    using Matrix<T_>::_col_ind;
    using Matrix<T_>::_ch;
    using Matrix<T_>::_dof_type;
    using Matrix<T_>::_is_diagonal;
    using Matrix<T_>::_theMesh;

 public:

/// \brief Default constructor.
/// \details Initialize a zero-dimension matrix
    SpMatrix();

/** \brief Constructor that initializes current instance as a dense matrix.
 *  \details Normally, for a dense matrix this is not the right class.
 *  @param [in] nr Number of matrix rows.
 *  @param [in] nc Number of matrix columns.
 */
    SpMatrix(size_t nr,
             size_t nc);

/** \brief Constructor that initializes current instance as a dense matrix.
 *  \details Normally, for a dense matrix this is not the right class.
 *  @param [in] size Number of matrix rows (and columns).
 *  @param [in] is_diagonal Boolean argument to say is the matrix is actually a diagonal matrix or not.
 */
    SpMatrix(size_t size,
             int    is_diagonal=false);

/** \brief Constructor using a Mesh instance.
 *  @param [in] mesh Mesh instance from which matrix graph is extracted.
 *  @param [in] dof Option parameter, with default value <tt>0</tt>.\n
 *  <tt>dof=1</tt> means that only one degree of freedom for each node (or element or side)
 *  is taken to determine matrix structure. The value <tt>dof=0</tt> means that matrix 
 *  structure is determined using all DOFs.
 *  @param [in] is_diagonal Boolean argument to say is the matrix is actually a diagonal matrix or not.
 */
    SpMatrix(class Mesh&  mesh,
                   size_t dof=0,
                   int    is_diagonal=false);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    SpMatrix(      size_t dof,
             class Mesh&  mesh,
                   int    code=0);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    SpMatrix(      size_t dof,
                   size_t nb_eq,
             class Mesh&  mesh);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/** \brief Constructor for a square matrix using non zero row and column indices
 *  @param [in] I Vector containing pairs of row and column indices
 *  @param [in] opt Flag indicating if vectors I is cleaned and ordered
 *  (opt=1) or not (opt=0). In the latter case, this vector can have
 *  the same contents more than once and are not necessarily ordered
 */
    SpMatrix(const class vector<std::pair<size_t,size_t> >& I,
                         int                                opt=1);

/** \brief Constructor for a square matrix using non zero row and column indices
 *  @param [in] I Vector containing pairs of row and column indices
 *  @param [in] a Vector containing matrix entries in the same order than
 *  the one given by <tt>I</tt>
 *  @param [in] opt Flag indicating if vector <tt>I</tt> is cleaned and ordered
 *  (<tt>opt=1</tt>: default) or not (<tt>opt=0</tt>). In the latter case, this vector can have
 *  the same contents more than once and are not necessarily ordered
 */
    SpMatrix(const class vector<std::pair<size_t,size_t> >& I,
             const class vector<T_>&                        a,
                         int                                opt=1);

/** \brief Constructor for a rectangle matrix.
 *  @param [in] nr Number of rows
 *  @param [in] nc Number of columns
 *  @param [in] row_ptr Vector of row pointers (See the above description of this class).
 *  @param [in] col_ind Vector of column indices (See the above description of this class).\n
 */
    SpMatrix(      size_t          nr,
                   size_t          nc,
             const vector<size_t>& row_ptr,
             const vector<size_t>& col_ind);

/** \brief Constructor for a rectangle matrix
 *  @param [in] nr Number of rows
 *  @param [in] nc Number of columns
 *  @param [in] row_ptr Vector of row pointers (See the above description of this class).
 *  @param [in] col_ind Vector of column indices (See the above description of this class).\n
 *  @param [in] a vector instance containing matrix entries stored columnwise
 */
    SpMatrix(size_t                nr,
             size_t                nc,
             const vector<size_t>& row_ptr,
             const vector<size_t>& col_ind,
             const vector<T_>&     a);

/** \brief Constructor for a rectangle matrix
 *  @param [in] row_ptr Vector of row pointers (See the above description of this class).
 *  @param [in] col_ind Vector of column indices (See the above description of this class).
 */
    SpMatrix(const vector<size_t>& row_ptr,
             const vector<size_t>& col_ind);

/** \brief Constructor for a rectangle matrix
 *  @param [in] row_ptr Vector of row pointers (See the above description of this class).
 *  @param [in] col_ind Vector of column indices (See the above description of this class).
 *  @param [in] a vector instance that contain matrix entries stored row by row.\n
 *  Number of rows is extracted from vector <tt>row_ptr</tt>.
 */
    SpMatrix(const vector<size_t>& row_ptr,
             const vector<size_t>& col_ind,
             const vector<T_>&     a);

/// \brief Copy constructor.
    SpMatrix(const SpMatrix& m);

/// \brief Destructor.
    ~SpMatrix(void);

/** \brief Determine mesh graph and initialize matrix.
 *  \details This member function is called by constructor with the same arguments
 *  @param [in] mesh Mesh instance for which matrix graph is determined.
 *  @param [in] dof Option parameter, with default value <tt>0</tt>.\n
 *  <tt>dof=1</tt> means that only one degree of freedom for each node (or element or side)
 *  is taken to determine matrix structure. The value <tt>dof=0</tt> means that matrix
 *  structure is determined using all DOFs.
 */
    void setMesh(class Mesh& mesh,
                      size_t dof=0);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(      size_t dof,
                 class Mesh&  mesh,
                       int    code=0);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(      size_t dof, 
                       size_t nb_eq,
                 class Mesh&  mesh);
    void setMesh(class Mesh&  mesh,
                       size_t dof,
                       size_t type);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** \brief Activate extended graph option.
 *  \details An extended graph is the one
 *  for which a node is linked not only to its neighbor nodes but also to the
 *  neighbors of the neighbors.
 */
    void ExtendedGraph();
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/// \brief Activate 1-DOF per node option.
    void setOneDOF();

/// \brief Activate Sides option.
    void setSides();

/// \brief Store diagonal entries in a separate internal vector.
    void setDiag();

/// \brief Get <tt>j</tt>-th column vector.
    Vect<T_> getColumn(size_t j) const;

/** \brief Impose by a diagonal method an essential boundary condition.
 *  \details This member function modifies diagonal terms in matrix and terms
 *  in vector that correspond to degrees of freedom with nonzero code
 *  in order to impose a boundary condition.
 *  The penalty parameter is defined by default equal to 1.e20.
 *  It can be modified by member function \b setPenal(..).
 *  @param [in] mesh Mesh instance from which information is extracted.
 *  @param [in,out] b Vect instance that contains right-hand side.
 *  @param [in] u Vect instance that conatins imposed valued at DOFs where they are to be imposed.
 */
    void DiagPrescribe(class Mesh&     mesh,
                             Vect<T_>& b,
                       const Vect<T_>& u);

/** \brief Impose by a diagonal method an essential boundary condition using the Mesh instance
 *  provided by the constructor
 *  \details This member function modifies diagonal terms in matrix and terms
 *  in vector that correspond to degrees of freedom with nonzero code
 *  in order to impose a boundary condition.
 *  The penalty parameter is defined by default equal to 1.e20.
 *  It can be modified by member function \b setPenal(..).
 *  @param [in,out] b Vect instance that contains right-hand side.
 *  @param [in] u Vect instance that conatins imposed valued at DOFs where they are to be imposed.
 */
    void DiagPrescribe(            Vect<T_>& b,
                       const class Vect<T_>& u);

/// \brief Set size of matrix (case where it's a square matrix).
/// @param [in] size Number of rows and columns.
    void setSize(size_t size);

/** \brief Set size (number of rows) of matrix
 *  @param [in] nr Number of rows
 *  @param [in] nc Number of columns
 */
   void setSize(size_t nr,
                size_t nc);

/** \brief Set graph of matrix by giving a vector of its nonzero entries
 *  @param [in] I Vector containing pairs of row and column indices
 *  @param [in] opt Flag indicating if vector <tt>I</tt> is cleaned and ordered
 *  (<tt>opt=1</tt>: default) or not (<tt>opt=0</tt>). In the latter case, this vector can have
 *  the same contents more than once and are not necessarily ordered
 */
    void setGraph(const vector<std::pair<size_t,size_t> >& I,
                        int                                opt=1);

/// \brief Get <tt>i</tt>-th row vector.
    Vect<T_> getRow(size_t i) const;

/** \brief Operator () (Non constant version)
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ & operator()(size_t i,
                    size_t j);

/** \brief Operator <tt>()</tt> (Constant version)
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ operator()(size_t i,
                  size_t j) const;

/** \brief Operator <tt>()</tt> with one argument (Constant version)
 *  \details Returns <tt>i</tt>-th position in the array storing matrix entries.
 *  The first entry is at location <tt>1</tt>.
 *  Entries are stored row by row.
 */
    const T_ operator()(size_t i) const;

/** \brief Operator <tt>[]</tt> (Constant version).
 *  \details Returns <tt>i</tt>-th position in the array storing matrix entries.
 *  The first entry is at location <tt>0</tt>.
 *  Entries are stored row by row.
 */
    const T_ operator[](size_t i) const;

/// \brief Get mesh instance whose reference will be stored in current instance of SpMatrix.
    void getMesh(class Mesh& mesh);

/** \brief Multiply matrix by vector and save in another one.
 *  @param [in] v Vector to multiply by matrix
 *  @param [out] w Vector that contains on output the result.
 */
    void Mult(const Vect<T_>& v,
                    Vect<T_>& w) const;

/** \brief Multiply matrix by vector <tt>x</tt> and add to <tt>y</tt>.
 *  @param [in] x Vector to multiply by matrix
 *  @param [out] y Vector to add to the result. <tt>y</tt> contains on output the result.
 */
    void MultAdd(const Vect<T_>& x,
                       Vect<T_>& y) const;

/** \brief Multiply matrix by vector <tt>a*x</tt> and add to <tt>y</tt>.
 *  @param [in] a Constant to multiply by matrix
 *  @param [in] x Vector to multiply by matrix
 *  @param [out] y Vector to add to the result. <tt>y</tt> contains on output the result.
 */
    void MultAdd(      T_        a,
                 const Vect<T_>& x,
                       Vect<T_>& y) const;

/** \brief Multiply transpose of matrix by vector <tt>x</tt> and save in <tt>y</tt>.
 *  @param [in] x Vector to multiply by matrix
 *  @param [out] y Vector that contains on output the result.
 */
    void TMult(const Vect<T_>& x,
                     Vect<T_>& y) const;
   
/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] m %Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(      T_            a,
              const SpMatrix<T_>& m);

/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] m Pointer to Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(      T_          a,
              const Matrix<T_>* m);

/** \brief Assign a value to an entry of the matrix
 *  @param [in] i Row index
 *  @param [in] j Column index
 *  @param [in] val Value to assign to <tt>a(i,j)</tt>
 */
    void set(      size_t i,
                   size_t j,
             const T_&    val);

/** \brief Add a value to an entry of the matrix
 *  @param [in] i Row index
 *  @param [in] j Column index
 *  @param [in] val Constant value to add to <tt>a(i,j)</tt>
 */
    void add(      size_t i,
                   size_t j,
             const T_&    val);

/// \brief Operator =.
/// \details Assign constant value <tt>x</tt> to all matrix entries.
    void operator=(const T_& x);

/// \brief Return storage information.
/// @return Column index of the <tt>i</tt>-th stored element in matrix
    size_t getColInd(size_t i) const;

/// \brief Return Row pointer at position <tt>i</tt>.
    size_t getRowPtr(size_t i) const;

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// \brief Factorize matrix
    int Factor();
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/** \brief Solve the linear system of equations.
 *  \details The default parameters are:
 *  <ul>
 *     <li><tt>CG_SOLVER</tt> for solver
 *     <li><tt>DIAG_PREC</tt> for preconditioner
 *     <li>Max. Number of iterations is 1000
 *     <li>Tolerance is 1.e-8
 *  </ul>
 *  To change these values, call function setSolver before this function
 *  @param [in,out] b Vector that contains right-hand side on input and solution on output
 *
 *  @return Number of actual performed iterations 
 */
    int Solve(class Vect<T_>& b);

/** \brief Solve the linear system of equations.
 *  \details The default parameters are:
 *  <ul>
 *     <li><tt>CG_SOLVER</tt> for solver
 *     <li><tt>DIAG_PREC</tt> for preconditioner
 *     <li>Max. Number of iterations is <tt>1000</tt>
 *     <li>Tolerance is <tt>1.e-8</tt>
 *  </ul>
 *  To change these values, call function setSolver before this function
 *  @param [in] b Vector that contains right-hand side
 *  @param [out] x Vector that contains the obtained solution
 *  @return Number of actual performed iterations 
 */
    int Solve(const Vect<T_>& b,
                    Vect<T_>& x);

/** \brief Choose solver and preconditioner for an iterative procedure
 *  @param [in] solver Option to choose iterative solver in an enumerated variable
 *  <ul>
 *     <li><tt>CG_SOLVER</tt>: Conjugate Gradient [default]
 *     <li><tt>CGS_SOLVER</tt>: Squared conjugate gradient
 *     <li><tt>BICG_SOLVER</tt>: Biconjugate gradient
 *     <li><tt>BICG_STAB_SOLVER</tt>: Biconjugate gradient stabilized
 *     <li><tt>GMRES_SOLVER</tt>: Generalized Minimal Residual
 *     <li><tt>QMR_SOLVER</tt>: Quasi-Minimal Residual
 *  </ul>
 *  Default value is <tt>CG_SOLVER</tt>
 *  @param [in] prec Option to choose preconditioner in an enumerated variable
 *  <ul>
 *     <li><tt>IDENT_PREC</tt>: Identity preconditioner (no preconditioning)
 *     <li><tt>DIAG_PREC</tt>: Diagonal preconditioner [default]
 *     <li><tt>ILU_PREC</tt>: ILU (Incomplete factorization) preconditioner
 *  </ul>
 *  Default value is <tt>DIAG_PREC</tt>
 *  @param [in] max_it Maximum number of allowed iterations. Default value is <tt>1000</tt>.
 *  @param [in] toler Tolerance for convergence. Default value is <tt>1.e-8</tt>
 */
    void setSolver(Iteration      solver=CG_SOLVER,
                   Preconditioner prec=DIAG_PREC,
                   int            max_it=1000,
                   real_t         toler=1.e-8);

/// \brief Return C-Array.
/// \details Non zero terms of matrix is stored row by row.
    T_ *get() const;

/** \brief  Return entry <tt>(i,j)</tt> of matrix if this one is stored, <tt>0</tt> otherwise
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ get(size_t i,
           size_t j) const;

 private:

    int _col_index(size_t i, size_t j) const
    {
       int k = 0;
       while (k<int(_row_ptr[i]-_row_ptr[i-1]))
          if (j==_col_ind[_row_ptr[i-1]+k-1])
             return k;
          else
             k++;
       return -1;
    }

   size_t _dof;
   int _type, _solver, _prec, _max_it;
   real_t _toler;
   int _extended, _one_dof, _sides;
};


///////////////////////////////////////////////////////////////////////////////
//                         I M P L E M E N T A T I O N                       //
///////////////////////////////////////////////////////////////////////////////

#ifndef DOXYGEN_SHOULD_SKIP_THIS

template<class T_>
SpMatrix<T_>::SpMatrix()
{
   _rmin = _cmin = 1, _rmax = _nb_rows, _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
   _is_diagonal = false;
}


template<class T_>
SpMatrix<T_>::SpMatrix(size_t nr,
                       size_t nc)
{
   register size_t i, j;
   _is_diagonal = false;
   _nb_rows = nr;
   _nb_cols = nc;
   _size = 0;
   if (_nb_rows == _nb_cols)
      _size = _nb_rows;
   _length = _nb_rows * _nb_cols;
   _row_ptr.resize(_nb_rows+1);
   _col_ind.resize(_length);
   _row_ptr[0] = 1;
   for (i=1; i<=_nb_rows; i++)
      _row_ptr[i] = _row_ptr[i-1] + _nb_cols;
   size_t l = 0;
   for (i=0; i<_nb_rows; i++)
      for (j=0; j<_nb_cols; j++)
         _col_ind[l++] = j+1;
   _a.resize(_length,T_(0.));
   _rmin = 1, _rmax = _nb_rows;
   _cmin = 1, _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File :" << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
SpMatrix<T_>::SpMatrix(size_t size,
                       int    is_diagonal)
{
   register size_t i;
   _is_diagonal = is_diagonal;
   _size = _nb_rows = _nb_cols = size;
   _length = _size * _size;
   _row_ptr.resize(_nb_rows+1);
   _col_ind.resize(_length);
   _row_ptr[0] = 1;
   for (i=1; i<=_nb_rows; i++)
      _row_ptr[i] = _row_ptr[i-1] + _nb_cols;
   size_t l = 0;
   for (i=0; i<_nb_rows; i++)
      for (size_t j=0; j<_nb_cols; j++)
         _col_ind[l++] = j+1;
   _a.resize(_length,T_(0.));
   _rmin = 1, _rmax = _nb_rows;
   _cmin = 1, _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
SpMatrix<T_>::SpMatrix(class Mesh&  mesh,
                             size_t dof,
                             int    is_diagonal)
{
   _is_diagonal = is_diagonal;
   setMesh(mesh,dof);
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__<< ", Line: " << __LINE__ << endl;
#endif
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
SpMatrix<T_>::SpMatrix(      size_t dof,
                       class Mesh&  mesh,
                             int    code)
{
   _is_diagonal = false;
   setMesh(dof,mesh,mesh.getDOFSupport());
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__<< ", Line: " << __LINE__ << endl;
#endif
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
SpMatrix<T_>::SpMatrix(      size_t dof,
                             size_t nb_eq,
                       class Mesh&  mesh)
{
   _is_diagonal = false;
   setMesh(dof,nb_eq,mesh);
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
   _rmin = _cmin = 1;
   _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
SpMatrix<T_>::SpMatrix(const class vector<std::pair<size_t,size_t> >& I,
                                   int                                opt)
{
   _is_diagonal = false;
   setGraph(I,opt);
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
}


template<class T_>
SpMatrix<T_>::SpMatrix(const class vector<std::pair<size_t,size_t> >& I,
                       const class vector<T_>&                        a,
                                   int                                opt)
{
   _is_diagonal = false;
   size_t n = I.size();
   std::vector<RC> pp(n);
   _nb_rows = _nb_cols = 0;
   for (size_t i=0; i<n; i++) {
      _nb_rows = max(_nb_rows,I[i].first);
      _nb_cols = max(_nb_cols,I[i].second);
      pp[i] = RC(I[i].first-1,I[i].second-1);
   }
   if (_nb_rows == _nb_cols)
      _size = _nb_rows;
   if (opt==0) {
      sort(pp.begin(),pp.end());
      std::vector<RC>::iterator new_end = std::unique(pp.begin(),pp.end());
      pp.erase(new_end,pp.end());
   }
   _row_ptr.resize(_size+1);
   _col_ind.resize(n);
   StoreGraph(_size,pp,_row_ptr,_col_ind);
   _length = pp.size();
   _a.resize(_length);
   for (size_t j=0; j<n; j++)
      _a[_row_ptr[I[j].first-1]+_col_index(I[j].first,I[j].second)-1] = a[j];
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
}


template<class T_>
SpMatrix<T_>::SpMatrix(            size_t          nr,
                                   size_t          nc,
                       const class vector<size_t>& row_ptr, 
                       const class vector<size_t>& col_ind)
             : _type(0), _dof(0)
{
   _is_diagonal = false;
   _nb_rows = nr;
   _nb_cols = nc;
   _size = 0;
   _length = col_ind.size();
   _row_ptr.resize(_nb_rows+1);
   _row_ptr = row_ptr;
   _col_ind.resize(_length);
   _col_ind = col_ind;
   _a.resize(_length,T_(0.));
   _diag.resize(_size);
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
SpMatrix<T_>::SpMatrix(            size_t          nr,
                                   size_t          nc,
                       const class vector<size_t>& row_ptr,
                       const class vector<size_t>& col_ind,
                       const class vector<T_>&     a)
             : _type(0), _dof(0)
{
   _is_diagonal = false;
   _size = 0;
   _nb_rows = nr; _nb_cols = nc;
   _length = col_ind.size();
   _row_ptr.resize(_nb_rows+1);
   _row_ptr = row_ptr;
   _col_ind.resize(_length);
   _col_ind = col_ind;
   _a.resize(_length);
   _a = a;
   _diag.resize(_size);
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
SpMatrix<T_>::SpMatrix(const class vector<size_t>& row_ptr,
                       const class vector<size_t>& col_ind)
             : _type(0), _dof(0)
{
   _is_diagonal = false;
   _size = _nb_rows = _nb_cols = row_ptr.size()-1;
   _length = col_ind.size();
   _row_ptr.resize(_size+1);
   _row_ptr = row_ptr;
   _col_ind.resize(_length);
   _col_ind = col_ind;
   _a.resize(_length,T_(0.));
   _diag.resize(_size);
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
SpMatrix<T_>::SpMatrix(const class vector<size_t>& row_ptr,
                       const class vector<size_t>& col_ind,
                       const class vector<T_>&     a)
             : _type(0), _dof(0)
{
   _is_diagonal = false;
   _size = _nb_rows = _nb_cols = row_ptr.size()-1;
   _length = col_ind.size();
   _row_ptr.resize(_size+1);
   _row_ptr = row_ptr;
   _col_ind.resize(_length);
   _col_ind = col_ind;
   _a.resize(_length);
   _a = a;
   _diag.resize(_size);
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
SpMatrix<T_>::SpMatrix(const SpMatrix& m)
{
   _is_diagonal = m._is_diagonal;
   _dof = m._dof;
   _size = m._size;
   _length = m._length;
   _row_ptr.resize(_size+1);
   _col_ind.resize(_length);
   _col_ind = m._col_ind;
   _row_ptr = m._row_ptr;
   _a.resize(_length);
   _a = m._a;
   _diag.resize(_size);
   _diag = m._diag;
   _rmin = m._rmin; _rmax = m._rmax;
   _cmin = m._cmin; _cmax = m._cmax;
   _solver = CG_SOLVER;
   _prec = DIAG_PREC;
   _max_it = 1000;
   _toler = 1.e-8;
   _solver = m._solver;
   _prec = m._prec;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
SpMatrix<T_>::~SpMatrix(void)
{
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class SpMatrix is deleted in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
void SpMatrix<T_>::setSize(size_t size)
{
   _nb_rows = _nb_cols = _size = size;
   _length = lsize_t(_nb_rows*_nb_cols);
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _row_ptr.resize(_nb_rows+1);
   _col_ind.resize(_length);
   _row_ptr[0] = 1;
   size_t i;
   for (i=1; i<=_nb_rows; i++)
      _row_ptr[i] = _row_ptr[i-1] + _nb_cols;
   size_t l = 0;
   for (i=0; i<_nb_rows; i++)
      for (size_t j=0; j<_nb_cols; j++)
         _col_ind[l++] = j+1;
   _a.resize(_length,T_(0.));
}


template<class T_>
void SpMatrix<T_>::setSize(size_t nr,
                           size_t nc)
{
   _nb_rows = nr;
   _nb_cols = nc;
   _size = 0;
   if (_nb_rows==_nb_cols)
      _size = _nb_rows;
   _length = lsize_t(_nb_rows*_nb_cols);
   _rmin = _cmin = 1; _rmax = _nb_rows; _cmax = _nb_cols;
   _row_ptr.resize(_nb_rows+1);
   _col_ind.resize(_length);
   _row_ptr[0] = 1;
   size_t i;
   for (i=1; i<=_nb_rows; i++)
      _row_ptr[i] = _row_ptr[i-1] + _nb_cols;
   size_t l = 0;
   for (i=0; i<_nb_rows; i++)
      for (size_t j=0; j<_nb_cols; j++)
         _col_ind[l++] = j+1;
   _a.resize(_length,T_(0.));
}


template<class T_>
void SpMatrix<T_>::setMesh(class Mesh&  mesh,
                                 size_t dof)
{
   _dof_type = mesh.getDOFSupport();
   Matrix<T_>::init_set_mesh(mesh,dof);
   if (_dof_type==NODE_DOF) {
      if (_dof_type && dof)
         _length = XGraph(mesh, _row_ptr, _col_ind);
      else if (dof)
         _length = NodeGraphScal(mesh, _row_ptr, _col_ind);
      else if (dof==0)
         _length = NodeGraph(mesh,_row_ptr,_col_ind);
   }

   else if (_dof_type==SIDE_DOF)
      _length = SideGraph(mesh,_row_ptr,_col_ind);

   else if (_dof_type==ELEMENT_DOF)
      _length = ElementGraph(mesh, _row_ptr,_col_ind);

   else
      ;

   _a.resize(_length,T_(0.));
   _diag.resize(_size);
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void SpMatrix<T_>::setMesh(class Mesh &mesh, size_t dof, size_t type)
{
   _dof_type = mesh.getDOFSupport();
   Matrix<T_>::init_set_mesh(mesh,dof);
   if (_dof_type==NODE_DOF) {
      if (type && dof)
         _length = XGraph(mesh, _row_ptr, _col_ind);
      else if (dof)
         _length = NodeGraphScal(mesh, _row_ptr, _col_ind);
      else if (type==0 && dof==0)
         _length = NodeGraph(mesh,_row_ptr,_col_ind);
   }
   else if (_dof_type==SIDE_DOF)
      _length = SideGraph(mesh,_row_ptr,_col_ind);

   else if (_dof_type==ELEMENT_DOF)
      _length = ElementGraph(mesh, _row_ptr,_col_ind);

   else;
   _a.resize(_length,T_(0.));
   _diag.resize(_size);
}


template<class T_>
void SpMatrix<T_>::setMesh(size_t dof, class Mesh &mesh, int code)
{
   _size = _nb_rows = _nb_cols = mesh.getNbEq();
   if (dof)
      _size = _nb_rows = _nb_cols = mesh.getNbNodes();
   if (code!=0)
      _length = XGraph(mesh, _row_ptr, _col_ind);
   else
      _length = NodeGraphScal(mesh, _row_ptr, _col_ind);
   _a.resize(_length,T_(0.));
   _diag.resize(_size);
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void SpMatrix<T_>::setMesh(size_t dof, size_t nb_eq, class Mesh &mesh)
{
   _type = 0;
   _dof = 0;
   _size = _nb_rows = _nb_cols = nb_eq;
   _length = NodeGraphScal(mesh, dof, nb_eq, _row_ptr, _col_ind);
   _a.resize(_length,T_(0.));
   _diag.resize(_size);
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void SpMatrix<T_>::ExtendedGraph()
{
   _extended = true;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */
   

template<class T_>
void SpMatrix<T_>::setGraph(const vector<std::pair<size_t,size_t> >& I,
                                  int                                opt)
{
   size_t i;
   _length = I.size();
   _nb_rows = _nb_cols = 0;
   vector<RC> pp(I.size());
   for (i=0; i<_length; i++) {
      _nb_rows = max(_nb_rows,I[i].first);
      _nb_cols = max(_nb_cols,I[i].second);
      pp[i] = RC(I[i].first-1,I[i].second-1);
   }
   _size = _nb_rows;
   if (opt==0) {
      sort(pp.begin(),pp.end());
      vector<RC>::iterator new_end = std::unique(pp.begin(),pp.end());
      pp.erase(new_end,pp.end());
   }
   _length = pp.size();
   _row_ptr.resize(_size+1);
   _col_ind.resize(_length);
   for (size_t i=0; i<_size; i++)
      _row_ptr[i] = 0;
   for (size_t j=0; j<_length; j++) {
      _row_ptr[pp[j].first]++;
      _col_ind[j] = pp[j].second + 1;
   }
   for (size_t k=1; k<_size; k++)
      _row_ptr[k] += _row_ptr[k-1];
   for (int ii=_size; ii>0; ii--)
      _row_ptr[ii] = _row_ptr[ii-1] + 1;
   _row_ptr[0] = 1;
   _a.resize(_length,0);
}


template<class T_>
void SpMatrix<T_>::setOneDOF()
{
   _one_dof = true;
}


template<class T_>
void SpMatrix<T_>::setSides()
{
   _sides = true; 
}


template<class T_>
void SpMatrix<T_>::setDiag()
{
   for (size_t i=0; i<_size; i++) {
      for (size_t j=0; j<_row_ptr[i+1]-_row_ptr[i]; j++) {
         if (i==j) {
            _diag[i] = _a[_row_ptr[i]+i-1];
            break;
         }
      }
   }
}


template<class T_>
Vect<T_> SpMatrix<T_>::getColumn(size_t j) const
{
   Vect<T_> v(_nb_rows);
   for (size_t i=1; i<=_nb_rows; i++)
      v(i) = get(i,j);           
   return v;
}


template<class T_>
void SpMatrix<T_>::DiagPrescribe(      class Mesh&     mesh,
                                             Vect<T_>& b,
                                 const       Vect<T_>& u)
{
   real_t p = 0;
   for (size_t j=1; j<=_nb_rows; j++)
      p = max(p,Abs(get(j,j)));

   size_t k=0;
   mesh_nodes(mesh) {
      for (size_t i=1; i<=The_node.getNbDOF(); ++i) {
         size_t ii = The_node.getDOF(i)-1;
         for (size_t j=0; j<_row_ptr[ii+1]-_row_ptr[ii]; j++,k++)
            if (The_node.getCode(i)>0) {
               b[ii] = p*u[ii];
               _a[k] = 0;
               if (ii+1==_col_ind[k])
                  _a[k] = p;
            }
      }
   }
}


template<class T_>
void SpMatrix<T_>::DiagPrescribe(      class Vect<T_>& b,
                                 const class Vect<T_>& u)
{
   real_t p = 0;
   for (size_t j=1; j<=_nb_rows; j++)
      p = max(p,Abs(get(j,j)));

   size_t k=0;
   MESH_ND {
      for (size_t i=1; i<=TheNode.getNbDOF(); ++i) {
         size_t ii = TheNode.getDOF(i)-1;
         for (size_t j=0; j<_row_ptr[ii+1]-_row_ptr[ii]; j++,k++)
            if (TheNode.getCode(i)>0) {
               b[ii] = p*u[ii];
               _a[k] = 0;
               if (ii+1==_col_ind[k])
                  _a[k] = p;
            }
      }
   }
}


template<class T_>
Vect<T_> SpMatrix<T_>::getRow(size_t i) const
{
   return getColumn(i);
}


template<class T_>
T_ & SpMatrix<T_>::operator()(size_t i,
                              size_t j)
{
   int k = _col_index(i,j);
   try {
      if (k<0)
         THROW_RT("set(i,j,x): Index pair : (" + itos(i) + "," + itos(j) + ") is not compatible "
                  "with sparse storage.");
      else
         return _a[_row_ptr[i-1]+k-1];
   }
   CATCH("SpMatrix");
   return _temp;
}


template<class T_>
T_ SpMatrix<T_>::operator()(size_t i,
                            size_t j) const
{
   int k = _col_index(i,j);
   if (k<0)
      return _zero;
   else
      return _a[_row_ptr[i-1]+k-1];
}


template<class T_>
const T_ SpMatrix<T_>::operator()(size_t i) const
{
   return _a[i-1];
}


template<class T_>
const T_ SpMatrix<T_>::operator[](size_t i) const
{
   return _a[i];
}


template<class T_>
void SpMatrix<T_>::getMesh(class Mesh& mesh)
{
   if (_sides)
      SideGraph(mesh,_row_ptr,_col_ind);
   else {
      if (_dof)
         _size = _nb_rows = _nb_cols = mesh.getNbNodes();
      else
         _size = _nb_rows = _nb_cols = mesh.getNbEq();
      if (_type)
         XGraph(mesh,_row_ptr,_col_ind);
//             XGraphScal(mesh,_row_ptr,_col_ind);
      else
         NodeGraph(mesh,_row_ptr,_col_ind);
   }
   _a.resize(_length,T_(0.));
   _diag.resize(_size);
}


template<class T_>
void SpMatrix<T_>::Mult(const Vect<T_>& x,
                              Vect<T_>& y) const
{
   y = static_cast<T_>(0);
   MultAdd(x,y);
}


template<class T_>
void SpMatrix<T_>::MultAdd(const Vect<T_>& x,
                                 Vect<T_>& y) const
{
   size_t l = 0;
   for (size_t i=0; i<_nb_rows; ++i)
      for (size_t j=0; j<_row_ptr[i+1]-_row_ptr[i]; ++j)
         y[i] += _a[_row_ptr[i]+j-1] * x(_col_ind[l++]);
}


template<class T_>
void SpMatrix<T_>::MultAdd(T_ a, const class Vect<T_> &x, class Vect<T_> &y) const
{
   size_t l = 0;
   for (size_t i=0; i<_nb_rows; i++)
      for (size_t j=0; j<_row_ptr[i+1]-_row_ptr[i]; j++)
         y[i] += a * _a[_row_ptr[i]+j-1] * x(_col_ind[l++]);
}


template<class T_>
void SpMatrix<T_>::TMult(const Vect<T_>& x,
                               Vect<T_>& y) const
{
   for (size_t i=0; i<_nb_rows; i++)
      for (size_t j=0; j<_row_ptr[i+1]-_row_ptr[i]; j++)
         y[_col_ind[j+_row_ptr[i]-1]-1] += _a[_row_ptr[i]+j-1] * x[i];
}


template<class T_>
void SpMatrix<T_>::set(      size_t i,
                             size_t j,
                       const T_&    val)
{
   int k = _col_index(i,j);
   try { 
      if (k<0)
         THROW_RT("set(i,j,x): Index pair : (" + itos(i) + "," + itos(j) + ") is not compatible "
                  "with sparse storage.");
      else
         _a[_row_ptr[i-1]+_col_index(i,j)-1] = val;
   }
   CATCH("SpMatrix");
}


template<class T_>
void SpMatrix<T_>::add(      size_t i,
                             size_t j,
                       const T_&    val)
{
   _a[_row_ptr[i-1]+_col_index(i,j)-1] += val;
}


template<class T_>
void SpMatrix<T_>::Axpy(      T_            a,
                        const SpMatrix<T_>& m)
{
   _a += a * m._a;
}


template<class T_>
void SpMatrix<T_>::Axpy(      T_          a,
                        const Matrix<T_>* m)
{
   for (size_t i=0; i<_length; i++)
      _a[i] += a * m->_a[i];
}


template<class T_>
void SpMatrix<T_>::operator=(const T_& x)
{
   for (size_t i=0; i<_length; i++)
      _a[i] = x;
}


template<class T_>
size_t SpMatrix<T_>::getColInd(size_t i) const
{
#ifdef _BOUNDS_DEBUG
   assert(i<=_length && i>0);
#endif
   return _col_ind[i-1];
}


template<class T_>
size_t SpMatrix<T_>::getRowPtr(size_t i) const
{
#ifdef _BOUNDS_DEBUG
   assert(i<=_nb_rows+1 && i>0);
#endif
   return _row_ptr[i-1];
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
int SpMatrix<T_>::Factor()
{
   return -1;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
int SpMatrix<T_>::Solve(class Vect<T_> &b)
{
   if (_is_diagonal) {
      for (size_t i=0; i<_size; i++)
         b[i] /= _a[i];
      return 0;
   }
   Vect<T_> x(b.size());
   int nb_it = Solve(b,x);
   b = x;
   return nb_it;
}


template<class T_>
int SpMatrix<T_>::Solve(const Vect<T_>& b,
                              Vect<T_>& x)
{
   if (_is_diagonal) {
      for (size_t i=0; i<_size; i++)
         x[i] = b[i]/_a[i];
      return 0;
   }
  
   Prec<T_> *p = new Prec<T_>(*this,_prec);
   int nb_it = 0;

   switch (_solver) {

      case CG_SOLVER:
         nb_it = CG(*this,*p,b,x,_max_it,_toler,1);
         break;

      case BICG_SOLVER:
         nb_it = BiCG(*this,*p,b,x,_max_it,_toler,1);
         break;

      case CGS_SOLVER:
         nb_it = CGS(*this,*p,b,x,_max_it,_toler,1);
         break;

      case BICG_STAB_SOLVER:
         nb_it = BiCGStab(*this,*p,b,x,_max_it,_toler,1);
         break;

      case GMRES_SOLVER:
         nb_it = GMRes(*this,*p,b,x,5,_max_it,_toler,1);
         break;

      case QMR_SOLVER:
         nb_it = QMR(*this,*p,*p,b,x,_max_it,_toler,1);
         break;
   }
   delete p;
   return nb_it;
}


template<class T_>
void SpMatrix<T_>::setSolver(Iteration      solver,
                             Preconditioner prec,
                             int            max_it,
                             real_t         toler)
{
   _solver = solver;
   _prec = prec;
   _max_it = max_it;
   _toler = toler;
}


template<class T_>
T_ *SpMatrix<T_>::get() const
{
   return &_a[0];
}


template<class T_>
T_ SpMatrix<T_>::get(size_t i, size_t j) const
{
   int k = _col_index(i,j);
   if (k<0)
      return _zero;
   else
      return _a[_row_ptr[i-1]+k-1];
}


///////////////////////////////////////////////////////////////////////////////
//                           ASSOCIATED  FUNCTIONS                           //
///////////////////////////////////////////////////////////////////////////////


/** \fn ostream & operator<<(ostream& s, const SpMatrix<T_> &a)
 *  \ingroup VectMat
 *  Output matrix in output stream
 */
template<class T_>
ostream& operator<<(      ostream&      s,
                    const SpMatrix<T_>& a)
{
   size_t rmin, rmax, cmin, cmax;
   a.getPrintView(rmin,rmax,cmin,cmax);
   s.setf(ios::right|ios::scientific);
   s << endl;
   for (size_t i=rmin; i<=rmax; i++) {
      s << "\nRow:  " << setw(6) << i << endl;
      for (size_t j=cmin; j<=cmax; j++)
         s << "  " << setprecision(8) << std::setfill(' ') << setw(18) << a(i,j);
      s << endl;
   }
   return s;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

} /* namespace OFELI */

#endif
