/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                   Template Class Matrix for tridiagonal matrices

  ==============================================================================*/


#ifndef _TR_MATRIX_H
#define _TR_MATRIX_H

#include "linear_algebra/Matrix.h"

namespace OFELI {

/*! \file TrMatrix.h
 *  \brief Definition file for class TrMatrix.
 */

template<class T_> class TrMatrix;

/*! \class TrMatrix
 *  \ingroup VectMat
 * \brief To handle tridiagonal matrices.
 *
 * This class enables storing and manipulating tridiagonal matrices.
 * The template parameter is the type of matrix entries
 *
 * \tparam <T_> Data type (double, float, complex<double>, ...)
 */


template<class T_>
class TrMatrix : public Matrix<T_>
{

 public:

   using Matrix<T_>::operator();
   using Matrix<T_>::_rmin;
   using Matrix<T_>::_cmin;
   using Matrix<T_>::_rmax;
   using Matrix<T_>::_cmax;
   using Matrix<T_>::_size;
   using Matrix<T_>::_nb_cols;
   using Matrix<T_>::_nb_rows;
   using Matrix<T_>::_length;
   using Matrix<T_>::_a;
   using Matrix<T_>::_fact;
   using Matrix<T_>::_zero;
   using Matrix<T_>::_temp;
   using Matrix<T_>::_ch;

/// \brief Default constructor.
/// \details Initialize a zero dimension tridiagonal matrix
    TrMatrix();

/// \brief Constructor for a tridiagonal matrix with <tt>size</tt> rows.
    TrMatrix(size_t size);

/// \brief Copy Constructor
    TrMatrix(const TrMatrix& m);

/// \brief Destructor
    ~TrMatrix();

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(class Mesh&  mesh,
                       size_t dof=0);
    void setMesh(class Mesh&  mesh,
                       size_t dof,
                       size_t type);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    virtual void setMesh(      size_t dof,
                         class Mesh&  mesh,
                               int    code=0);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    virtual void setMesh(      size_t dof,
                               size_t nb_eq,
                         class Mesh&  mesh);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setDiag() { }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/// \brief Set size (number of rows) of matrix.
/// @param [in] size Number of rows and columns.
    void setSize(size_t size);

/// \brief Multiply matrix by vector <tt>x</tt> and add result to <tt>y</tt>.
    void MultAdd(const Vect<T_>& x, Vect<T_>& y) const;

/// \brief Multiply matrix by vector <tt>a*x</tt> and add result to <tt>y</tt>.
    void MultAdd(      T_        a,
                 const Vect<T_>& x,
                       Vect<T_>& y) const;

/// \brief Multiply matrix by vector <tt>x</tt> and save result in <tt>y</tt>.
    void Mult(const Vect<T_>& x,
                    Vect<T_>& y) const;

/// \brief Multiply transpose of matrix by vector <tt>x</tt> and save result in <tt>y</tt>.
    void TMult(const Vect<T_>& x,
                     Vect<T_>& y) const;
   
/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] m %Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(      T_            a,
              const TrMatrix<T_>& m);
   
/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] m %Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(      T_          a,
              const Matrix<T_>* m);

/// \brief Assign constant <tt>val</tt> to an entry <tt>(i,j)</tt> of the matrix.
    void set(      size_t i,
                   size_t j,
             const T_&    val);

/// \brief Add constant <tt>val</tt> value to an entry <tt>(i,j)</tt> of the matrix.
    void add(      size_t i,
                   size_t j,
             const T_&    val);

/** \brief Operator () (Constant version).
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ operator()(size_t i,
                  size_t j) const;

/** \brief Operator () (Non constant version).
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ & operator()(size_t i,
                    size_t j);

/// \brief Operator =.
/// \details Copy matrix <tt>m</tt> to current matrix instance.
    TrMatrix<T_> & operator=(const TrMatrix<T_>& m);

/// \brief Operator =.
/// \details Assign constant value <tt>x</tt> to all matrix entries.
    TrMatrix<T_> & operator=(const T_& x);

/// \brief Operator *=.
/// \details Premultiply matrix entries by constant value <tt>x</tt>.
    TrMatrix<T_> & operator*=(const T_& x);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    int Factor();
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/** \brief Solve a linear system with current matrix (forward and back substitution).
 *  @param [in] b Vect instance that contains right-hand side on input and solution on output.
 *  @return
 *  <ul>
 *     <li><tt>0</tt> if solution was normally performed,
 *     <li><tt>n</tt> if the <tt>n</tt>-th pivot is null.
 *  </ul>
 * 
 *  \b Warning: %Matrix is modified after this function.
 */
    int Solve(Vect<T_>& b);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// Return storage information
    size_t getColInd(size_t i) const;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// Return storage information
    size_t getRowPtr(size_t i) const;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/// \brief Return C-Array.
    T_ *get() const { return &_a[0]; }

/// \brief Return entry <tt>(i,j)</tt> of matrix
    T_ get(size_t i, size_t j) const;

 private:

   size_t _imin, _imax;
};

///////////////////////////////////////////////////////////////////////////////
//                         I M P L E M E N T A T I O N                       //
///////////////////////////////////////////////////////////////////////////////

template<class T_>
TrMatrix<T_>::TrMatrix()
{
   _size = 0;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class TrMatrix is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}

template<class T_>
TrMatrix<T_>::TrMatrix(size_t size)
{
   _size = _nb_rows = _nb_cols = size;
   _length = 3*_size;
   _a.resize(_length,static_cast<T_>(0));
   _ch.resize(_size);
   _ch[0] = 0;
   for (size_t i=1; i<_size; i++)
      _ch[i] = i+1;
}

template<class T_>
TrMatrix<T_>::TrMatrix(const TrMatrix& m)
{
   _size = m._size;
   _length = m._length;
   _a.resize(_length);
   _a = m._a;
}

template<class T_>
TrMatrix<T_>::~TrMatrix()
{
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class TrMatrix is destructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void TrMatrix<T_>::setMesh(class Mesh&  mesh,
                                 size_t dof)
{
   try {
      THROW_RT("setMesh(Mesh,size_t): This member function is not valid for class TrMatrix");
   }
   CATCH("TrMatrix");
}

template<class T_>
void TrMatrix<T_>::setMesh(class Mesh&  mesh,
                                 size_t dof,
                                 size_t type)
{
   try {
      THROW_RT("setMesh(Mesh,size_t,size_t): This member function is not valid for class TrMatrix");
   }
   CATCH("TrMatrix");
}


template<class T_>
void TrMatrix<T_>::setMesh(      size_t dof,
                           class Mesh&  mesh,
                                 int    code)
{
   try {
      THROW_RT("setMesh(Mesh,Mesh,int): This member function is not valid for class TrMatrix");
   }
   CATCH("TrMatrix");
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void TrMatrix<T_>::setMesh(      size_t dof,
                                 size_t nb_eq, 
                           class Mesh&  mesh)
{
// This is just to avoid warning on unused variable
   dof = 0;
   nb_eq = 0;
   if (mesh.getDim()==0) { }
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

template<class T_>
void TrMatrix<T_>::setSize(size_t size)
{
   _nb_rows = _nb_cols = _size = size;
   _length = 3 * _nb_rows;
   _a.resize(_length);
}


template<class T_>
void TrMatrix<T_>::MultAdd(const Vect<T_>& x,
                                 Vect<T_>& y) const
{
   y(1) += _a[1]*x(1) + _a[2]*x(2);
   for (size_t i=2; i<_size; ++i)
      y(i) += _a[3*i-3]*x(i-1) + _a[3*i-2]*x(i) + _a[3*i-1]*x(i+1);
   y(_size) += _a[3*_size-3]*x(_size-1) + _a[3*_size-2]*x(_size);
}


template<class T_>
void TrMatrix<T_>::MultAdd(      T_        a,
                           const Vect<T_>& x,
                                 Vect<T_>& y) const
{
   y(1) += a*(_a[1]*x(1) + _a[2]*x(2));
   for (size_t i=2; i<_size; ++i)
      y(i) += a * (_a[3*i-3]*x(i-1) + _a[3*i-2]*x(i) + _a[3*i-1]*x(i+1));
   y(_size) += a * (_a[3*_size-3]*x(_size-1) + _a[3*_size-2]*x(_size));
}


template<class T_>
void TrMatrix<T_>::Mult(const Vect<T_>& x,
                              Vect<T_>& y) const
{
   y = static_cast<T_>(0);
   MultAdd(x,y);
}


template<class T_>
void TrMatrix<T_>::TMult(const Vect<T_>& x,
                               Vect<T_>& y) const
{
   y(1) = _a[1]*x(1) + _a[3]*x(2);
   for (size_t i=2; i<_size; ++i)
      y(i) = _a[3*i-4]*x(i-1) + _a[3*i-2]*x(i) + _a[3*i]*x(i+1);
   y(_size) = _a[3*_size-4]*x(_size-1) + _a[3*_size-2]*x(_size);
}


template<class T_>
void TrMatrix<T_>::Axpy(      T_            a,
                        const TrMatrix<T_>& m)
{
  //  Axpy(a,m._a,_a);
   _a += a * m._a;
}


template<class T_>
void TrMatrix<T_>::Axpy(      T_          a,
                        const Matrix<T_>* m)
{
   for (size_t i=0; i<_length; i++)
      _a[i] += a * m->_a[i];
}


template<class T_>
void TrMatrix<T_>::set(      size_t i,
                             size_t j,
                       const T_&    val)
{
   if (i==j)
      _a[3*i-2] = val;
   else if (i==j+1)
      _a[3*i-3] = val;
   else if (i==j-1)
      _a[3*i-1] = val;
}


template<class T_>
void TrMatrix<T_>::add(      size_t i,
                             size_t j,
                       const T_&    val)
{
   if (i==j)
      _a[3*i-2] += val;
   else if (i==j+1)
      _a[3*i-3] += val;
   else if (i==j-1)
      _a[3*i-1] += val;
}


template<class T_>
T_ TrMatrix<T_>::operator()(size_t i,
                            size_t j) const
{
#ifdef _BOUNDS_DEBUG
   assert(i==j || i==j+1 || i==j-1);
#endif
   if (i==j)
      return _a[3*i-2];
   else if (i==j+1)
      return _a[3*i-3];
   else if (i==j-1)
      return _a[3*i-1];
   else
      return _zero;
}


template<class T_>
T_ & TrMatrix<T_>::operator()(size_t i,
                              size_t j)
{
#ifdef _BOUNDS_DEBUG
   assert(i==j || i==j+1 || i==j-1);
#endif
   try {
      if (i==j)
         return _a[3*i-2];
      else if (i==j+1)
         return _a[3*i-3];
      else if (i==j-1)
         return _a[3*i-1];
      else {
         THROW_RT("Operator(): Index pair (" + itos(i) + "," + itos(j) +
                  ") is not compatible with tridiagonal structure.");
      }
   }
   CATCH("TrMatrix");
   return _zero;
}


template<class T_>
TrMatrix<T_> & TrMatrix<T_>::operator=(const TrMatrix<T_>& m)
{
   _a = m._a;
   return *this;
}


template<class T_>
TrMatrix<T_> & TrMatrix<T_>::operator=(const T_& x)
{
   Clear(_a);
   for (size_t i=1; i<=_size; ++i)
      _a[3*i-2] = x;
   return *this;
}


template<class T_>
TrMatrix<T_> & TrMatrix<T_>::operator*=(const T_& x)
{
   for (size_t i=1; i<=_length; i++)
      _a[i-1] *= x;
   return *this;
}

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
int TrMatrix<T_>::Factor()
{
   try {
      THROW_RT("Factor(): Matrix cannot be fatorized separately Call Solve(b) directly.");
   }
   CATCH("TrMatrix");
   return 1;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
int TrMatrix<T_>::Solve(Vect<T_>& b)
{
   T_ p=0;
   for (size_t i=1; i<_size; ++i) {
      try {
         if (Abs(_a[3*i-2]) < OFELI_EPSMCH)
            THROW_RT("Solve(Vect<T_>): The " + itos(i) + "-th pivot is null."); 
         else
            p = _a[3*i]/_a [3*i-2];
      }
      CATCH("TrMatrix");
      _a[3*i+1] -= p*_a[3*i-1];
      b.add(i+1,-p*b(i));
   }
   try {
      if (Abs(_a[3*_size-2]) < OFELI_EPSMCH)
         THROW_RT("Solve(Vect<T_>): The " + itos(_size) + "-th pivot is null.");
      else
         b(_size) /= _a[3*_size-2];
   }
   CATCH("TrMatrix");
   for (int j=_size-1; j>0; j--)
      b(j) = (b(j)-_a[3*j-1]*b(j+1))/_a[3*j-2];
   return 0;
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
size_t TrMatrix<T_>::getColInd(size_t i) const
{
#ifdef _BOUNDS_DEBUG
   assert(i<=_length && i>0);
#endif
   return _ch[i-1];
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
size_t TrMatrix<T_>::getRowPtr(size_t i) const
{
#ifdef _BOUNDS_DEBUG
   assert(i<=_nb_rows+1 && i>0);
#endif
   return _ch[i-1];
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
T_ TrMatrix<T_>::get(size_t i,
                     size_t j) const
{
   if (i==j)
      return _a[3*i-2];
   else if (i==j+1)
      return _a[3*i-3];
   else if (i==j-1)
      return _a[3*i-1];
   else
      return _zero;
}

///////////////////////////////////////////////////////////////////////////////
//                           ASSOCIATED  FUNCTIONS                           //
///////////////////////////////////////////////////////////////////////////////


/** \fn TrMatrix<T_> operator*(T_ a, const TrMatrix<T_> &A)
 *  \brief Operator * (Premultiplication of matrix by constant)
 *  \ingroup VectMat
 *  @return a*A
 */
template<class T_>
TrMatrix<T_> operator*(      T_            a,
                       const TrMatrix<T_>& A)
{
   TrMatrix<T_> v(A);
   for (size_t i=0; i<A.getLength(); ++i)
      v[i] *= a;
   return v;
}

/** \fn ostream& operator<<(ostream& s, const TrMatrix<T_>& a)
 *  \ingroup VectMat
 *  \brief Output matrix in output stream
 */
template<class T_>
ostream& operator<<(      ostream&      s,
                    const TrMatrix<T_>& a)
{
   size_t rmin, rmax, cmin, cmax;
   a.getPrintView(rmin,rmax,cmin,cmax);
   s.setf(ios::right|ios::scientific);
   s << endl;
   for (size_t i=rmin; i<=rmax; i++) {
      s << "\nRow  " << setw(6) << i << endl;
      for (size_t j=cmin; j<=cmax; j++)
         s << "  " << setprecision(8) << std::setfill(' ') << setw(18) << a(i,j);
      s << endl;
   }
   return s;
}

} /* namespace OFELI */

#endif
