/*==============================================================================

                                    O  F  E  L  I

                            Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                  Definition of class 'PETScMatrix' for Sparse matrix
                              using the PETSc library

  ==============================================================================*/


#ifndef __PETSC_MATRIX_H
#define __PETSC_MATRIX_H

#ifdef USE_PETSC

#include "petscmat.h"
#include "petsc.h"

#include "mesh/Mesh.h"
#include "linear_algebra/GraphOfMatrix.h"
#include "linear_algebra/petsc/PETScVect.h"
#include "util/util.h"

namespace OFELI {

/*! \file PETScMatrix.h
 *  \brief Definition file for class PETScMatrix.
 */


/*! \class PETScMatrix
 *  \ingroup VectMat
 *  \brief To handle matrices in sparse storage format.
 *
 * This template class enables to store and manipulate a sparse matrix,
 * <i>i.e.</i> only nonzero terms are stored. Internally, the matrix is stored as a
 * vector instance and uses for the definition of its graph a <tt>Vect<size_t></tt>
 * instance row_ptr and a Vect<size_t> instance <tt>col_ind</tt> that contains
 * respectively addresses of first element of each row and column indices.
 *
 * To illustrate this, consider the matrix
 * \verbatim
            1   2   0
            3   4   0
            0   5   0
   \endverbatim
 *
 * Such a matrix is stored in the vector<real_t> instance {1,2,3,4,5}.
 * The vectors <tt>row_ptr</tt> and <tt>col_ind</tt> are respectively:
 * <tt>{1,3,5,6}</tt>, <tt>{1,2,1,2,2}</tt>
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 */

template<class T_> class PETScVect;

template<class T_> class PETScMatrix
{

 public:

/// \brief Default constructor.
/// \details Initialize a zero-dimension matrix
    PETScMatrix();

/** \brief Constructor that initializes current instance as a dense matrix.
 *  \details Normally, for a dense matrix this is not the right class.
 *  @param [in] nr Number of matrix rows.
 *  @param [in] nc Number of matrix columns.
 */
    PETScMatrix(size_t nr,
                size_t nc);

/** \brief Constructor that initializes current instance as a dense matrix.
 *  \details Normally, for a dense matrix this is not the right class.
 *  @param [in] size Number of matrix rows (and columns).
 */
    PETScMatrix(size_t size);

/** \brief Constructor using a Mesh instance
 *  @param [in] mesh Mesh instance from which matrix graph is extracted.
 *  @param [in] dof Option parameter, with default value <tt>0</tt>.\n
 *  <tt>dof=1</tt> means that only one degree of freedom for each node (or element or side)
 *  is taken to determine matrix structure. The value <tt>dof=0</tt> means that matrix 
 *  structure is determined using all DOFs.
 */
    PETScMatrix(class Mesh&  mesh,
                      size_t dof=0);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    PETScMatrix(      size_t dof,
                class Mesh&  mesh,
                      int    code=0);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    PETScMatrix(      size_t dof,
                      size_t nb_eq,
                class Mesh&  mesh);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/** \brief Constructor for a square matrix using non zero row and column indices
 *  @param [in] I Vector containing pairs of row and column indices
 *  @param [in] opt Flag indicating if vectors I is cleaned and ordered
 *  (opt=1) or not (opt=0). In the latter case, this vector can have
 *  the same contents more than once and are not necessarily ordered
 */
    PETScMatrix(const class vector<std::pair<size_t,size_t> >& I,
                            int                                opt=1);

/// \brief Copy constructor
    PETScMatrix(const PETScMatrix& m);

/// \brief Destructor
    ~PETScMatrix(void);

/** \brief Determine mesh graph and initialize matrix.
 *  \details This member function is called by constructor with the same arguments
 *  @param [in] mesh Mesh instance for which matrix graph is determined.
 *  @param [in] dof Option parameter, with default value <tt>0</tt>.\n
 *  <tt>dof=1</tt> means that only one degree of freedom for each node (or element or side)
 *  is taken to determine matrix structure. The value <tt>dof=0</tt> means that matrix
 *  structure is determined using all DOFs.
 */
    void setMesh(class Mesh& mesh,
                      size_t dof=0);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(      size_t dof,
                 class Mesh&  mesh,
                       int    code=0);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    void setMesh(      size_t dof, 
                       size_t nb_eq,
                 class Mesh&  mesh);
    void setMesh(class Mesh&  mesh,
                       size_t dof,
                       size_t type);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** \brief Activate extended graph option.
 *  \details An extended graph is the one
 *  for which a node is linked not only to its neighbor nodes but also to the
 *  neighbors of the neighbors.
 */
    void ExtendedGraph();
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/// \brief Activate 1-DOF per node option.
    void setOneDOF();

/// \brief Activate Sides option.
    void setSides();

/** \brief Impose by a diagonal method an essential boundary condition.
 *  \details This member function modifies diagonal terms in matrix and terms
 *  in vector that correspond to degrees of freedom with nonzero code
 *  in order to impose a boundary condition.
 *  The penalty parameter is defined by default equal to 1.e20.
 *  It can be modified by member function \b setPenal(..).
 *  @param [in] mesh Mesh instance from which information is extracted.
 *  @param [in,out] b PETScVect instance that contains right-hand side.
 *  @param [in] u PETScVect instance that conatins imposed valued at DOFs where they are to be imposed.
 */
    void DiagPrescribe(class Mesh&          mesh,
                             PETScVect<T_>& b,
                       const PETScVect<T_>& u);

/** \brief Impose by a diagonal method an essential boundary condition using the Mesh instance
 *  provided by the constructor
 *  \details This member function modifies diagonal terms in matrix and terms
 *  in vector that correspond to degrees of freedom with nonzero code
 *  in order to impose a boundary condition.
 *  The penalty parameter is defined by default equal to 1.e20.
 *  It can be modified by member function \b setPenal(..).
 *  @param [in,out] b PETScVect instance that contains right-hand side.
 *  @param [in] u PETScVect instance that conatins imposed valued at DOFs where they are to be imposed.
 */
    void DiagPrescribe(      PETScVect<T_>& b,
                       const PETScVect<T_>& u);

/// \brief Set size of matrix (case where it's a square matrix).
/// @param [in] size Number of rows and columns.
    void setSize(size_t size);

/** \brief Set size (number of rows) of matrix
 *  @param [in] nr Number of rows
 *  @param [in] nc Number of columns
 */
   void setSize(size_t nr,
                size_t nc);

/** \brief Set graph of matrix by giving a vector of its nonzero entries
 *  @param [in] I Vector containing pairs of row and column indices
 *  @param [in] opt Flag indicating if vector <tt>I</tt> is cleaned and ordered
 *  (<tt>opt=1</tt>: default) or not (<tt>opt=0</tt>). In the latter case, this vector can have
 *  the same contents more than once and are not necessarily ordered
 */
    void setGraph(const vector<std::pair<size_t,size_t> >& I,
                        int                                opt=1);

/** \brief Operator <tt>()</tt>
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ operator()(size_t i,
                  size_t j) const;

/// \brief Return number of matrix rows
    size_t getNbRows() const { return _nb_rows; }

/// \brief Return number of matrix columns
    size_t getNbColumns() const { return _nb_cols; }

/// \brief Return length of matrix
/// \details The length is the total number of stored elements in the matrix
    size_t getLength() const { return _length; }

/// \brief Get mesh instance whose reference will be stored in current instance of PETScMatrix.
    void getMesh(class Mesh& mesh);

/** \brief Multiply matrix by vector and save in another one.
 *  @param [in] x Vector to multiply by matrix
 *  @param [out] y Vector that contains on output the result.
 */
    void Mult(const PETScVect<T_>& x,
                    PETScVect<T_>& y) const;

/** \brief Multiply matrix by vector <tt>x</tt> and add to <tt>y</tt>.
 *  @param [in] x Vector to multiply by matrix
 *  @param [out] y Vector to add to the result. <tt>y</tt> contains on output the result.
 */
    void MultAdd(const PETScVect<T_>& x,
                       PETScVect<T_>& y) const;

/** \brief Multiply matrix by vector <tt>a*x</tt> and add to <tt>y</tt>.
 *  @param [in] a Constant to multiply by matrix
 *  @param [in] x Vector to multiply by matrix
 *  @param [out] y Vector to add to the result. <tt>y</tt> contains on output the result.
 */
    void MultAdd(      T_        a,
                 const PETScVect<T_>& x,
                       PETScVect<T_>& y) const;

/** \brief Multiply transpose of matrix by vector <tt>x</tt> and save in <tt>y</tt>.
 *  @param [in] x Vector to multiply by matrix
 *  @param [out] y Vector that contains on output the result.
 */
    void TMult(const PETScVect<T_>& x,
                     PETScVect<T_>& y) const;

/** \brief Add to matrix the product of a matrix by a scalar
 *  @param [in] a Scalar to premultiply
 *  @param [in] m %Matrix by which <tt>a</tt> is multiplied. The result is added
 *  to current instance
 */
    void Axpy(      T_               a,
              const PETScMatrix<T_>& m);

/** \brief Assign a value to an entry of the matrix
 *  @param [in] i Row index
 *  @param [in] j Column index
 *  @param [in] a Value to assign to <tt>a(i,j)</tt>
 */
    void set(      size_t i,
                   size_t j,
             const T_&    a);

/** \brief Add a value to an entry of the matrix
 *  @param [in] i Row index
 *  @param [in] j Column index
 *  @param [in] a Constant value to add to <tt>a(i,j)</tt>
 */
    void add(      size_t i,
                   size_t j,
             const T_&    a);
/**
 * \brief Assign values to a portion of the matrix
 * @param [in] nr Number of rows to assign
 * @param [in] ir Vector of row indexes to assign
 * @param [in] nc Number of columns to assign
 * @param [in] ic Vector of column indexes to assign
 * @param [in] val Vector of values to assign
 */
    void set(size_t     nr,
             Vect<int>& ir,
             size_t     nc,
             Vect<int>& ic,
             Vect<T_>&  val);

/// \brief Operator <tt>=</tt>
/// \details Assign constant value <tt>a</tt> to matrix diagonal entries
    void operator=(const T_& a);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// \brief Factor matrix
    int Factor();
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/** \brief Solve the linear system of equations.
 *  \details The default parameters are: the Conjugate Gradient method and the Jacobi
 *  method for preconditioner.
 *  To change these values, call function setSolver before this function
 *  @param [in,out] b Vector that contains right-hand side on input and solution on output
 *
 *  @return Number of actual performed iterations 
 */
    int Solve(class PETScVect<T_>& b);

/** \brief Solve the linear system of equations.
 *  \details The default parameters are: the Conjugate Gradient method and the Jacobi
 *  method for preconditioner.
 *  To change these values, call function setSolver before this function
 *  @param [in] b Vector that contains right-hand side
 *  @param [out] x Vector that contains the obtained solution
 *  @return Number of actual performed iterations 
 */
    int Solve(const PETScVect<T_>& b,
                    PETScVect<T_>& x);

/** \brief Choose solver and preconditioner for an iterative procedure
 *  @param [in] solver Option to choose iterative solver among the macros (see PETSc documentation for
 *  more details):
 *  <ul>
 *     <li><tt>KSPRICHARDSON</tt>: The Richardson iterative method (Default damping parameter is
 *                                 <tt>1.0</tt>)
 *     <li><tt>KSPCHEBYSHEV</tt>: The Chebyshev iterative method
 *     <li><tt>KSPCG</tt>: The conjugate gradient method [Default]
 *     <li><tt>KSPCGNE</tt>: The CG method for normal equations (without explicitly forming the product
 *         <tt>A^TA</tt>
 *     <li><tt>KSPGMRES</tt>: The GMRES iterative method (see A Generalized Minimal Residual
 *         Algorithm for Solving Nonsymmetric Linear Systems. Y. Saad and M. H. Schultz, SIAM J. Sci.
 *         Stat. Comput. Vo|. 7, No. 3, July 1986, pp. 856-869)
 *     <li><tt>KSPFGMRES</tt>: The Flexible GMRES method (with restart)
 *     <li><tt>KSPLGMRES</tt>: The 'augmented' standard GMRES method where the subspace uses
 *         approximations to the error from previous restart cycles
 *     <li><tt>KSPTCQMR</tt>: A variant of QMR (quasi minimal residual) developed by Tony Chan
 *     <li><tt>KSPBCGS</tt>: The BiCGStab (Stabilized version of BiConjugate Gradient Squared) method
 *     <li><tt>KSPIBCGS</tt>: The IBiCGStab (Improved Stabilized version of BiConjugate Gradient
 *         Squared) method in an alternative form to have only a single global reduction operation
 *         instead of the usual 3 (or 4)
 *     <li><tt>KSPFBCGS</tt>: The flexible BiCGStab method.
 *     <li><tt>KSPCGS</tt>: The CGS (Conjugate Gradient Squared) method
 *     <li><tt>KSPTFQMR</tt>: A transpose free QMR (quasi minimal residual)
 *     <li><tt>KSPCR</tt>: The conjugate residuals method
 *     <li><tt>KSPLSQR</tt>: The LSQR method
 *     <li><tt>KSPBICG</tt>: The Biconjugate gradient method (similar to running the conjugate
 *         gradient on the normal equations)
 *     <li><tt>KSPMINRES</tt>: The MINRES (Minimum Residual) method
 *     <li><tt>KSPSYMMLQ</tt>: The SYMMLQ method
 *     <li><tt>KSPGCR</tt>: The Generalized Conjugate Residual method
 *  </ul>
 *  Default value is <tt>CG_SOLVER</tt>
 *  @param [in] prec Option to choose preconditioner in an enumerated variable
 *  <ul>
 *     <li><tt>PCJACOBI<tt>: [Default] Jacobi (<i>i.e.</i> diagonal scaling) preconditioning
 *     <li><tt>PCBJACOBI<tt>: Block Jacobi preconditioning, each block is (approximately) solved
 *         with its own KSP object
 *     <li><tt>PCSOR<tt>: (S)SOR (successive over relaxation, Gauss-Seidel) preconditioning
 *     <li><tt>PCEISENSTAT<tt>: An implementation of SSOR (symmetric successive over relaxation,
 *         symmetric Gauss-Seidel) preconditioning that incorporates Eisenstat's trick to reduce
 *         the amount of computation needed
 *     <li><tt>PCICC<tt>: Incomplete Cholesky factorization preconditioners
 *     <li><tt>PCILU<tt>: Incomplete factorization preconditioners
 *     <li><tt>PCASM<tt>: Use the (restricted) additive Schwarz method, each 
 *         block is (approximately) solved with its own KSP object
 *     <li><tt>PCLU<tt>: Uses a direct solver, based on LU factorization, as a preconditioner
 *     <li><tt>PCCHOLESKY<tt>: Uses a direct solver, based on Cholesky factorization, as a preconditioner
 *  </ul>
 *  @param [in] max_it Maximum number of allowed iterations [Default: <tt>1000</tt>]
 *  @param [in] toler Tolerance for convergence [Default: <tt>1.e-8</tt>]
 */
    void setSolver(string solver,
                   string prec,
                   int    max_it=1000,
                   real_t toler=1.e-8);

/// \brief Return C-Array.
/// \details Non zero terms of matrix is stored row by row.
    T_ *get() const;

/** \brief  Return entry <tt>(i,j)</tt> of matrix if this one is stored, <tt>0</tt> otherwise
 *  @param [in] i Row index
 *  @param [in] j Column index
 */
    T_ get(size_t i,
           size_t j) const;

/** \brief Casting operator
 *  \details This casting operator is used only when \b Petsc is used. It allows
 *  converting a PETScMatrix instance into matrix variable <tt>Mat</tt>. 
 */
    operator Mat() const { return _A; }

/// \brief Get 1-norm of matrix
    PetscReal getNorm1() const;

/// \brief Get Frobenius norm of matrix
    PetscReal getFrobeniusNorm() const;

/// \brief Get infinity norm of matrix
    PetscReal getNormMax() const;

/** \brief Assembly of element matrix into global matrix.
 *  @param [in] el Pointer to element instance
 *  @param [in] a Element matrix as a C-array
 */
    void Assembly(const class Element* el,
                              T_*      a);

 private:

   Mesh *_theMesh;
   string _solver, _prec;
   int _type, _max_it;
   bool _extended, _one_dof, _sides;
   int _dof_type, *_nnz, *_col_ind, *_cnnz;
   bool _set_nodes, _set_elements, _set_sides;
   size_t _nb_rows, _nb_cols, _size, _length, _dof;
   real_t _temp, _toler;
   Mat _A;

   void NodeGraph();

   static Mat create(size_t   nr,
                     size_t   nc,
                     MPI_Comm comm=PETSC_COMM_WORLD)
   {
      Mat m;
      MatCreate(comm,&m);
      MatSetSizes(m,PETSC_DECIDE,PETSC_DECIDE,nr,nc);
      MatSetFromOptions(m);
      MatSetUp(m);
      return m;
    }

    int col_index(size_t i, size_t j) const
    {
       for (int ii=0; ii<_nnz[i-1]; ii++) {
          size_t k = _col_ind[_cnnz[i-1]+ii];
          if (j==k+1)
             return k;
       }
       return -1;
    }
};


///////////////////////////////////////////////////////////////////////////////
//                         I M P L E M E N T A T I O N                       //
///////////////////////////////////////////////////////////////////////////////

#ifndef DOXYGEN_SHOULD_SKIP_THIS

template<class T_>
PETScMatrix<T_>::PETScMatrix()
{
   _solver = KSPCG;
   _prec = PCJACOBI;
   _max_it = 1000;
   _toler = 1.e-8;
   _nb_rows = _nb_cols = 0;
   _nnz = _col_ind = _cnnz = NULL;
}


template<class T_>
PETScMatrix<T_>::PETScMatrix(size_t nr,
                             size_t nc)
{
   _nb_rows = nr;
   _nb_cols = nc;
   _size = 0;
   if (_nb_rows == _nb_cols)
      _size = _nb_rows;
   _length = _nb_rows * _nb_cols;
   _A = create(_nb_rows,_nb_cols);
   _solver = KSPCG;
   _prec = PCJACOBI;
   _max_it = 1000;
   _toler = 1.e-8;
   _nnz = _col_ind = _cnnz = NULL;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class PETScMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
PETScMatrix<T_>::PETScMatrix(size_t size)
{
   _size = _nb_rows = _nb_cols = size;
   _length = _size * _size;
   _A = create(_nb_rows,_nb_cols);
   _solver = KSPCG;
   _prec = PCJACOBI;
   _max_it = 1000;
   _toler = 1.e-8;
   _nnz = _col_ind = _cnnz = NULL;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class PETScMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
PETScMatrix<T_>::PETScMatrix(class Mesh&  mesh,
                                   size_t dof)
{
   setMesh(mesh,dof);
   //   MatAssemblyBegin(_A,MAT_FINAL_ASSEMBLY);
   //   MatAssemblyEnd(_A,MAT_FINAL_ASSEMBLY);
   _solver = KSPCG;
   _prec = PCJACOBI;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class PETScMatrix is constructed in\n";
   std::clog << "File: " << __FILE__<< ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
PETScMatrix<T_>::PETScMatrix(      size_t dof,
                             class Mesh&  mesh,
                                   int    code)
{
   setMesh(dof,mesh,mesh.getDOFSupport());
   _solver = KSPCG;
   _prec = PCJACOBI;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class PETScMatrix is constructed in\n";
   std::clog << "File: " << __FILE__<< ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
PETScMatrix<T_>::PETScMatrix(      size_t dof,
                                   size_t nb_eq,
                             class Mesh&  mesh)
{
   setMesh(dof,nb_eq,mesh);
   _solver = KSPCG;
   _prec = PCJACOBI;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class PETScMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
PETScMatrix<T_>::PETScMatrix(const class vector<std::pair<size_t,size_t> >& I,
                                         int                                opt)
{
   setGraph(I,opt);
   _solver = KSPCG;
   _prec = PCJACOBI;
   _max_it = 1000;
   _toler = 1.e-8;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class PETScMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
PETScMatrix<T_>::PETScMatrix(const PETScMatrix& m)
{
   _dof = m._dof;
   _size = m._size;
   _length = m._length;
   _nnz = new PetscInt [_size+1];
   _cnnz = new PetscInt [_size+1];
   _col_ind = new PetscInt [_length];
   _col_ind = m._col_ind;
   _nnz = m._nnz;
   _cnnz = m._cnnz;
   _solver = KSPCG;
   _prec = PCJACOBI;
   _max_it = 1000;
   _toler = 1.e-8;
   _solver = m._solver;
   _prec = m._prec;
   MatCopy(m._A,_A,SAME_NONZERO_PATTERN);
   //   MatAssemblyBegin(_A,MAT_FINAL_ASSEMBLY);
   //   MatAssemblyEnd(_A,MAT_FINAL_ASSEMBLY);
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class PETScMatrix is constructed in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
PETScMatrix<T_>::~PETScMatrix()
{
   if (_nnz)
      delete [] _nnz;
   if (_cnnz)
      delete [] _cnnz;
   if (_col_ind)
      delete [] _col_ind;
   MatDestroy(&_A);
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class PETScMatrix is deleted in\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
void PETScMatrix<T_>::setSize(size_t size)
{
   _nb_rows = _nb_cols = _size = size;
   _length = lsize_t(_nb_rows*_nb_cols);
}


template<class T_>
void PETScMatrix<T_>::setSize(size_t nr,
                              size_t nc)
{
   _nb_rows = nr;
   _nb_cols = nc;
   _size = 0;
   if (_nb_rows==_nb_cols)
      _size = _nb_rows;
   _length = lsize_t(_nb_rows*_nb_cols);
   static int n=0;
   MatCreateSeqAIJ(PETSC_COMM_SELF,_nb_rows,_nb_cols,0,&n,&_A);
   MatSetFromOptions(_A);
}


template<class T_>
void PETScMatrix<T_>::setMesh(class Mesh&  mesh,
                                    size_t dof)
{
   _dof_type = mesh.getDOFSupport();
   _theMesh = &mesh;
   _dof = dof;
   _size = _nb_rows = _nb_cols = _theMesh->getNbEq();
   if (_dof_type==NODE_DOF) {
      if (_dof)
         _size = _nb_rows = _nb_cols = _theMesh->getNbNodes();
      else
         _size = _nb_rows = _nb_cols = _theMesh->getNbEq();
   }
   else if (_dof_type==SIDE_DOF) {
      if (_dof)
         _size = _nb_rows = _nb_cols = _theMesh->getNbSides();
      else
         _size = _nb_rows = _nb_cols = _theMesh->getNbEq();
   }
   else if (_dof_type==ELEMENT_DOF)
      _size = _nb_rows = _nb_cols = _theMesh->getNbElements();
   NodeGraph();
   MatCreateSeqAIJ(PETSC_COMM_SELF,_nb_rows,_nb_cols,PETSC_DEFAULT,_nnz,&_A);
   MatSeqAIJSetColumnIndices(_A,_col_ind);
   MatSetFromOptions(_A);
}


template<class T_>
void PETScMatrix<T_>::NodeGraph()
{
   size_t l=0, size = _theMesh->getNbEq();
   vector<size_t> ch(size);
   NodeSkyline(*_theMesh,ch);

   size_t ip = 0;
   mesh_elements(*_theMesh) {
      for (size_t in=1; in<=The_element.getNbNodes(); in++) {
         Node *nd1 = The_element(in);
         for (size_t k=1; k<=nd1->getNbDOF(); k++) {
            if (nd1->getDOF(k)!=0) {
               for (size_t jn=1; jn<=The_element.getNbNodes(); jn++) {
                  Node *nd2 = The_element(jn);
                  for (l=1; l<=nd2->getNbDOF(); l++) {
                     if (nd2->getDOF(l)!=0) {
                        ip++;
                        if (nd1->getDOF(k)!=nd2->getDOF(l))
                           ip++;
                     }
                  }
               }
            }
         }
      }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_elements(*_theMesh) {
      for (size_t in=1; in<=The_element.getNbNodes(); in++) {
         Node *nd1 = The_element(in);
         for (size_t k=1; k<=nd1->getNbDOF(); k++) {
            if (nd1->getDOF(k)!=0) {
               for (size_t jn=1; jn<=The_element.getNbNodes(); jn++) {
                  Node *nd2 = The_element(jn);
                  for (l=1; l<=nd2->getNbDOF(); l++) {
                     if (nd2->getDOF(l)!=0) {
                        *(it++) = RC(nd1->getDOF(k)-1,nd2->getDOF(l)-1);
                        if (nd1->getDOF(k)!=nd2->getDOF(l))
                           *(it++) = RC(nd2->getDOF(l)-1,nd1->getDOF(k)-1);
                     }
                  }
               }
            }
         }
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());

   _length = pp.size();
   _col_ind = new int [_length];
   _nnz = new int [size+1];
   for (size_t i=0; i<size; i++)
      _nnz[i] = 0;
   for (size_t j=0; j<_length; j++) {
      _nnz[pp[j].first]++;
      _col_ind[j] = pp[j].second;
   }
   _cnnz = new int [size+1];
   _cnnz[0] = 0;
   for (size_t i=1; i<=size; i++)
      _cnnz[i] = _cnnz[i-1] + _nnz[i-1];
}


template<class T_>
void PETScMatrix<T_>::setGraph(const vector<std::pair<size_t,size_t> >& I,
                                     int                                opt)
{
   size_t i;
   _length = I.size();
   _nb_rows = _nb_cols = 0;
   vector<RC> pp(I.size());
   for (i=0; i<_length; i++) {
      _nb_rows = max(_nb_rows,I[i].first);
      _nb_cols = max(_nb_cols,I[i].second);
      pp[i] = RC(I[i].first-1,I[i].second-1);
   }
   _size = _nb_rows;
   if (opt==0) {
      sort(pp.begin(),pp.end());
      vector<RC>::iterator new_end = std::unique(pp.begin(),pp.end());
      pp.erase(new_end,pp.end());
   }
   _length = pp.size();
   _col_ind = new PetscInt [_length];
   _nnz = new PetscInt [_nb_rows];
   for (size_t i=0; i<_size; i++)
      _nnz[i] = 0;
   for (size_t j=0; j<_length; j++)
      _nnz[pp[j].first]++;
   for (size_t k=1; k<_size; k++)
      _nnz[k] += _nnz[k-1];
   for (int ii=_size; ii>0; ii--)
      _nnz[ii] = _nnz[ii-1] + 1;
   MatCreateSeqAIJ(PETSC_COMM_SELF,_nb_rows,_nb_cols,PETSC_DEFAULT,_nnz,&_A);
   MatSeqAIJSetColumnIndices(_A,_col_ind);
   MatSetFromOptions(_A);
}


template<class T_>
void PETScMatrix<T_>::setOneDOF()
{
   _one_dof = true;
}


template<class T_>
void PETScMatrix<T_>::setSides()
{
   _sides = true; 
}


template<class T_>
void PETScMatrix<T_>::DiagPrescribe(      class Mesh&          mesh,
                                                PETScVect<T_>& b,
                                    const       PETScVect<T_>& u)
{
   real_t p = 0;
   for (size_t j=1; j<=_nb_rows; j++)
      p = max(p,Abs(get(j,j)));
 /*  const Node *nd;
         size_t k=0;
      for (mesh.topNode(); (nd=mesh.getNode());) {
      for (size_t i=1; i<=nd->getNbDOF(); ++i) {
         size_t ii = nd->getDOF(i)-1;
         for (size_t j=0; j<_row_ptr[ii+1]-_row_ptr[ii]; j++,k++)
            if (nd->getCode(i)>0) {
               b[ii] = p*u[ii];
               _a[k] = 0;
               if (ii+1==_col_ind[k])
                  _a[k] = p;
            }
      }
      }*/
}


template<class T_>
void PETScMatrix<T_>::DiagPrescribe(      PETScVect<T_>& b,
                                    const PETScVect<T_>& u)
{
   real_t p = 0;
   for (size_t j=1; j<=_nb_rows; j++)
      p = max(p,Abs(get(j,j)));
 /* size_t k=0;
   MESH_ND {
      for (size_t i=1; i<=TheNode.getNbDOF(); ++i) {
         size_t ii = TheNode.getDOF(i)-1;
         for (size_t j=0; j<_row_ptr[ii+1]-_row_ptr[ii]; j++,k++)
            if (TheNode.getCode(i)>0) {
               b[ii] = p*u[ii];
               _a[k] = 0;
               if (ii+1==_col_ind[k])
                  _a[k] = p;
            }
      }
      }*/
}


template<class T_>
T_ PETScMatrix<T_>::operator()(size_t i,
                               size_t j) const
{
   PetscInt ii=i-1, jj=col_index(i,j);
   PetscScalar a;
   MatGetValues(_A,1,&ii,1,&jj,&a);
   return a;
}


template<class T_>
void PETScMatrix<T_>::getMesh(class Mesh& mesh)
{
  /*   if (_sides)
      SideGraph(mesh,_row_ptr,_col_ind);
   else {
      if (_dof)
         _size = _nb_rows = _nb_cols = mesh.getNbNodes();
      else
         _size = _nb_rows = _nb_cols = mesh.getNbEq();
      if (_type)
         XGraph(mesh,_row_ptr,_col_ind);
//             XGraphScal(mesh,_row_ptr,_col_ind);
else*/
   NodeGraph(mesh,_nnz,_col_ind);
         //   }
   //   _a.resize(_length,T_(0.));
}


template<class T_>
void PETScMatrix<T_>::Mult(const PETScVect<T_>& x,
                                 PETScVect<T_>& y) const
{
   MatMult(_A,x,y);
   VecAssemblyBegin(y);
   VecAssemblyEnd(y);
}


template<class T_>
void PETScMatrix<T_>::MultAdd(const PETScVect<T_>& x,
                                    PETScVect<T_>& y) const
{
   PETScVect<T_> z(y);
   MatMult(_A,x,y);
   y += z;
   VecAssemblyBegin(y);
   VecAssemblyEnd(y);
}


template<class T_>
void PETScMatrix<T_>::MultAdd(            T_             a,
                              const class PETScVect<T_>& x,
                                    class PETScVect<T_>& y) const
{
   PETScVect<T_> z(y);
   MatMult(_A,x,y);
   y += a*z;
   VecAssemblyBegin(y);
   VecAssemblyEnd(y);
}


template<class T_>
void PETScMatrix<T_>::TMult(const PETScVect<T_>& x,
                                  PETScVect<T_>& y) const
{
  /*   for (size_t i=0; i<_nb_rows; i++)
      for (size_t j=0; j<_row_ptr[i+1]-_row_ptr[i]; j++)
      y[_col_ind[j+_row_ptr[i]-1]-1] += _a[_row_ptr[i]+j-1] * x[i];*/
}


template<class T_>
void PETScMatrix<T_>::set(      size_t i,
                                size_t j,
                          const T_&    a)
{
   MatSetValue(_A,i-1,col_index(i,j),a,INSERT_VALUES);
}


template<class T_>
void PETScMatrix<T_>::add(      size_t i,
                                size_t j,
                          const T_&    a)
{
   MatSetValue(_A,i-1,col_index(i,j),a,ADD_VALUES);
}


template<class T_>
void PETScMatrix<T_>::set(size_t     nr,
                          Vect<int>& ir,
                          size_t     nc,
                          Vect<int>& ic,
                          Vect<T_>&  val)
{
   MatSetValues(_A,nr,&ir[0],nc,&ic[0],&val[0],INSERT_VALUES);
   MatAssemblyBegin(_A,MAT_FINAL_ASSEMBLY);
   MatAssemblyEnd(_A,MAT_FINAL_ASSEMBLY);
}


template<class T_>
void PETScMatrix<T_>::Axpy(      T_               a,
                           const PETScMatrix<T_>& m)
{
  //   _a += a * m._a;
}


template<class T_>
void PETScMatrix<T_>::operator=(const T_& a)
{
   for (size_t i=1; i<=_nb_rows; i++)
      MatSetValue(_A,i-1,col_index(i,i),a,INSERT_VALUES);
   MatAssemblyBegin(_A,MAT_FINAL_ASSEMBLY);
   MatAssemblyEnd(_A,MAT_FINAL_ASSEMBLY);
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
int PETScMatrix<T_>::Factor()
{
   return -1;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
int PETScMatrix<T_>::Solve(class PETScVect<T_>& b)
{
   PETScVect<T_> x(b.size());
   int nb_it = Solve(b,x);
   b = x;
   VecAssemblyBegin(b);
   VecAssemblyEnd(b);
   return nb_it;
}


template<class T_>
int PETScMatrix<T_>::Solve(const PETScVect<T_>& b,
                                 PETScVect<T_>& x)
{
   return 0;
}


template<class T_>
void PETScMatrix<T_>::setSolver(string solver,
                                string prec,
                                int    max_it,
                                real_t toler)
{
   _solver = solver;
   _prec = prec;
   _max_it = max_it;
   _toler = toler;
}


#endif /* DOXYGEN_SHOULD_SKIP_THIS */


template<class T_>
PetscReal PETScMatrix<T_>::getNorm1() const
{
   PetscReal n;
   MatNorm(_A,NORM_1,&n);
   return n;
}


template<class T_>
PetscReal PETScMatrix<T_>::getFrobeniusNorm() const
{
   PetscReal n;
   MatNorm(_A,NORM_FROBENIUS,&n);
   return n;
}


template<class T_>
PetscReal PETScMatrix<T_>::getNormMax() const
{
   PetscReal n;
   MatNorm(_A,NORM_INFINITY,&n);
   return n;
}


template<class T_>
void PETScMatrix<T_>::Assembly(const class Element* el,
                                           T_*      a)
{
   size_t kk=0;
   for (size_t i=1; i<=el->getNbNodes(); ++i) {
      Node *nd1 = (*el)(i);
      for (size_t k=1; k<=nd1->getNbDOF(); ++k) {
         for (size_t j=1; j<=el->getNbNodes(); ++j) {
            Node *nd2 = (*el)(j);
            for (size_t l=1; l<=nd2->getNbDOF(); ++l) {
               if (nd1->getDOF(k)!=0 && nd2->getDOF(l)!=0)
                  add(nd1->getDOF(k),nd2->getDOF(l),a[kk]);
               kk++;
            }
         }
      }
   }
}


///////////////////////////////////////////////////////////////////////////////
//                           ASSOCIATED  FUNCTIONS                           //
///////////////////////////////////////////////////////////////////////////////


/** \fn PETScVect<T_> operator*(const PETScMatrix<T_>& A, const PETScVect<T_>& x)
 *  \ingroup VectMat
 *  Multiply a matrix by a vector
 *  @param [in] A %Matrix to multiply by (Instance of class PETScMatrix)
 *  @param [in] x Vector to multiply by (Instance of class PETScVect)
 *  @return Vector product <tt>y = Ax</tt>
 */
template <class T_>
PETScVect<T_> operator*(const PETScMatrix<T_>& A,
                        const PETScVect<T_>&   x)
{
   VecAssemblyBegin(x);
   VecAssemblyEnd(x);
   PETScVect<T_> y(x.size());
   MatMult(A,x,y);
   return y;
}


/** \fn ostream & operator<<(ostream& s, const PETScMatrix<T_> &a)
 *  \ingroup VectMat
 *  \brief Output matrix in output stream
 */
template<class T_>
ostream& operator<<(      ostream&         s,
                    const PETScMatrix<T_>& A)
{
   MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
   MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);
   PetscViewer viewer;
   PetscViewerASCIIOpen(PETSC_COMM_WORLD,PETSC_NULL,&viewer);
   if (A.getNbRows() > 500)
      PetscViewerPushFormat(viewer,PETSC_VIEWER_ASCII_INFO);
   MatView(A,viewer);
   PetscViewerDestroy(&viewer);
   return s;
}

} /* namespace OFELI */

#endif

#endif
