/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzaniset

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                    Definition of Template class PETScWrapper

  ==============================================================================*/

#ifndef __PETSC_WRAPPER_H
#define __PETSC_WRAPPER_H

#ifdef USE_PETSC

#include "linear_algebra/petsc/PETScVect.h"
#include "linear_algebra/petsc/PETScMatrix.h"
#include "petscksp.h"

namespace OFELI {

template<class T_>
class PETScWrapper
{

 public:

/**
 * Constructor with program arguments
 * @param [in] argc
 * @param [in] args
 * @param [in] help String that contains message to display when argument
 *             <tt>-v</tt> is used
 */
    PETScWrapper(int    argc,
                 char** args,
                 string help="") 
    {
       PetscInt n;
       PetscInitialize(&argc,&args,(char *)0,help.c_str());
       MPI_Comm_size(PETSC_COMM_WORLD,&_size);
       MPI_Comm_rank(PETSC_COMM_WORLD,&_rank);
       PetscOptionsGetInt(PETSC_NULL,"-n",&n,PETSC_NULL);
       _ls = _nls = false;
       _tol = 1.e-14;
    }

/// \brief Destructor
/// \details Destroy the KSP context and release memory allocated by petsc
    ~PETScWrapper() 
    {
       if (_ls)
          KSPDestroy(&_ksp);
       _err = PetscFinalize();
    }

/// \brief Return wrapper size, i.e. number of processors
    PetscMPIInt size() const { return _size; }

/** \brief Define problem matrix
 *  @param [in] A PETScMatrix instance that contains matrix
 */
    void setLS(PETScMatrix<T_>& A)
    {
       _ls = true;
       _A = &A;
       KSPCreate(PETSC_COMM_WORLD,&_ksp);
       KSPSetOperators(_ksp,*_A,*_A,DIFFERENT_NONZERO_PATTERN);

/**    Set linear solver defaults for this problem (optional).
      - By extracting the KSP and PC contexts from the KSP context,
        we can then directly call any KSP and PC routines to set
        various options.
      - The following four statements are optional; all of these
        parameters could alternatively be specified at runtime via
        KSPSetFromOptions();
*/
       KSPGetPC(_ksp,&_pc);
       PCSetType(_pc,PCJACOBI);
       KSPSetTolerances(_ksp,1.e-5,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);
       KSPSetInitialGuessNonzero(_ksp,PETSC_TRUE);
    }

/** \brief Choose preconditioner for the iterative procedure
 *  @param [in] p Option to choose preconditioner among the macros (see PETSc documentation for
 *  more details):
 *  <ul>
 *     <li><tt>PCJACOBI<tt>: [Default] Jacobi (<i>i.e.</i> diagonal scaling) preconditioning
 *     <li><tt>PCBJACOBI<tt>: Block Jacobi preconditioning, each block is (approximately) solved
 *         with its own KSP object
 *     <li><tt>PCSOR<tt>: (S)SOR (successive over relaxation, Gauss-Seidel) preconditioning
 *     <li><tt>PCEISENSTAT<tt>: An implementation of SSOR (symmetric successive over relaxation,
 *         symmetric Gauss-Seidel) preconditioning that incorporates Eisenstat's trick to reduce
 *         the amount of computation needed
 *     <li><tt>PCICC<tt>: Incomplete Cholesky factorization preconditioners
 *     <li><tt>PCILU<tt>: Incomplete factorization preconditioners
 *     <li><tt>PCASM<tt>: Use the (restricted) additive Schwarz method, each 
 *         block is (approximately) solved with its own KSP object
 *     <li><tt>PCLU<tt>: Uses a direct solver, based on LU factorization, as a preconditioner
 *     <li><tt>PCCHOLESKY<tt>: Uses a direct solver, based on Cholesky factorization, as a preconditioner
 *  </ul>
 */
    void setPreconditioner(string p)
    {
       PCSetType(_pc,p.c_str());
    }

/** \brief Choose iteration parameters
 *  @param [in] max_it Maximum number of linear solver iterations
 *  @param [in] tol Tolerance for convergence of iteration process
 *              [Default: <tt>1.e-12</tt>]
 */
    void setIterationParameters(size_t max_it,
                                double tol)
    {
       _tol = tol;
       KSPSetTolerances(_ksp,_tol,PETSC_DEFAULT,PETSC_DEFAULT,max_it);
    }

/** \brief Choose the iterative method
 *  @param [in] s Option to choose iterative solver among the macros (see PETSc documentation for
 *  more details):
 *  <ul>
 *     <li><tt>KSPRICHARDSON</tt>: The Richardson iterative method (Default damping parameter is
 *                                 <tt>1.0</tt>)
 *     <li><tt>KSPCHEBYSHEV</tt>: The Chebyshev iterative method
 *     <li><tt>KSPCG</tt>: The conjugate gradient method [Default]
 *     <li><tt>KSPCGNE</tt>: The CG method for normal equations (without explicitly forming the product
 *         <tt>A^TA</tt>
 *     <li><tt>KSPGMRES</tt>: [Default] The GMRES iterative method (see A Generalized Minimal Residual
 *         Algorithm for Solving Nonsymmetric Linear Systems. Y. Saad and M. H. Schultz, SIAM J. Sci.
 *         Stat. Comput. Vo|. 7, No. 3, July 1986, pp. 856-869)
 *     <li><tt>KSPFGMRES</tt>: The Flexible GMRES method (with restart)
 *     <li><tt>KSPLGMRES</tt>: The 'augmented' standard GMRES method where the subspace uses
 *         approximations to the error from previous restart cycles
 *     <li><tt>KSPTCQMR</tt>: A variant of QMR (quasi minimal residual) developed by Tony Chan
 *     <li><tt>KSPBCGS</tt>: The BiCGStab (Stabilized version of BiConjugate Gradient Squared) method
 *     <li><tt>KSPIBCGS</tt>: The IBiCGStab (Improved Stabilized version of BiConjugate Gradient
 *         Squared) method in an alternative form to have only a single global reduction operation
 *         instead of the usual 3 (or 4)
 *     <li><tt>KSPFBCGS</tt>: The flexible BiCGStab method.
 *     <li><tt>KSPCGS</tt>: The CGS (Conjugate Gradient Squared) method
 *     <li><tt>KSPTFQMR</tt>: A transpose free QMR (quasi minimal residual)
 *     <li><tt>KSPCR</tt>: The conjugate residuals method
 *     <li><tt>KSPLSQR</tt>: The LSQR method
 *     <li><tt>KSPBICG</tt>: The Biconjugate gradient method (similar to running the conjugate
 *         gradient on the normal equations)
 *     <li><tt>KSPMINRES</tt>: The MINRES (Minimum Residual) method
 *     <li><tt>KSPSYMMLQ</tt>: The SYMMLQ method
 *     <li><tt>KSPGCR</tt>: The Generalized Conjugate Residual method
 *  </ul>
 */
    void setIterationMethod(string method)
    {
       KSPSetType(_ksp,method.c_str());
    }

/** \brief Solve the linear system
 *  \details If the member functions setIterationMethod and setPreconditioner have not been used,
 *  default methods are used
 *  @param [in] b Vector containing the right-hand side
 *  @param [in,out] x Vector containing the initial guess on input and, if convergence is achieved,
 *  the solution on output
 */
    void solveLS(const PETScVect<T_>& b,
                       PETScVect<T_>& x)
    {
       MatAssemblyBegin(*_A,MAT_FINAL_ASSEMBLY);
       MatAssemblyEnd(*_A,MAT_FINAL_ASSEMBLY);
       _x = &x;
       KSPSolve(_ksp,b,*_x);
    }

/**
 * \brief Check residual error
 * \details This function computes the residual A*x - b and outputs the
 * number of iterations
 * @param [out] u Residual vector
 */
    void checkError(PETScVect<T_>& u)
    {
       PetscReal norm;
       PetscInt  it;
       VecAXPY(*_x,-1.0,u);
       VecNorm(*_x,NORM_2,&norm);
       KSPGetIterationNumber(_ksp,&it);
       cout << "Number of iterations: " << it << endl;
       cout << "Norm of error: " << norm << endl;
    }

/// \brief Return the number of iterations
    int getIterationNumber() const {
       int it;
       KSPGetIterationNumber(_ksp,&it);
       return it;
    }

/// \brief Return the rank of the current processor
    PetscMPIInt getRank() const { return _rank; }

/// \brief Output wrapper information
    template<class S_>
    friend ostream &operator<<(      ostream&          s,
                               const PETScWrapper<S_>& w);

 private:
   bool             _ls, _nls;
   PetscErrorCode   _err;
   PetscInt         _size;
   PetscMPIInt      _rank;
   KSP              _ksp;
   PC               _pc;
   real_t           _tol;
   PETScVect<T_>*   _x;
   PETScMatrix<T_>* _A;

};


/** \fn ostream &operator<<(ostream &s, const PETScWrapper<T_> &w)
 *  \brief Output Petsc Wrapper data in output stream
 *  \ingroup VectMat
 */
template<class T_>
ostream &operator<<(      ostream&          s, 
                    const PETScWrapper<T_>& w)
{
   if (w._ls) {
      s << "Linear Solver data:" << endl;
      KSPView(w._ksp,PETSC_VIEWER_STDOUT_SELF);
   }
   return s;
}

} /* namespace OFELI */

/*! @} */

#endif

#endif

/*

Available preconditioners:
PCJACOBI, PCBJACOBI, PCSOR, PCEISENSTAT, PCICC, PCILU, PCASM, PCKSP, PCLU, PCCHOLESKY

Available iterative solvers:
KSPRICHARDSON, KSPCHEBYSHEV, KSPCG, KSPGROPPCG, KSPPIPECG, KSPCGNE, KSPNASH, KSPSTCG
KSPGLTR, KSPGMRES, KSPFGMRES, KSPLGMRES, KSPDGMRES, KSPPGMRES, KSPTCQMR, KSPBCGS
KSPIBCGS, KSPFBCGS, KSPFBCGSR, KSPBCGSL, KSPCGS, KSPTFQMR, KSPCR, KSPPIPECR, KSPLSQR
KSPPREONLY, KSPQCG, KSPBICG, KSPMINRES, KSPSYMMLQ, KSPLCD, KSPPYTHON, KSPGCR, KSPSPECEST

*/
