/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                         Definition of class Partition

  ==============================================================================*/

#ifndef __PARTITION_H
#define __PARTITION_H


#include "OFELI_Config.h"
#include "mesh/Mesh.h"
#include "io/output.h"

#define WITH_METIS

#ifdef WITH_METIS
#ifdef __cplusplus
extern "C"
{
#include "mesh/metis/metis.h"
}
#endif
#endif


namespace OFELI {

/*! \class Partition
 *  \ingroup Mesh
 *  \brief To partition a finite element mesh into balanced submeshes
 *
 * Class Partition enables partitioning a given mesh into a given number
 * of submeshes with a minimal connectivity. Partition uses the well known
 * <tt>metis</tt> library that is included in the OFELI library. A more detailed
 * description of metis can be found in the web site:\n
 * http://www.csit.fsu.edu/~burkardt/c_src/metis/metis.html
*/

class Partition
{

#ifndef DOXYGEN_SHOULD_SKIP_THIS
 typedef struct { size_t side1, sd, side; } Interface;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

 public:

//----------------------------   BASIC OPERATIONS   -----------------------------

/// Constructor to get interface information. Partition information
/// is read in <tt>file</tt>
    Partition(Mesh& mesh,
              char* file);

/// Constructor to partition a mesh into n submeshes
    Partition(Mesh&  mesh,
              size_t n);

/// Destructor
    ~Partition();

//-----------------------------   INSPECTORS  -----------------------------------

/// Return number of submeshes
    size_t getNbSubMeshes() const
       { return _nb_submesh; }

/// Return number of nodes in given submesh
    size_t getNbNodes(size_t i) const
       { return _theSubMesh[i]->getNbNodes(); }

/// Return number of elements in given submesh
    size_t getNbElements(size_t i) const
       { return _theSubMesh[i]->getNbElements(); }

/// Return the submesh of label i
    Mesh &getMesh(size_t i) const
       { return *_theSubMesh[i]; }

/// Return node label in subdomain by giving its label in initial mesh
    size_t getNodeLabelInSubMesh(size_t sm,
                                 size_t label) const
       { return _m2sm_node[sm][label-1]; }

/// Return element label in subdomain by giving its label in initial mesh
    size_t getElementLabelInSubMesh(size_t sm,
                                    size_t label) const
       { return _m2sm_element[sm][label-1]; }

/// Return node label in initial mesh by giving its label in submesh
    size_t getNodeLabelInMesh(size_t sm,
                              size_t label) const
       { return _sm2m_node[sm][label-1]; }

/// Return element label in initial mesh by giving its label in submesh
    size_t getElementLabelInMesh(size_t sm,
                                 size_t label) const
       { return _sm2m_element[sm][label-1]; }

/// Return Nb. of interface sides for a given sub-mesh
    size_t getNbInterfaceSides(size_t sm) const
       { return _nb_interface_sides[sm]; }

/// Return i-th side label in sub-mesh <tt>sm</tt>
    size_t getFirstSideLabel(size_t sm,
                             size_t i) const
       { return _interface_side[sm][i-1].side1; }

/// Return submesh that contains the <tt>i</tt>-th side label in sub-mesh <tt>sm</tt>
    size_t getSubMesh(size_t sm,
                      size_t i) const
       { return _interface_side[sm][i-1].sd; }

/// Return side label in the neighbouring submesh corresponding to <tt>i</tt>-th side
/// label in sub-mesh <tt>sm</tt>
    size_t getSecondSideLabel(size_t sm,
                              size_t i) const
       { return _interface_side[sm][i-1].side; }

/// Set Message Level
    void setVerbose(size_t verb)
       { _verbose = verb; }

/// Save <tt>n</tt>-th submesh in file
    void Put(size_t n,
             char*  file) const
       { _theSubMesh[n]->Put(file); }

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// Save <tt>n</tt>-th submesh in binary file
    void BPut(size_t n,
              char* file) const
       { _theSubMesh[n]->BPut(file); }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

//-----------------------------   MODIFIERS  -----------------------------------

/// Get interface information
    void Init(int* epart);

//------------------------   ASSOCIATED OPERATORS  -----------------------------

/// Output class information
    friend ostream & operator<<(      ostream&   s,
                                const Partition& p);

 protected:
#ifndef DOXYGEN_SHOULD_SKIP_THIS
    int  _nb_submesh, _verbose;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

 private:
   Mesh           *_theMesh, *_theSubMesh[MAX_NB_SUBMESHES];
   vector<size_t> _nb_interface_sides;
   size_t         **_sm2m_node, **_m2sm_node, **_sm2m_element, **_m2sm_element;
   Interface      **_interface_side;
   bool EqualSides(size_t sm1, const Side *sd1, size_t sm2, const Side *sd2);
};

} /* namespace OFELI */

#endif
