/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                 Template function for Conjugate Gradient Method
                   CG is inspired from the SIAM Templates book

  ==============================================================================*/

#ifndef __CG_H
#define __CG_H

#include <iostream>
using std::ostream;
using std::cout;
using std::endl;

#include <iomanip>
using std::setw;
using std::ios;
using std::setprecision;


#include "OFELI_Config.h"
#include "linear_algebra/Vect.h"
#include "solvers/Precond.h"
#include "util/util.h"
#include "io/output.h"

namespace OFELI {

/*! \file CG.h
 *  \brief Functions to solve a symmetric positive definite linear system of equations
 *  using the Conjugate Gradient method.
 *
 *  Preconditioning is possible using a preconditioning class.
 *
 */

/** \fn int CG(const M_ &A, const P_ &P, const Vect<T_> &b, Vect<T_> &x, int max_it, double &toler, int verbose)
 *  \ingroup Solver
 *  \brief Conjugate gradient solver function.
 *  @param [in] A Problem matrix (Instance of abstract class \b M_).
 *  @param [in] P Preconditioner (Instance of abstract class \b P_).
 *  @param [in] b Right-hand side vector (class Vect)
 *  @param [in,out] x Vect instance containing initial solution guess in input and solution
 *  of the linear system in output (If iterations have succeeded).
 *  @param [in] max_it Maximum number of iterations.
 *  @param [in] toler Tolerance for convergence (measured in relative weighted 2-Norm) in input,
 *  effective discrepancy in output.
 *  @param [in] verbose Information output parameter
 *    - 0: No output
 *    - 1: Output iteration information,
 *    - 2 and greater: Output iteration information and solution at each iteration.
 *  @return Number of performed iterations,
 *
 * \tparam <T_> Data type (double, float, complex<double>, ...)
 * \tparam <M_> %Matrix storage class
 * \tparam <P_> Preconditioner class
 */
template<class T_, class M_, class P_ >
int CG(const M_&       A,
       const P_&       P,
       const Vect<T_>& b,
             Vect<T_>& x,
             int       max_it,
             double&   toler,
             int       verbose)
{
   int it;
   size_t size = x.size();
   double res, nrm=b.getNorm2();
   T_ rho=0, rho_1=T_(1), beta=0;
   if (nrm == 0)
      nrm = 1;
   Vect<T_> r(size), p(size), q(size), z(size);
   A.Mult(x,r);
   r = b - r;
   if ((res = r.getNorm2()/nrm) <= toler) {
      if (verbose > 1)
         cout << "Convergence after 0 iterations." << endl;
      return 0;
   }

   for (it=1; it<=max_it; it++) {
       P.Solve(r,z);
       rho = Dot(r,z);
       if (it==1)
          p = z;
       else {
          beta = rho/rho_1;
          p = z + beta * p;
       }
       A.Mult(p,q);
       T_ alpha = rho/Dot(p,q);
       x += alpha * p;
       r -= alpha * q;
       res = r.getNorm2()/nrm;

       if (verbose > 1) {
          cout << "Iteration: " << setw(4) << it;
          cout << "  ... Residual: " << res << endl;
       }
       if (verbose > 2)
          cout << x;
       if (res <= toler) {
          toler = res;
          if (verbose)
             cout << "Convergence after " << it << " iterations." << endl;
          return it;
       }
       rho_1 = rho;
   }
   if (verbose)
      cout << "No Convergence after " << it << " iterations." << endl;
   return -it;
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
int CG(      Matrix<T_> *A,
             Prec<T_> *P,
       const Vect<T_> &b,
             Vect<T_> &x,
             int max_it,
             double toler,
             int verbose)
{
   int it=0;
   size_t size = x.size();
   double nrm=b.getNorm2(), rho=0, res=0, beta=0, rho_1=0;

   if (nrm == 0)
      nrm = 1;
   Vect<T_> r(size), z(size), p(size), q(size);
   A->Mult(x,r);
   r = b - r;
   if ((res = r.getNorm2()/nrm) <= toler) {
      if (verbose > 1)
         cout << "Convergence of the CG method after " << it << " iterations." << endl;
      return it;
   }

   for (it=1; it<=max_it; it++) {
      P->Solve(r,z);
      rho = Dot(r,z);
      if (it==1)
         p = z;
      else {
         beta = rho/rho_1;
         p = z + beta * p;
      }
      A->Mult(p,q);
      double alpha = rho/Dot(p,q);
      x += alpha * p;
      r -= alpha * q;
      res = r.getNorm2()/nrm;

      if (verbose > 1) {
         cout << "Iteration: " << setw(4) << it;
         cout << "  ... Residual: " << res << endl;
      }
      if (verbose > 2)
         cout << x;
      if (res <= toler) {
         if (verbose)
            cout << "Convergence of the CG method after " << it << " iterations." << endl;
         return it;
      }
      rho_1 = rho;
   }
   if (verbose)
      cout << "No Convergence of the CG method after " << it << " iterations." << endl;
   return -it;
}


template<class T_, class M_, class P_>
int CG(const P_&P,
       const Vect<T_>& b,
             Vect<T_>& x,
             int       max_it,
             double    toler,
             int       verbose)
{
   int it=0;
   size_t size = x.Dim();
   double nrm=b.getNorm2(), rho=0, res=0, beta=0, rho_1=0;
   if (nrm == 0)
      nrm = 1;
   Vect<T_> r(size), z(size), p(size), q(size);
   r = b;
   r.MultAdd(Mxv(x),-1);

   if ((res = r.getNorm2() / nrm) <= toler) {
      if (verbose)
         cout << "Convergence of the CG method after " << it << " iterations." << endl;
      return it;
   }

   for (it=1; it<=max_it; it++) {
      P.Solve(r,z);
      rho = Dot(r,z);
      if (it==1)
         p = z;
      else {
         beta = rho/rho_1;
         p *= beta;
         p.MultAdd(z);
      }

      q = Mxv(p);
      double alpha = rho/Dot(p,q);
      x.MultAdd(p, alpha);
      r.MultAdd(q,-alpha);
      res = r.Norm2()/nrm;

      if (verbose > 1)
         cout << "\nIteration: " << it << "  ..." << endl;
      if (verbose > 2)
         cout << x;
      if (res <= toler) {
         if (verbose)
            cout << "Convergence of the CG method after " << it << " iterations." << endl;
         return it;
      }
      rho_1 = rho;
   }
   if (verbose)
      cout << "No Convergence of the CG method after " << it << " iterations." << endl;
   return -it;
}


template<class T_>
int CG(const Vect<T_> &b, Vect<T_> &x, void (*Axb)(const Vect<T_> &, Vect<T_> &),
       int max_it, double toler, int verbose)
{
   int it=0;
   size_t size = x.size();
   double nrm=b.Norm2(), rho=0, res=0, beta=0, rho_1=0;
   if (nrm == 0)
      nrm = 1;
   Vect<T_> r(size), z(size), p(size), q(size);
   (*Axb)(x,r);
   for (size_t i=1; i<=size; i++)
     r(i) = b(i) - r(i);

   if ((res=r.Norm2() / nrm) <= toler) {
      if (verbose)
         cout << "Convergence of the CG method after " << it << " iterations." << endl;
      return it;
   }

   for (it=1; it<=max_it; it++) {
      z = r;
      rho = Dot(r,z);
      if (it==1)
         p = z;
      else {
         beta = rho/rho_1;
         p *= beta;
         p.MultAdd(z);
      }

      (*Axb)(p,q);
      double alpha = rho/Dot(p,q);
      x.MultAdd(p, alpha);
      r.MultAdd(q,-alpha);
      res = r.Norm2()/nrm;

      if (verbose > 1)
         cout << "\nIteration: " << it << "  ..." << endl;
      if (verbose > 2)
         cout << x;
      if (res <= toler) {
         if (verbose)
            cout << "Convergence of the CG method after " << it << " iterations." << endl;
         return it;
      }
      rho_1 = rho;
   }
   if (verbose)
      cout << "No Convergence of the CG method after " << it << " iterations." << endl;
   return -it;
}


template<class T_>
int CG(const Prec<T_> &P, const Vect<T_> &b, Vect<T_> &x,
       void (*Axb)(const Vect<T_> &, Vect<T_> &),
       int max_it, double toler, int verbose)
{
   int it=0;
   size_t size = x.Dim();
   double nrm=b.Norm2(), rho=0, res=0, rho_1=0;
   if (nrm == 0)
      nrm = 1;
   Vect<T_> r(size), z(size), p(size), q(size);
   (*Axb)(x,r);
   for (size_t i=1; i<=size; i++)
      r(i) = b(i) - r(i);

   if ((res=r.Norm2() / nrm) <= toler) {
      if (verbose)
         cout << "Convergence of the CG method after " << it << " iterations." << endl;
      return it;
   }

   for (it=1; it<=max_it; it++) {
      P.Solve(r,z);
      rho = Dot(r,z);
      if (it==1)
         p = z;
      else {
         p *= rho/rho_1;
         p.MultAdd(z);
      }

      (*Axb)(p,q);
      double alpha = rho/Dot(p,q);
      x.MultAdd(p, alpha);
      r.MultAdd(q,-alpha);
      res = r.Norm2()/nrm;

      if (verbose > 1)
         cout << "\nIteration: " << it << "  ..." << endl;
      if (verbose > 2)
         cout << x;
      if (res <= toler) {
         if (verbose)
            cout << "Convergence of the CG method after " << it << " iterations." << endl;
         return it;
      }
      rho_1 = rho;
   }
   if (verbose)
      cout << "No Convergence of the CG method after " << it << " iterations." << endl;
   return -it;
}


template<class T_>
int _CG(const Prec<T_>* P,
        const Vect<T_>& b,
              Vect<T_>& x,
              int       max_it,
              double    toler, 
              int       verbose)
{
   int it=0;
   size_t size = x.size();
   double nrm=b.Norm2(), rho=0, res=0, beta=0, rho_1=0;
   if (nrm == 0)
      nrm = 1;
   Vect<T_> r(size), z(size), p(size), q(size);
   r = b;
   r.MultAdd(Mxv(x),-1);

   if ((res = r.Norm2() / nrm) <= toler) {
      if (verbose)
         cout << "Convergence of the CG method after " << it << " iterations." << endl;
      return it;
   }

   for (it=1; it<=max_it; it++) {
      P->Solve(r,z);
      rho = Dot(r,z);
      if (it==1)
         p = z;
      else {
         beta = rho/rho_1;
         p *= beta;
         p.MultAdd(z);
      }

      q = Mxv(p);
      double alpha = rho/Dot(p,q);
      x.MultAdd(p, alpha);
      r.MultAdd(q,-alpha);
      res = r.Norm2()/nrm;

      if (verbose > 1)
         cout << "\nIteration: " << it << "  ..." << endl;
      if (verbose > 2)
         cout << x;
      if (res <= toler) {
         if (verbose)
            cout << "Convergence of the CG method after " << it << " iterations." << endl;
         return it;
      }
      rho_1 = rho;
   }
   if (verbose)
      cout << "No Convergence of the CG method after " << it << " iterations." << endl;
   return -it;
}

/*
template<class T_>
int CG(void (*Axv)(), const Vect<T_> &b, Vect<T_> &x, int max_it,
       double toler, int verbose)
{
   int it=0;
   size_t size = x.Dim();
   double nrm=b.Norm2(), rho=0, res=0, beta=0, rho_1=0;
   if (nrm == 0)
      nrm = 1;
   Vect<T_> r(size), z(size), p(size), q(size);
   r = b;
   r.MultAdd(Mxv(x),-1);

   if ((res = r.Norm2() / nrm) <= toler) {
      if (verbose)
         cout << "Convergence of the CG method after " << it << " iterations." << endl;
      return it;
   }

   for (it=1; it<=max_it; it++) {
      P->Solve(r,z);
      rho = Dot(r,z);
      if (it==1)
         p = z;
      else {
         beta = rho/rho_1;
         p *= beta;
         p.MultAdd(z);
      }

      q = Mxv(p);
      double alpha = rho/Dot(p,q);
      x.MultAdd(p, alpha);
      r.MultAdd(q,-alpha);
      res = r.Norm2()/nrm;

      if (verbose > 1)
         cout << "\nIteration: " << it << "  ..." << endl;
      if (verbose > 2)
         cout << x;
      if (res <= toler) {
         if (verbose)
            cout << "Convergence of the CG method after " << it << " iterations." << endl;
         return it;
      }
      rho_1 = rho;
   }
   if (verbose)
      cout << "No Convergence of the CG method after " << it << " iterations." << endl;
   return -it;
   }*/

#endif /* DOXYGEN_SHOULD_SKIP_THIS */

} /* namespace OFELI */

#endif
