/*==============================================================================

                                    O  F  E  L  I

                            Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

      Class to find the smallest eigenvalues and corresponding eigenvectors
     in a generalized eigenproblem using the Bathe subspace iteration method

  ==============================================================================*/

#ifndef __EIGEN_H
#define __EIGEN_H

#include "OFELI_Config.h"

#include "linear_algebra/Vect.h"
#include "linear_algebra/DMatrix.h"
#include "linear_algebra/DSMatrix.h"
#include "linear_algebra/SkSMatrix.h"

namespace OFELI {

/*! \file Eigen.h
 *  \brief Class to find the smallest eigenvalues and corresponding
 *  eigenvectors in a generalized eigenproblem using the Bathe subspace 
 *  iteration method
 */

/*! \class Eigen
 *  \ingroup Solver
 *  \brief Class to find the smallest eigenvalues and corresponding
 *  eigenvectors in a generalized eigenproblem using the Bathe subspace 
 *  iteration method
 */

class Eigen {

public:

/// \brief Default constructor.
/// Sets default parameters
    Eigen();

/** \brief Constructor for Symmetric Skyline Matrices.
 *  @param [in] K Stiffness matrix
 *  @param [in] M "Mass" matrix
 *  \note The generalized eigenvalue problem is defined by <tt>Kx = aMx</tt>,
 *  where <tt>K</tt> and <tt>M</tt> are referred to as stiffness and mass matrix.
 *  The mass matrix can be diagonal (lumped) or not (consistent). In the 
 *  first case, it is better to choose a constructor where this matrix is stored
 *  in a vector.
 */
    Eigen(SkSMatrix<double>& K,
          SkSMatrix<double>& M);

/** \brief Constructor for Symmetric Skyline Matrices.
 *  @param [in] K "Stiffness" matrix
 *  @param [in] M Diagonal "Mass" matrix stored as a Vect instance
 *  \note The generalized eigenvalue problem is defined by <tt>Kx = aMx</tt>,
 *  where <tt>K</tt> and <tt>M</tt> are referred to as stiffness and mass matrix.
 */
    Eigen(SkSMatrix<double>& K,
          Vect<double>&      M);

/// \brief Constructor for a dense matrix.
/// @param [in] K Matrix for which eigenmodes are sought.
    Eigen(DSMatrix<double>& K);

/// \brief Destructor
    ~Eigen();

/** \brief Set matrix instances  (Symmetric matrices).
 *  \details This function is to be used when the default constructor is applied.
 *  Case where the mass matrix is consistent.
 *  @param [in] K Stiffness matrix instance
 *  @param [in] M Mass matrix instance
 */
    void setMatrix(SkSMatrix<double>& K,
                   SkSMatrix<double>& M);

/** \brief Set matrix instances (Symmetric matrices).
 *  \details This function is to be used when the default constructor is applied.
 *  Case where the mass matrix is (lumped) diagonal and stored in a vector.
 *  @param [in] K Stiffness matrix instance
 *  @param [in] M Mass matrix instance where diagonal terms are stored as a vector.
 */
    void setMatrix(SkSMatrix<double>& K,
                   Vect<double>&      M);

/** \brief Set matrix instance (Symmetric matrix).
 *  \details This function is to be used when the default constructor is applied.
 *  Case of a standard (not generalized) eigen problem is to be solved
 *  @param [in] K Stiffness matrix instance
 */
    void setMatrix(DSMatrix<double>& K);

/** \brief Run the subspace iteration solver.
 *  @param [in] nb_eigv Number of eigenvalues to be extracted
 *  @param [in] ss_dim Subspace dimension. Must be at least equal to the number
 *  eigenvalues to seek. [Default: <tt>nb_eigv</tt>
 *  @return   1: Normal execution. Convergence has been achieved. 
 *            2: Convergence for eigenvalues has not been attained. 
 */ 
    int runSubSpace(int nb_eigv,
                    int ss_dim=0);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    int runJacobi(DMatrix<double>& wm);
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/** \brief Define the subspace dimension.
 *  @param [in] dim Subspace dimension. Must be larger or equal to the
 *  number of wanted eigenvalues.
 *  By default this value will be set to the number of wanted eigenvalues
 */
    void setSubspaceDimension(int dim) { _dim = dim; }

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/// \brief set
    void setSubspaceOption(int opt) { _opt = opt; }
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/// \brief set maximal number of iterations.
    void setMaxIter(int max_it) { _max_it = max_it; }

/// \brief Return actual number of performed iterations
    int getNbIter() const { return _max_it; }

/// \brief set tolerance value
///  @param [in] eps Convergence tolerance for eigenvalues [Default: 1.e-8]
    void setTolerance(double eps) { _epsj = _epsv = eps; }

/** \brief Check how many eigenvalues have been found using Sturm sequence method
 *  @param [out] nb_found number of eigenvalues actually found
 *  @param [out] nb_lost  number of eigenvalues missing
 *  @return   - 0, Successful completion of subroutine. 
 *            - 1, No convergent eigenvalues found. 
 */
    int checkSturm(int& nb_found,
                   int& nb_lost);

/// \brief Return the n-th eigenvalue
    double getEigenValue(int n) const;

/** \brief Return the n-th eigenvector
 *  @param [in] n Label of eigenvector (They are stored in ascending order of eigenvalues)
 *  @param [in] v Vect instance where the eigenvector is stored.
 */
    void getEigenVector(int           n,
                        Vect<double>& v) const;

 private:
   bool               _lM_alloc, _diag;
   int                _max_it, _opt, _dim, _nb_eq, _nb_eigv;
   double             _epsv, _epsj;
   Matrix<double>     *_K, *_M;
   Vect<double>       *_lM, _rc, _eigv;
   DMatrix<double>    _ev;
   DSMatrix<double>   _pK, _pM;

   int init();
   void Mxv(const Vect<double>& b,
                  Vect<double>& c);
};

}  /* Namespace OFELI */

#endif
