/*==============================================================================

                                    O  F  E  L  I

                          Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                       Definition of class LinearSolver
       to select iterative or direct solver for a linear system of equations

  ==============================================================================*/

#ifndef __LINEAR_SOLVER_H
#define __LINEAR_SOLVER_H

#include <string>
#include <iostream>
using std::ostream;
using std::endl;

#include <iomanip>
using std::setw;

#include "OFELI_Config.h"
#include "util/util.h"
#include "io/output.h"
#include "solvers/CG.h"
#include "solvers/CGS.h"
#include "solvers/BiCG.h"
#include "solvers/BiCGStab.h"
#include "solvers/GMRes.h"
#include "solvers/QMR.h"

namespace OFELI {

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_> class LinearSolver
{

public:

/// \brief Default Constructor.
/// \details Initializes default parameters and pointers to 0.
    LinearSolver();

/** \brief Constructor with iteration parameters
 *  @param [in] max_it Maximal number of iterations
 *  @param [in] toler Tolerance for convergence (measured in relative weighted 2-Norm) in input,
 *  effective discrepancy in output.
 *  @param [in] verbose Information output parameter
 *  <ul>
 *    <li><tt>0</tt>: No output
 *    <li><tt>1</tt>: Output iteration information,
 *    <li><tt>2</tt> and greater: Output iteration information and solution at each iteration.
 *  </ul>
 */
    LinearSolver(int max_it, real_t tolerance, int verbose);

/** \brief Constructor using matrix, right-hand side and solution vector
 *  @param [in] A Pointer to matrix. The usage of abstract class Matrix allows any matrix class
 *  @param [in] b Vect instance that contains the right-hand side
 *  @param [in,out] x Vect instance that contains initial guess on input and solution on output
 */
    LinearSolver(      Matrix<T_>* A,
                 const Vect<T_>&   b,
                       Vect<T_>&   x);

/** \brief Constructor using matrix, right-hand side and solution vector
 *  @param [in] A SkMatrix instance that contains matrix
 *  @param [in] b Vect instance that contains the right-hand side
 *  @param [in,out] x Vect instance that contains initial guess on input and solution on output
 */
    LinearSolver(      SkMatrix<T_>& A,
                 const Vect<T_>&     b,
                       Vect<T_>&     x);

/** \brief Constructor using matrix, right-hand side
 *  @param [in] A SkMatrix instance that contains matrix
 *  @param [in,out] b Vect instance that contains the right-hand side on input and solution on output.
 *  The initial guess is the 0-vector
 */
    LinearSolver(SkMatrix<T_>& A,
                 Vect<T_>&     b);

/// Destructor
    virtual ~LinearSolver();

/// \brief Set message level
/// \details Default value is 0
    void setVerbose(int verb) { _verbose = verb; }

/// \brief Set Maximum number of iterations
/// \details Default value is 100
    void setMaxIter(int m) { _max_it = m; }

/// \brief Set tolerance value
    void setTolerance(real_t tol) { _toler = tol; }

/// \brief Set solution vector
    void setSolution(Vect<T_>& x) { _x = &x; }

/// \brief Set right-hand side vector
    void setRHS(Vect<T_>& b) { _b = &b; }

/// \brief Set matrix in the case of a pointer to matrix
/// @param [in] A Pointer to abstract Matrix class
    void setMatrix(Matrix<T_>* A) { _A = A; _matrix_set = true; }

/// \brief Set matrix in the case of a skyline matrix
/// @param [in] A %Matrix as instance of class SkMatrix
    void setMatrix(SkMatrix<T_>& A) { _A = &A; _matrix_set = true; }

/** \brief Set solver and preconditioner
 *  @param [in] s Solver identification parameter.
 *  To be chosen in the enumeration variable Iteration:\n 
 *  <tt>DIRECT_SOLVER</tt>, <tt>CG_SOLVER</tt>, <tt>CGS_SOLVER<tt>, <tt>BICG_SOLVER</tt>,
 *  <tt>BICG_STAB_SOLVER</tt>, <tt>GMRES_SOLVER</tt>, <tt>QMR_SOLVER</tt>
 *  @param [in] p Preconditioner identification parameter.
 *  To be chosen in the enumeration variable Preconditioner:\n
 *  <tt>IDENT_PREC</tt>, <tt>DIAG_PREC</tt>, <tt>SSOR_PREC</tt>, <tt>ILU_PREC</tt>
 *  @note The argument <tt>p</tt> has no effect if the solver is <tt>DIRECT_SOLVER</tt>
 */
    void setSolver(int s, 
                   int p=IDENT_PREC)
   { _s = s; _p = p; }

/// Return solver code
    int getSolver() const { return _s; }

/// Solve equations
    int solve();

/// Factorize matrix
    void setFact() { _fact = true; }

/// Do not factorize matrix
    void setNoFact() { _fact = false; }

protected:

#ifndef DOXYGEN_SHOULD_SKIP_THIS
   int              _fact, _verbose, _max_it, _s, _p, _matrix_set;
   real_t           _toler;
   Vect<T_>         *_x;
   mutable Vect<T_> *_b;
   Matrix<T_>       *_A;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

};


///////////////////////////////////////////////////////////////////////////////
//                         I M P L E M E N T A T I O N                       //
///////////////////////////////////////////////////////////////////////////////


template<class T_>
LinearSolver<T_>::LinearSolver()
{
   _max_it = 100;
   _verbose = 0;
   _toler = sqrt(OFELI_EPSMCH);
   _A = NULL;
   _b = _x = NULL;
   _s = 0;
   _p = ILU_PREC;
   _fact = true;
   _matrix_set = false;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class LinearSolver is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
LinearSolver<T_>::LinearSolver(int    max_it,
                               real_t tolerance,
                               int    verbose)
{
   _max_it = max_it; 
   _verbose = verbose;
   _toler = tolerance;
   _A = NULL;
   _b = _x = NULL;
   _s = DIRECT_SOLVER;
   _p = ILU_PREC;
   _matrix_set = false;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class LinearSolver is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
LinearSolver<T_>::LinearSolver(      Matrix<T_>* A,
                               const Vect<T_>&   b,
                                     Vect<T_>&   x)
{
   _A = A;
   _b = &b;
   _x = &x;
   _s = CG_SOLVER;
   _p = ILU_PREC;
   _max_it = 100;
   _toler = 1.e-8;
   _matrix_set = false;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class LinearSolver is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
LinearSolver<T_>::LinearSolver(      SkMatrix<T_>& A,
                               const Vect<T_>&     b,
                                     Vect<T_>&     x)
{
   _A = &A;
   _b = &b;
   _x = &x;
   _s = DIRECT_SOLVER;
   _p = ILU_PREC;
   _matrix_set = false;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class LinearSolver is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
LinearSolver<T_>::LinearSolver(SkMatrix<T_>& A,
                               Vect<T_>&     b)
{
   _A = &A;
   _b = &b;
   _x = NULL;
   _s = DIRECT_SOLVER;
   _p = ILU_PREC;
   _matrix_set = false;
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class LinearSolver is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
LinearSolver<T_>::~LinearSolver()
{
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class LinearSolver is destructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_>
int LinearSolver<T_>::solve()
{
   int ret = 0;
   try {
      if (!_matrix_set)
         THROW_RT("solve(): No matrix has been defined.");
   }
   CATCH("LinearSolver");
      
   if (_A->isDiagonal()) {
      if (_x) {
         for (size_t i=1; i<=_A->getNbRows(); i++)
            (*_x)(i) = (*_b)(i)/(*_A)(i,i);
      }
      else {
         for (size_t i=1; i<=_A->getNbRows(); i++)
            (*_x)(i) /= (*_A)(i,i);
      }
      return ret;
   }
   if (_s == DIRECT_SOLVER) {
      if (_fact)
         _A->Factor();
      if (_x)
         _A->Solve(*_b,*_x);
      else
         _A->Solve(*_b);
   }
   else {
      Prec<T_> p(_A,_p);
      if (_s == CG_SOLVER)
         ret = CG(_A,&p,*_b,*_x,_max_it,_toler,_verbose);
      else if (_s == GMRES_SOLVER)
         ret = GMRes(_A,&p,*_b,*_x,10,_max_it,_toler,_verbose);
      else if (_s == CGS_SOLVER)
         ret = _CGS(_A,&p,*_b,*_x,_max_it,_toler,_verbose);
      else if (_s == BICG_SOLVER)
         ret = BiCG(_A,&p,*_b,*_x,_max_it,_toler,_verbose);
      else if (_s == BICG_STAB_SOLVER)
         ret = BiCGStab(_A,&p,*_b,*_x,_max_it,_toler,_verbose);
      else if (_s == QMR_SOLVER)
         ret = QMR(_A,&p,&p,*_b,*_x,_max_it,_toler,_verbose);
   }
   return ret;
}

#endif /* DOXYGEN_SHOULD_SKIP_THIS */

} /* namespace OFELI */

#endif
