/*==============================================================================

                                    O  F  E  L  I

                          Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                             Preconditioner classes

  ==============================================================================*/

#ifndef __PRECOND_H
#define __PRECOND_H

#include <iostream>
using std::ostream;
using std::endl;

#include <iomanip>
using std::setw;

#include "OFELI_Config.h"
#include "linear_algebra/Vect.h"
#include "util/util.h"
#include "linear_algebra/Matrix.h"

namespace OFELI {

/*! \file Precond.h
 *  \brief Definition file for preconditioning classes.
 */

template<class T_> class SpMatrix;

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_,class M_>
int inv_diag(size_t      n,
             const M_&   A,
             vector<T_>& diag)
{
   for (int i=1; i<=int(n); i++) {
      if (A(i,i) == static_cast<T_>(0))
         return i;
      if ((diag[i-1]=static_cast<T_>(1)/A(i,i)) == static_cast<T_>(0))
         return -i;
   }
   return 0;
}

template<class T_>
int inv_diag(      size_t      n,
             const Matrix<T_>* A,
                   vector<T_>& diag)
{
   for (size_t i=1; i<=n; i++) {
      T_ d = (*A)(i,i);
      if (d == static_cast<T_>(0))
         return i;
      diag[i-1] = static_cast<T_>(1)/d;
      if (diag[i-1] == static_cast<T_>(0))
         return -i;
   }
   return 0;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

/*! \class Prec
 *  \ingroup Solver
 *  \brief To set a preconditioner.
 *  \details The preconditioner type is chosen in the constructor
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 */

template<class T_> class Prec
{

 public:

/// \brief Default constructor.
    Prec() {
       _type = -1;
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class Prec is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/** \brief Constructor that chooses preconditioner.
 *  @param [in] type Preconditioner type:
 *  <ul>
 *    <li><tt>IDENT_PREC</tt>: Identity preconditioner (No preconditioning)
 *    <li><tt>DIAG_PREC</tt>: Diagonal preconditioner
 *    <li><tt>ILU_PREC</tt>: Icomplete factorization preconditioner
 *    <li><tt>SSOR_PREC</tt>: SSOR preconditioner
 *  </ul>
 */
    Prec(int type) {
       _type = type;
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class Prec is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/** \brief Constructor using matrix of the linear system to precondition
 *  @param [in] A %Matrix to precondition
 *  @param [in] type Preconditioner type:
 *  <ul>
 *    <li><tt>IDENT_PREC</tt>: Identity preconditioner (No preconditioning)
 *    <li><tt>DIAG_PREC</tt>: Diagonal preconditioner
 *    <li><tt>ILU_PREC</tt>: Icomplete factorization preconditioner
 *    <li><tt>SSOR_PREC</tt>: SSOR preconditioner
 *  </ul>
 */
   Prec(const SpMatrix<T_>& A,
              int           type=IDENT_PREC) {
       _type = type;
       setMatrix(A);
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class Prec is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}

/** \brief Constructor using matrix of the linear system to precondition
 *  @param [in] A Pointer to abstract Matrix class to precondition
 *  @param [in] type Preconditioner type:
    <ul>
       <li> <tt>IDENT_PREC</tt>: Identity preconditioner (No preconditioning)
       <li> <tt>DIAG_PREC</tt>: Diagonal preconditioner
       <li> <tt>ILU_PREC</tt>: Icomplete factorization preconditioner
       <li> <tt>SSOR_PREC</tt>: SSOR preconditioner
    </ul>
 */
    Prec(const Matrix<T_>* A,
               int         type=IDENT_PREC) {
       _type = type;
       setMatrix(A);
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class Prec is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// \brief Destructor
    ~Prec()
    {
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class Prec is destructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/** \brief Define preconditioner type
 *  @param [in] type Preconditioner type:
    <ul>
       <li> <tt>IDENT_PREC</tt>: Identity preconditioner (No preconditioning)
       <li> <tt>DIAG_PREC</tt>: Diagonal preconditioner
       <li> <tt>ILU_PREC</tt>: Icomplete factorization preconditioner
       <li> <tt>SSOR_PREC</tt>: SSOR preconditioner
    </ul>
 */
    void setType(int type) { _type = type; }

/// \brief Define pointer to matrix for preconditioning (if this one is abstract)
/// @param [in] A %Matrix to precondition
    void setMatrix(const Matrix<T_>* A)
    {
       try {
          if (_type==-1)
             THROW_RT("setMatrix(Matrix<T_> *): Choose preconditioner before setting matrix !");
       }
       CATCH("Prec");
       _a = (SpMatrix<T_> *)A;
       _size = _a->size();
       _length = _a->getLength();

       switch (_type) {

          case IDENT_PREC:
             break;

          case DIAG_PREC:
             _pivot.resize(_size);
             int i;
             try {
                if ((i=inv_diag()) != 0)
                   THROW_RT("setMatrix(Matrix<T_> *): Zero pivot detected in row "+itos(i));
             }
             CATCH("Prec");
             break;

          case ILU_PREC:
             {
               size_t k=1;
               _pivot.resize(_size);
               _id.resize(_size);
               _row_ptr.resize(_size+1);
               _col_ind.resize(_length);
               for (size_t i=0; i<=_size; i++)
                  _row_ptr[i] = _a->getRowPtr(i+1);
               for (size_t i=0; i<_length; i++)
                  _col_ind[i] = _a->getColInd(i+1);
               for (size_t i=0; i<_size; i++) {
                  for (size_t j=_row_ptr[i]; j<_row_ptr[i+1]; j++,k++) {
                     if (_col_ind[j-1]==i+1) {
                        _id[i] = k;
                        try {
                           if ((*_a)(k) == static_cast<T_>(0))
                              THROW_RT("setMatrix(Matrix<T_> *): Zero pivot detected in row "+itos(i+1));
                        }
                        CATCH("Prec");
                        _pivot[i] = (*_a)(k);
                     }
                  }
               }
               bool found;
               T_ c = 0;
               for (size_t i=0; i<_size; i++) {
                  _pivot[i] = 1./_pivot[i];
                  for (size_t j=_id[i]+1; j<=_row_ptr[i+1]-1; j++) {
                     found = false;
                     size_t l = _col_ind[j-1];
                     for (k=_row_ptr[l-1]; k<=_id[l-1]-1; k++) {
                        if (_col_ind[k-1]==i+1) {
                           found = true;
                           c = (*_a)(k);
                        }
                     }
                     if (found)
                        _pivot[l-1] -= c*_pivot[i]*(*_a)(j);
                  }
               }
             }
             break;

          case SSOR_PREC:
             {
               size_t k=1;
               _id.resize(_size);
               _row_ptr.resize(_size+1);
               _col_ind.resize(_length);
               for (size_t i=0; i<=_size; i++)
                  _row_ptr[i] = _a->getRowPtr(i+1);
               for (size_t i=0; i<_length; i++)
                  _col_ind[i] = _a->getColInd(i+1);
               for (size_t i=0; i<_size; i++) {
                  for (size_t j=_row_ptr[i]; j<_row_ptr[i+1]; j++,k++) {
                     if (_col_ind[j-1]==i+1) {
                        _id[i] = k;
                        try {
                           if ((*_a)(k) == static_cast<T_>(0))
                              THROW_RT("setMatrix(Matrix<T_> *): Zero pivot detected in row "+itos(i+1));
                        }
                        CATCH("Prec");
                     }
                  }
               }
             }
         }
    }

/// \brief Define the matrix for preconditioning
/// @param [in] A %Matrix to precondition (instance of class SpMatrix)
    void setMatrix(const SpMatrix<T_>& A)
    {
       try {
          if (_type==-1)
             THROW_RT("setMatrix(SpMatrix<T_>): Choose preconditioner before setting matrix !");
       }
       CATCH("Prec");
       _a = &A;
       _size = _a->size();
       _length = _a->getLength();

       switch (_type) {

          case IDENT_PREC:
             break;

          case DIAG_PREC:
             _pivot.resize(_size);
             int i;
             try {
                if ((i=inv_diag()) != 0)
                   THROW_RT("setMatrix(SpMatrix<T_>): Zero pivot detected in row "+itos(i));
             }
             CATCH("Prec");
             break;

          case ILU_PREC:
             {
               size_t k=1;
               _pivot.resize(_size);
               _id.resize(_size);
               _row_ptr.resize(_size+1);
               _col_ind.resize(_length);
               for (size_t i=0; i<=_size; i++)
                  _row_ptr[i] = _a->getRowPtr(i+1);
               for (size_t i=0; i<_length; i++)
                  _col_ind[i] = _a->getColInd(i+1);
               for (size_t i=0; i<_size; i++) {
                  for (size_t j=_row_ptr[i]; j<_row_ptr[i+1]; j++, k++) {
                     if (_col_ind[j-1]==i+1) {
                        _id[i] = k;
                        try {
                           if ((*_a)(k) == static_cast<T_>(0))
                              THROW_RT("setMatrix(SpMatrix<T_>): Zero pivot detected in row "+itos(i+1));
                        }
                        CATCH("Prec");
                        _pivot[i] = (*_a)(k);
                     }
                  }
               }
               bool found;
               T_ c = static_cast<T_>(0);
               for (size_t i=0; i<_size; i++) {
                  _pivot[i] = static_cast<T_>(1)/_pivot[i];
                  for (size_t j=_id[i]+1; j<=_row_ptr[i+1]-1; j++) {
                     found = false;
                     size_t l = _col_ind[j-1];
                     for (k=_row_ptr[l-1]; k<=_id[l-1]-1; k++) {
                        if (_col_ind[k-1]==i+1) {
                           found = true;
                           c = (*_a)(k);
                        }
                     }
                     if (found)
                        _pivot[l-1] -= c*_pivot[i]*(*_a)(j);
                  }
               }
          }

          case SSOR_PREC:
             {
               size_t k=1;
               _id.resize(_size);
               _row_ptr.resize(_size+1);
               _col_ind.resize(_length);
               for (size_t i=0; i<=_size; i++)
                  _row_ptr[i] = _a->getRowPtr(i+1);
               for (size_t i=0; i<_length; i++)
                  _col_ind[i] = _a->getColInd(i+1);
               for (size_t i=0; i<_size; i++) {
                  for (size_t j=_row_ptr[i]; j<_row_ptr[i+1]; j++, k++) {
                     if (_col_ind[j-1]==i+1) {
                        _id[i] = k;
                        try {
                           if ((*_a)(k) == static_cast<T_>(0))
                              THROW_RT("setMatrix(SpMatrix<T_>): Zero pivot detected in row "+itos(i+1));
                        }
                        CATCH("Prec");
                     }
                  }
               }
             }
       }
    }

/// \brief Solve a linear system with preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void Solve(Vect<T_>& x) const
    {
      switch (_type) {

         case DIAG_PREC:
            for (size_t i=0; i<_size; i++)
               x[i] *= _pivot[i];
            break;

         case ILU_PREC:
            {
              Vect<T_> z(_size);
              for (size_t i=0; i<_size; i++) {
                 T_ s = 0;
                 for (size_t j=_row_ptr[i]; j<_id[i]; j++)
                    s += (*_a)[j-1] * z(_col_ind[j-1]);
                 z[i] = _pivot[i] * (x[i]-s);
              }
              for (int i=int(_size)-1; i>=0; i--) {
                 T_ s = 0;
                 for (size_t j=_id[i]+1; j<_row_ptr[i+1]; j++) {
                    s += (*_a)[j-1] * x(_col_ind[j-1]);
                    x[i] = z[i] - _pivot[i] * s;
                 }
              }
            }

         case SSOR_PREC:
            {
              Vect<T_> z(_size);
              for (size_t i=0; i<_size; i++) {
                 T_ s = 0;
                 for (size_t j=_row_ptr[i]; j<_id[i]; j++)
                    s += (*_a)[j-1] * z(_col_ind[j-1]);
                 z[i] = (x[i]-s)/(*_a)(_id[i]);
              }
              for (int i=int(_size)-1; i>=0; i--) {
                 T_ s = 0;
                 for (size_t j=_id[i]+1; j<_row_ptr[i+1]; j++) {
                    s += (*_a)[j-1] * x(_col_ind[j-1]);
                    x[i] = z[i] - s/(*_a)(_id[i]);
                 }
              }
            }
      }
    }

/** \brief Solve a linear system with preconditioning matrix.
 *  @param [in] b Right-hand side
 *  @param [out] x Solution vector
 */
    void Solve(const Vect<T_>& b,
                     Vect<T_>& x) const
    {

      switch (_type) {

         case IDENT_PREC:
            x = b;
            break;

         case DIAG_PREC:
            for (size_t i=0; i<_size; i++)
               x[i] = b[i] * _pivot[i];
            break;

         case ILU_PREC:
            {
              Vect<T_> z(_size);
              for (size_t i=0; i<_size; i++) {
                 T_ s = 0;
                 for (size_t j=_row_ptr[i]; j<_id[i]; ++j)
                    s += (*_a)[j-1] * z[_col_ind[j-1]-1];
                 z[i] = _pivot[i] * (b[i]-s);
              }
              for (size_t i=0; i<_size; i++) {
                 T_ s = 0;
                 for (size_t j=_id[_size-i-1]+1; j<_row_ptr[_size-i]; j++)
                    s += (*_a)[j-1] * x(_col_ind[j-1]);
                 x[_size-i-1] = z[_size-i-1] - _pivot[_size-i-1] * s;
              }
            }
            break;

          case SSOR_PREC:
             {
               Vect<T_> z(_size);
               for (size_t i=0; i<_size; i++) {
                  T_ s = 0;
                  for (size_t j=_row_ptr[i]; j<_id[i]; j++)
                     s += (*_a)[j-1] * z(_col_ind[j-1]);
                  z[i] = (b[i]-s)/(*_a)(_id[i]);
               }
               for (int i=int(_size-1); i>=0; i--) {
                  T_ s = 0;
                  for (size_t j=_id[i]+1; j<_row_ptr[i+1]; j++)
                     s += (*_a)[j-1] * x(_col_ind[j-1]);
                  x[i] = z[i] - s/(*_a)(_id[i]);
               }
            }
            break;
       }
    }

/// \brief Solve a linear system with transposed preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void TransSolve(Vect<T_>& x) const
    {
       switch (_type) {

          case DIAG_PREC:
             Solve(x);
             break;

          case ILU_PREC:
             {
                Vect<T_> z(_size), temp(x);
                T_ tmp;
                for (size_t i=0; i<_size; i++) {
                   z[i] = temp[i];
                   tmp = _pivot[i] * z[i];
                   for (size_t j=_id[i]+1; j<_row_ptr[i+1]; j++)
                      temp(_col_ind[j-1]) -= tmp * (*_a)(j);
                }
                for (int i=_size-1; i>=0; i--) {
                   x[i] = _pivot[i]*z[i];
                   for (size_t j=_row_ptr[i]; j<_id[i]; j++)
                      z(_col_ind[j-1]) -= (*_a)(j) * x[i];
                }
             }
             break;
       }
    }


/** \brief Solve a linear system with transposed preconditioning matrix.
 *  @param [in] b Right-hand side vector
 *  @param [out] x Solution vector
 */
    void TransSolve(const Vect<T_>& b,
                          Vect<T_>& x) const { x = b; TransSolve(x); }

/// Return i-th pivot of preconditioning matrix
    T_ & getPivot(size_t i) const { return _pivot[i-1]; }

 private:
   vector<T_>         _pivot;
   const SpMatrix<T_> *_a;
   size_t             _size, _length;
   vector<size_t>     _id, _row_ptr, _col_ind;
   int                _type;

   int inv_diag()
   {
      for (size_t i=1; i<=_a->getNbRows(); i++) {
         if ((*_a)(i,i) == static_cast<T_>(0))
            return i;
         if ((_pivot[i-1]=static_cast<T_>(1)/(*_a)(i,i)) == static_cast<T_>(0))
            return -int(i);
      }
      return 0;
   }
};

//==============================================================================
//                                Class Precond
//==============================================================================

/*! \class Precond
 *  \ingroup Solver
 *  \brief An abstract class from which derive all preconditioning classes.
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 */

template<class T_> class Precond
{

 public :

/// \brief Default constructor
    Precond() { }

/// \brief Constructor using matrix (A pointer to abstract Matrix class)
    Precond (const Matrix<T_>* A) { }

/// \brief Destructor
    virtual ~Precond() { }

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    virtual void Solve(Vect<T_>& x) const = 0;
    virtual void Solve(const Vect<T_>& b,
                             Vect<T_>& x) const = 0;
    virtual void TransSolve(Vect<T_>& x) const = 0;
    virtual void TransSolve(const Vect<T_>& b,
                                  Vect<T_>& x) const = 0;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

 protected:
#ifndef DOXYGEN_SHOULD_SKIP_THIS
   vector<T_>        _pivot;
   vector<size_t>    _id, _col_ind, _row_ptr;
   const Matrix<T_>* _a;
   size_t            _size, _length;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

};

/*-----------------------------------------------------------------------------
                             Identity Preconditioner
  -----------------------------------------------------------------------------*/

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/*! \class IdentPrec
 *  \ingroup Solver
 *  \brief To set identity preconditioner (No preconditioning)
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 * \tparam <M_> %Matrix class
 */

template<class T_,class M_> class IdentPrec
{

 public:

/// \brief Constructor using matrix of the linear system to precondition
    IdentPrec(const M_& A);

/// \brief Destructor
    ~IdentPrec();

/// \brief Solve a linear system with preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void Solve(Vect<T_>& x) const;

/** \brief Solve a linear system with preconditioning matrix.
 *  @param [in] b Right-hand side.
 *  @param [out] x Solution vector.
 */
    void Solve(const Vect<T_>& b,
                     Vect<T_>& x) const;

/// \brief Solve a linear system with transposed preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void TransSolve(Vect<T_>& x) const { }

/** \brief Solve a linear system with transposed preconditioning matrix
 *  @param [in] b Right-hand side vector
 *  @param [out] x Solution vector
 */
    void TransSolve(const Vect<T_>& b,
                          Vect<T_>& x) const;

};
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


//==============================================================================
//                                Class IdentPrecond
//==============================================================================


#ifndef DOXYGEN_SHOULD_SKIP_THIS
/*! \class IdentPrecond
 *  \ingroup Solver
 *  \brief To set identity preconditioner (No preconditioning)
 *
 * This class uses one template argument.
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 */

template<class T_> class IdentPrecond : virtual public Precond<T_>
{

 public:

/// \brief Constructor using matrix of the linear system to precondition
    IdentPrecond(const Matrix<T_>* A);

/// \brief Destructor
    ~IdentPrecond();

/// \brief Solve a linear system with preconditioning matrix.
/// @param [in] x Right-hand side in input and solution in output.
    void Solve(Vect<T_>& x) const { }

/** \brief Solve a linear system with preconditioning matrix.
    @param [in] b Right-hand side
    @param [out] x Solution vector
*/
    void Solve(const Vect<T_>& b,
                     Vect<T_>& x) const;

/// \brief Solve a linear system with transposed preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void TransSolve(Vect<T_>& x) const { }

/** \brief Solve a linear system with transposed preconditioning matrix.
    @param [in] b Right-hand side.
    @param [out] x Solution vector.
*/
    void TransSolve(const Vect<T_>& b,
                          Vect<T_>& x) const;

};
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


//==============================================================================
//                                CLASS DiagPrec
//==============================================================================

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/*! \class DiagPrec
 *  \ingroup Solver
 *  \brief To set diagonal preconditioner.
 *  All matrix elements are divided by diagonal elements.
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 * \tparam <M_> %Matrix class
 */

template<class T_, class M_> class DiagPrec
{

 public:

/// \brief Constructor using matrix of the linear system to precondition
    DiagPrec(const M_& a)
    {
       _size = a.size();
       _pivot.resize(_size);
       _a = &a;
       int i;
       if ((i = inv_diag(_size, *_a, _pivot)) != 0) {
          cerr << "\n*** Fatal Error in OFELI ***\nIn DiagPrec::DiagPrec(M_ &) : ";
          cerr << "Zero detected in element " << i << endl;
          cerr << "Program stops\n";
          exit(1);
       }
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class DiagPrec is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// \brief Destructor
    ~DiagPrec()
    {
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class DiagPrec is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// \brief Solve a linear system with preconditioning matrix.
/// @param x [in,out] Right-hand side in input and solution in output.
    void Solve(Vect<T_>& x) const
    {
       for (size_t i=0; i<_size; i++)
          x[i] *= _pivot[i];
    }

/** \brief Solve a linear system with preconditioning matrix.
    @param [in] b Right-hand side
    @param [out] x Solution vector
*/
    void Solve(const Vect<T_>& b,
                     Vect<T_>& x) const
    {
       for (size_t i=0; i<_size; i++)
          x[i] = b[i] * _pivot[i];
    }

/// \brief Solve a linear system with transposed preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void TransSolve(Vect<T_>& x) const { Solve(x); }

/// \brief Solve a linear system with transposed preconditioning matrix.
/// @param [in] b Right-hand side.
/// @param [out] x Solution vector.
    void TransSolve(const Vect<T_>& b,
                          Vect<T_>& x) const { Solve(b,x); }

/// \brief Return i-th diagonal entry of preconditioning matrix
    const T_ & getDiag(size_t i) const { return _pivot[i-1]; }

/// \brief Return i-th pivot of preconditioning matrix
    T_ & getPivot(size_t i) const { return _pivot[i-1]; }

 private:
   vector<T_>     _pivot;
   const M_       *_a;
   size_t         _size, _length;
   vector<size_t> _row_ptr, _col_ind;
};
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


//==============================================================================
//                                CLASS DiagPrecond
//==============================================================================

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_> class DiagPrecond : virtual public Precond<T_>
{

 public:

/// Constructor
    DiagPrecond(const Matrix<T_>* a)
    {
      _size = a->size();
      _pivot.resize(_size);
      _a = a;
      int i;
      if ((i = inv_diag(_size, _a, _pivot)) != 0) {
        cerr << "\n*** Fatal Error in OFELI ***\nIn DiagPrecond::DiagPrecond(...) : ";
        cerr << "Zero detected in element " << i << endl;
        cerr << "Program stops\n";
        exit(1);
      }
#ifdef _OFELI_DEBUG
      std::clog << "An instance of class DiagPrecond is constructed.\n";
      std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// Destructor
    ~DiagPrecond()
    {
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class DiagPrecond is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// Solve a linear system with preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void Solve(Vect<T_>& x) const
    {
       for (size_t i=0; i<_size; i++)
          x[i] *= _pivot[i];
    }

/** Solve a linear system with preconditioning matrix.
    @param [in] b Right-hand side
    @param [out] x Solution vector
*/
    void Solve(const Vect<T_>& b, Vect<T_>& x) const
    {
       for (size_t i=0; i<_size; i++)
          x[i] = b[i] * _pivot[i];
    }

/// Solve a linear system with transposed preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void TransSolve(Vect<T_>& x) const { Solve(x); }

/** Solve a linear system with transposed preconditioning matrix.
    @param [in] b Right-hand side.
    @param [out]x  Solution vector.
*/
    void TransSolve(const Vect<T_>& b,
                          Vect<T_>& x) const { Solve(b,x); }

/// Return i-th diagonal entry of preconditioning matrix.
    const T_ & getDiag(size_t i) const { return _pivot[i-1]; }

/// Return i-th pivot of preconditioning matrix.
    T_ & getPivot(size_t i) const { return _pivot[i-1]; }

   using Precond<T_>::_size;
   using Precond<T_>::_pivot;
   using Precond<T_>::_a;
};
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


//==============================================================================
//                                CLASS ILUPrec
//==============================================================================

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/*! \class ILUPrec
 *  \ingroup Solver
 *  \brief To set ILU (Incomplete LU factorization) preconditioner
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 * \tparam <M_> %Matrix class
 */

template<class T_, class M_> class ILUPrec
{

 public:

/// Default constructor.
/// Does nothing: Just to declare an instance
    ILUPrec() {
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class ILUPrec is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// Constructor using matrix of the linear system to precondition
    ILUPrec(const M_& a) {
       setMatrix(a);
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class ILUPrec is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// Destructor
    ~ILUPrec()
    {
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class ILUPrec is destructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// Define the matrix for preconditioning
/// @param [in] a %Matrix for preconditioning
    void setMatrix(const M_& a)
    {
       size_t k=1;
       _a = &a;
       _size = _a->size();
       _length = _a->getLength();
       _pivot.resize(_size);
       _id.resize(_size);
       _row_ptr.resize(_size+1);
       _col_ind.resize(_length);
       for (size_t i=0; i<=_size; i++)
          _row_ptr[i] = _a->getRowPtr(i+1);
       for (size_t i=0; i<_length; i++)
          _col_ind[i] = _a->getColInd(i+1);
       for (size_t i=0; i<_size; i++) {
          for (size_t j=_row_ptr[i]; j<_row_ptr[i+1]; j++,k++) {
             if (_col_ind[j-1]==i+1) {
                _id[i] = k;
                if ((*_a)(k)==0) {
                   cerr << "\n*** Fatal Error in OFELI ***\nIn ILUPrec::setMatrix(M_ &) : ";
                   cerr << "Zero detected in element " << i+1 << endl;
                   cerr << "Program stops\n";
                   exit(1);
                }
                _pivot[i] = (*_a)(k);
             }
          }
       }

       bool found;
       T_ c = 0;
       for (size_t i=0; i<_size; i++) {
          _pivot[i] = 1./_pivot[i];
          for (size_t j=_id[i]+1; j<_row_ptr[i+1]; j++) {
             found = false;
             size_t l = _col_ind[j-1];
             for (size_t k=_row_ptr[l-1]; k<_id[l-1]; k++) {
                if (_col_ind[k-1]==i+1) {
                   found = true;
                   c = (*_a)(k);
                }
             }
             if (found)
                _pivot[l-1] -= c * _pivot[i] * (*_a)(j);
          }
       }
    }

/// Solve a linear system with preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void Solve(Vect<T_>& x) const
    {
       Vect<T_> z(_size);
       for (size_t i=0; i<_size; i++) {
          T_ s = 0;
          for (size_t j=_row_ptr[i]; j<_id[i]; j++)
             s += (*_a)[j-1] * z(_col_ind[j-1]);
          z[i] = _pivot[i] * (x[i]-s);
       }
       for (int i=_size-1; i>=0; i--) {
          T_ s = 0;
          for (int j=_id[i]+1; j<_row_ptr[i+1]; j++) {
             s += (*_a)[j-1] * x(_col_ind[j-1]);
             x[i] = z[i] - _pivot[i] * s;
          }
       }
    }

/** Solve a linear system with preconditioning matrix.
 *  @param [in] b Right-hand side vector
 *  @param [out] x Solution vector
 */
    void Solve(const Vect<T_>& b,
                     Vect<T_>& x) const
    {
       Vect<T_> z(_size);
       for (size_t i=0; i<_size; i++) {
          T_ s = 0;
          for (size_t j=_row_ptr[i]; j<_id[i]; j++)
             s += (*_a)[j-1] * z(_col_ind[j-1]);
          z[i] = _pivot[i] * (b[i]-s);
       }
       for (int i=int(_size-1); i>=0; i--) {
          T_ s = 0;
          for (size_t j=_id[i]+1; j<_row_ptr[i+1]; j++)
             s += (*_a)[j-1] * x(_col_ind[j-1]);
          x[i] = z[i] - _pivot[i] * s;
       }
    }

/// Solve a linear system with transposed preconditioning matrix
/// @param [in,out] x Right-hand side vector on input and solution vector on output
    void TransSolve(Vect<T_>& x) const
    {
       int i, j;
       Vect<T_> z(_size), temp(x);
       T_ tmp;

       for (i=0; i<_size; i++) {
          z[i] = temp[i];
          tmp = _pivot[i] * z[i];
          for (j=_id[i]+1; j<_row_ptr[i+1]; j++)
             temp(_a->getColInd(j)) -= tmp * (*_a)(j);
       }

       for (i=_size-1; i>=0; i--) {
          x[i] = _pivot[i] * z[i];
          for (j=_row_ptr[i]; j<_id[i]; j++)
             z(_a->getColInd(j)) -= (*_a)(j) * x[i];
       }
    }

/// Solve a linear system with transposed preconditioning matrix.
/// @param [in] b Right-hand side vector
/// @param [out] x Solution vector
    void TransSolve(const Vect<T_>& b,
                          Vect<T_>& x) const { x = b; TransSolve(x); }

/// Return i-th pivot of preconditioning matrix
    T_ & getPivot(size_t i) const { return _pivot[i-1]; }

 private:
  vector<T_>        _pivot;
  const M_         *_a;
   size_t         _size, _length;
   vector<size_t> _row_ptr, _col_ind, _id;
};
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

//==============================================================================
//                                CLASS ILUPrecond
//==============================================================================

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_> class ILUPrecond : virtual public Precond<T_>
{

 public:

/// Constructor
    ILUPrecond(const Matrix<T_>* A)
    {
       size_t i, j, k=1;
       _size = A->size();
       _pivot.resize(_size);
       _id.resize(_size);
       _a = A;

       for (i=0; i<_size; i++) {
          for (j=_a->getRowPtr(i+1); j<_a->getRowPtr(i+2); j++) {
             size_t jj = _a->getColInd(j);
             if (jj==i+1) {
                _id[i] = k;
                if ((*_a)(i+1,i+1)==0) {
                   cerr << "\n*** Fatal Error in OFELI ***\nIn ILUPrecond::ILUPrecond(...) : ";
                   cerr << "Zero detected in element " << i+1 << endl;
                   cerr << "Program stops\n";
                   exit(1);
                }
                _pivot[i] = (*_a)(i+1,i+1);
             }
             k++;
          }
       }
       T_ c = 0;
       for (i=0; i<_size; i++) {
          _pivot[i] = static_cast<T_>(1)/_pivot[i];
          for (j=_id[i]+1; j<=_a->getRowPtr(i+2)-1; j++) {
             bool found = false;
             size_t l = _a->getColInd(j);
             for (k=_a->getRowPtr(l); k<=_id[l-1]-1; k++) {
                if (_a->getColInd(k)==i+1) {
                   found = true;
                   c = (*_a)(i+1,i+1);
                }
             }
             if (found)
                _pivot[l-1] -= c * _pivot[i] * (*_a)(i+1,l);
          }
       }
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class ILUPrecond is constructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// Destructor
    ~ILUPrecond()
    {
#ifdef _OFELI_DEBUG
       std::clog << "An instance of class ILUPrecond is destructed.\n";
       std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
    }

/// Solve a linear system with preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void Solve(Vect<T_>& x) const
    {
       size_t i, j;
       Vect<T_> z(_size);
       for (i=0; i<_size; i++) {
          T_ s = 0;
          for (j=_a->getRowPtr(i+1); j<=_id[i]-1; j++) {
             size_t k = _a->getColInd(j);
             s += (*_a)(i+1,k) * z[k-1];
          }
          z[i] = _pivot[i] * (x[i]-s);
       }

       for (int ii=int(_size-1); ii>=0; ii--) {
          T_ s = 0;
          for (j=_id[ii]+1; j<=_a->getRowPtr(ii+2)-1; j++) {
             size_t k = _a->getColInd(j);
             s += (*_a)(ii+1,k) * x[k-1];
          }
          x[ii] = z[ii] - _pivot[ii] * s;
       }
    }

/** Solve a linear system with preconditioning matrix.
 *  @param [in] b Right-hand side.
 *  @param [out] x Solution vector.
 */
    void Solve(const Vect<T_>& b,
                     Vect<T_>& x) const { x = b; Solve(x); }

/// Solve a linear system with transposed preconditioning matrix.
/// @param [in,out] x Right-hand side in input and solution in output.
    void TransSolve(Vect<T_>& x) const
    {
//     This is just to avoid warning on unused variable
       if (x.size()==0) { }
    }

/** Solve a linear system with transposed preconditioning matrix.
 *  @param [in] b Right-hand side.
 *  @param [out] x Solution vector.
 */
    void TransSolve(const Vect<T_>& b,
                          Vect<T_>& x) const { x = b; TransSolve(x); }

/// Return i-th pivot of preconditioning matrix
    T_ & getPivot(size_t i) const { return _pivot[i-1]; }


    using Precond<T_>::_size;
    using Precond<T_>::_pivot;
    using Precond<T_>::_id;
    using Precond<T_>::_a;
};
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


///////////////////////////////////////////////////////////////////////////////
//                         I M P L E M E N T A T I O N                       //
///////////////////////////////////////////////////////////////////////////////

#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_,class M_>
IdentPrec<T_,M_>::IdentPrec(const M_& A)
{
   A.getNbRows();
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class IdentPrec is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_,class M_>
IdentPrec<T_,M_>::~IdentPrec()
{
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class IdentPrec is destructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


template<class T_,class M_>
void IdentPrec<T_,M_>::Solve(Vect<T_>& x) const
{
   x.size();
}


template<class T_,class M_>
void IdentPrec<T_,M_>::Solve(const Vect<T_>& b,
                                   Vect<T_>& x) const
{
   x = b;
}

template<class T_,class M_>
void IdentPrec<T_,M_>::TransSolve(const Vect<T_>& b,
                                        Vect<T_>& x) const
{
   x = b;
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
IdentPrecond<T_>::IdentPrecond(const Matrix<T_>* A)
{
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class IdentPrecond is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
IdentPrecond<T_>::~IdentPrecond()
{
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class IdentPrecond is destructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void IdentPrecond<T_>::Solve(const Vect<T_>& b,
                                   Vect<T_>& x) const
{
   x = b;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template<class T_>
void IdentPrecond<T_>::TransSolve(const Vect<T_>& b,
                                        Vect<T_>& x) const
{
   x = b;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

} /* namespace OFELI */

#endif
