/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

             Template function for Quasi-Minimal Residual Method
                  QMR is inspired from the SIAM Templates book

  ------------------------------------------------------------------------------
     Return Values
         0        Convergence within max_iter iterations
        -1        No convergence after max_iter iterations
                  Breakdown in :
        -2          rho
        -3          beta
        -4          gamma
        -5          delta
        -6          ep
        -7          xi
  ==============================================================================*/

#ifndef __QMR_H
#define __QMR_H

#include <iostream>
using std::ostream;
using std::endl;

#include <iomanip>
using std::setw;
using std::ios;
using std::setprecision;


#include "OFELI_Config.h"
#include "solvers/Precond.h"
#include "util/util.h"
#include "io/output.h"

namespace OFELI {

/*! \file QMR.h
 *  \brief Function to solve a linear system of equations
 *  using the Quasi-Minimal Residual method.
 *
 */

/** \fn int QMR(const M_ &A, const P1_ &P1, const P2_ &P2, const Vect<real_t> &b, Vect<real_t> &x, int max_it, real_t &toler, int verbose)
 *  \ingroup Solver
 * \brief QMR solver function.
 * @param [in] A Problem matrix (Instance of abstract class \b M_).
 * @param [in] P1 First preconditioner for the Conjugate Gradient Squared.
 * @param [in] P2 Second preconditioner for the Conjugate Gradient Squared.
 * @param [in] A Problem %matrix (Instance of abstract class \b M_).
 * @param [in] b Right-hand side vector (class Vect)
 * @param [in,out] x Vect instance containing initial solution guess in input and solution
 * of the linear system in output (If iterations have succeeded).
 * @param [in] max_it Maximum number of iterations.
 * @param [in] toler Tolerance for convergence (measured in relative weighted 2-Norm).
 * @param [in] verbose Information output parameter (0: No output, 1: Output iteration information,
 * 2 and greater : Output iteration information and solution at each iteration.
 * @return Number of iterations:
 * <ul>
 *   <li><tt>0</tt>, Convergence within max_iter iterations
 *   <li><tt>-1</tt>, No convergence after max_iter iterations
 *   <li><tt>-2</tt>, Breakdown in rho
 *   <li><tt>-3</tt>, Breakdown in beta
 *   <li><tt>-4</tt>, Breakdown in gamma
 *   <li><tt>-5</tt>, Breakdown in delta
 *   <li><tt>-6</tt>, Breakdown in ep
 *   <li><tt>-7</tt>, Breakdown in xi
 * </ul>
 *
 * Template Arguments:
 *
 *  \arg \b M_  %Matrix storage class\n
 *  \arg \b P1_ Class of first preconditioner\n
 *  \arg \b P2_ Class of second preconditioner
 *
 */
template<class M_, class P1_ , class P2_>
int QMR(const M_&           A,
        const P1_&          P1,
        const P2_&          P2,
        const Vect<real_t>& b,
              Vect<real_t>& x,
              int           max_it,
              real_t&       toler,
              int           verbose)
{
   real_t res, rho, rho_1, xi, gamma, gamma_1;
   real_t theta, theta_1, beta=0, ep=0, delta=0, eta=0;
   size_t size = x.size();
   Vect<real_t> r(size), v(size), vv(size), y(size), yy(size);
   Vect<real_t> w(size), ww(size), z(size), zz(size);
   Vect<real_t> p(size), q(size), pp(size), d(size), s(size);
   real_t nrm = b.getNorm2();
   A.Mult(x,r);
   r = b - r;
   if (nrm == 0.0)
      nrm = 1;

   if ((res = r.getNorm2() / nrm) <= toler) {
      toler = res;
      return 0;
   }

   vv = r;
   P1.Solve(vv,y);
   rho = y.getNorm2();

   ww = r;
   P2.TransSolve(ww,z);
   xi = z.getNorm2();
   gamma = 1.0;
   eta = -1.0;
   theta = 0.0;
   verbose = 0;

   int it;
   for (it=1; it<=max_it; it++) {
      if (rho == 0.0)
         return -2;
      if (xi == 0.0)
         return -7;

      v = 1./rho*vv;
      y = 1./rho*y;
      w = 1./xi*ww;
      z = 1./xi*z;

      delta = Dot(z,y);
      if (delta == 0.0)
         return -5;

      P2.Solve(y,yy);
      P1.TransSolve(z,zz);

      if (it > 1) {
         p = yy - (xi*delta/ep)*p;
         q = zz - (rho*delta/ep)*q;
      } else {
         p = yy;
         q = zz;
      }

      A.Mult(p,pp);
      ep = Dot(q,pp);
      if (ep == 0.0)
         return -6;

      beta = ep/delta;
      if (beta == 0.0)
         return -3;
      vv = pp - beta*v;
      P1.Solve(vv,y);
      rho_1 = rho;
      rho = y.getNorm2();
      A.TMult(q,ww);
      ww -= beta*w;
      P2.TransSolve(ww,z);
      xi = z.getNorm2();

      gamma_1 = gamma;
      theta_1 = theta;
      theta = rho / (gamma_1 * beta);
      gamma = 1.0 / sqrt(1.0 + theta*theta);
      if (gamma == 0.0)
         return -4;

      eta = -eta*rho_1*gamma*gamma / (beta*gamma_1*gamma_1);

      if (it > 1) {
         d = eta*p  + (theta_1*theta_1*gamma*gamma) * d;
	 s = eta*pp + (theta_1*theta_1*gamma*gamma) * s;
      } else {
         d = eta*p;
         s = eta*pp;
      }
      x += d;
      r -= s;

      if ((res = r.getNorm2() / nrm) <= toler) {
         toler = res;
         return it;
      }
   }
   toler = res;
   return -it;
}


#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** \fn int QMR(const M_ &A, const P1_ &P1, const P2_ &P2, const Vect<complex_t> &b, Vect<complex_t> &x, int max_it, real_t &toler, int verbose)
 *  \ingroup Solver
 * \brief QMR solver function.
 * @param [in] A Problem matrix (Instance of abstract class \b M_).
 * @param [in] P1 First preconditioner for the Conjugate Gradient Squared.
 * @param [in] P2 Second preconditioner for the Conjugate Gradient Squared.
 * @param [in] A Problem %matrix (Instance of abstract class \b M_).
 * @param [in] b Right-hand side vector (class Vect)
 * @param [in,out] x Vect instance containing initial solution guess in input and solution
 * of the linear system in output (If iterations have succeeded).
 * @param [in] max_it Maximum number of iterations.
 * @param [in] toler Tolerance for convergence (measured in relative weighted 2-Norm).
 * @param [in] verbose Information output parameter (0: No output, 1: Output iteration information,
 * 2 and greater : Output iteration information and solution at each iteration.
 * @return Number of iterations
 *    -  0, Convergence within max_iter iterations
 *    - -1, No convergence after max_iter iterations
 *    - -2, Breakdown in rho
 *    - -3, Breakdown in beta
 *    - -4, Breakdown in gamma
 *    - -5, Breakdown in delta
 *    - -6, Breakdown in ep
 *    - -7, Breakdown in xi
 *
 * Template Arguments:
 *
 *  \arg \b M_  %Matrix storage class\n
 *  \arg \b P1_ Class of first preconditioner\n
 *  \arg \b P2_ Class of second preconditioner
 *
 */
template<class M_, class P1_ , class P2_>
int QMR(const M_&              A,
        const P1_&             P1,
        const P2_&             P2,
        const Vect<complex_t>& b,
        Vect<complex_t>&       x,
        int                    max_it,
        real_t&                toler,
        int                    verbose)
{
   int it;
   complex_t res, rho, rho_1, xi, gamma, gamma_1, theta, theta_1, beta, omega;
   complex_t ep=0, delta=0, eta=0;
   size_t size = x.size();
   Vect<complex_t> r(size), v(size), vv(size), y(size), yy(size);
   Vect<complex_t> w(size), ww(size), z(size), zz(size);
   Vect<complex_t> p(size), q(size), pp(size), d(size), s(size);
   real_t nrm = b.getNorm2();
   A.Mult(x,r);
   r = b - r;
   if (nrm == 0.0)
      nrm = 1;

   if (Abs((res = r.getNorm2()) / Abs(nrm)) <= toler) {
      toler = Abs(res);
      return 0;
   }

   vv = r;
   P1.Solve(vv,y);
   rho = y.getNorm2();

   ww = r;
   P2.TransSolve(ww,z);
   xi = z.getNorm2();
   gamma =  1.0;
   eta   = -1.0;
   theta =  0.0;

   for (it=1; it<=max_it; it++) {
      if (rho == 0.0)
         return -2;
      if (xi == 0.0)
         return -7;

      v = complex_t(1./rho)*vv;
      y = complex_t(1./rho)*y;
      w = complex_t(1./xi)*ww;
      z = complex_t(1./xi)*z;

      delta = Dot(z,y);
      if (delta == 0.0)
         return -5;

      P2.Solve(y,yy);
      P1.TransSolve(z,zz);

      if (it > 1) {
         p = yy - (xi*delta/ep)*p;
         q = zz - (rho*delta/ep)*q;
      } else {
         p = yy;
         q = zz;
      }

      A.Mult(p,pp);
      ep = Dot(q,pp);
      if (ep == 0.0)
         return -6;

      beta = ep/delta;
      if (beta == 0.0)
         return -3;
      vv = pp - complex_t(beta)*v;
      P1.Solve(vv,y);
      rho_1 = rho;
      rho = y.getNorm2();
      A.TMult(q,ww);
      ww -= complex_t(beta)*w;
      P2.TransSolve(ww,z);
      xi = z.getNorm2();

      gamma_1 = gamma;
      theta_1 = theta;
      theta = rho / (gamma_1 * beta);
      gamma = 1.0 / sqrt(1.0 + theta*theta);
      if (gamma == 0.0)
         return -4;

      eta = -eta*rho_1*gamma*gamma / (beta*gamma_1*gamma_1);

      if (it > 1) {
         omega = theta_1*theta_1*gamma*gamma;
         d = eta*p  + complex_t(omega)*d;
	 s = eta*pp + complex_t(omega)*s;
      } else {
         d = eta*p;
         s = eta*pp;
      }
      x += d;
      r -= s;

      if ((Abs(res = r.getNorm2()) / Abs(nrm)) <= toler) {
         toler = Abs(res);
         return it;
      }
   }

   toler = Abs(res);
   return -it;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS

inline int QMR(      Matrix<real_t>* A,
                     Prec<real_t>*   P1,
                     Prec<real_t>*   P2,
               const Vect<real_t>&   b,
                     Vect<real_t>&   x,
                     int             max_it,
                     real_t&         toler,
                     int             verbose)
{
   verbose = 0;
   real_t res, rho, rho_1, xi, gamma, gamma_1;
   real_t theta, theta_1, ep=0, delta=0, eta=0, beta=0;
   size_t size = x.size();
   Vect<real_t> r(size), v(size), vv(size), y(size), yy(size);
   Vect<real_t> w(size), ww(size), z(size), zz(size);
   Vect<real_t> p(size), q(size), pp(size), d(size), s(size);
   real_t nrm = b.getNorm2();
   A->Mult(x,r);
   r = b - r;
   if (nrm == 0.0)
      nrm = 1;

   if ((res = r.getNorm2() / nrm) <= toler) {
      toler = res;
      return 0;
   }

   vv = r;
   P1->Solve(vv,y);
   rho = y.getNorm2();

   ww = r;
   P2->TransSolve(ww,z);
   xi = z.getNorm2();
   gamma = 1.0; eta = -1.0; theta = 0.0;

   int it;
   for (it=1; it<=max_it; it++) {
      if (rho == 0.0)
         return -2;
      if (xi == 0.0)
         return -7;

      v = (1./rho)*vv;
      y = (1./rho)*y;
      w = (1./xi)*ww;
      z = (1./xi)*z;

      delta = Dot(z,y);
      if (delta == 0.0)
         return -5;

      P2->Solve(y,yy);
      P1->TransSolve(z,zz);

      if (it > 1) {
         p = yy - (xi*delta/ep)*p;
         q = zz - (rho*delta/ep)*q;
      } else {
         p = yy;
         q = zz;
      }

      A->Mult(p,pp);
      ep = Dot(q,pp);
      if (ep == 0.0)
         return -6;

      beta = ep/delta;
      if (beta == 0.0)
         return -3;
      vv = pp - beta*v;
      P1->Solve(vv,y);
      rho_1 = rho;
      rho = y.getNorm2();
      A->TMult(q,ww);
      ww -= beta*w;
      P2->TransSolve(ww,z);
      xi = z.getNorm2();

      gamma_1 = gamma; theta_1 = theta;
      theta = rho / (gamma_1 * beta);
      gamma = 1.0 / sqrt(1.0 + theta*theta);
      if (gamma == 0.0)
         return -4;

      eta = -eta*rho_1*gamma*gamma / (beta*gamma_1*gamma_1);

      if (it > 1) {
         d = eta*p + (theta_1*theta_1*gamma*gamma) * d;
         s = eta*pp + (theta_1*theta_1*gamma*gamma) * s;
      } else {
         d = eta*p;
         s = eta*pp;
      }
      x += d;
      r -= s;

      if ((res = r.getNorm2() / nrm) <= toler) {
         toler = res;
         return it;
      }
   }

   toler = res;
   return -it;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

} /* namespace OFELI */

#endif
