/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================*/

#ifndef __SCHUR_H
#define __SCHUR_H

#include "OFELI_Config.h"
#include "mesh/Mesh.h"
#include "linear_algebra/Vect.h"

namespace OFELI {


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template <class T_, class MAT_A_, class MAT_U_, class MAT_D_>
void Schur(const Mesh &ms, MAT_A_ &A, MAT_U_ &U, MAT_U_ &L, MAT_D_ &D, Vect<T_> &b, Vect<T_> &y)
{
   size_t i, j, k;
   size_t n=U.getNbRows(), ns=U.getNbColumns();

   int ret = A.Factor();
   Vect<real_t> c(b);
   A.Solve(b);
   L.Mult(b,y);

   Vect<real_t>  v(n), w(ns);
   for (j=1; j<=ns; j++) {
      v = 0;
      for (k=1; k<=ms.getNbNodes(); k++) {
         if (ms.getPtrNode(k)->getCode(1)) {
            v(2*k-1) = U(2*k-1,j);
            v(2*k  ) = U(2*k  ,j);
         }
      }
      A.Solve(v);
      L.Mult(v,w);
      for (i=1; i<=ns; i++)
         D.Add(i,j,-w(i));
   }

   ret = D.Factor();
   D.Solve(y);

   U.Mult(y,v);
   for (i=1; i<=n; i++)
      b(i) = c(i) + v(i);
   A.Solve(b);
   y *= -1;
}

#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template <class T_, class MAT_A_, class MAT_U_, class MAT_D_>
inline void Schur(size_t     nb,
                  Mesh**     ms,
                  MAT_A_**   A,
                  MAT_U_**   U,
                  MAT_U_**   L,
                  MAT_D_*    D,
                  Vect<T_>** b,
                  Vect<T_>*  c,
                  int        fact)
/*-----------------------------------------------------------------------------
 *           Solve a linear arrow block system by block factorization
 *
 *     The system is of the form:
 *
 *  @verbatim
       | A[0]   0    ...  0    U[0] |  | x[0] |     | b[0] |
       |  0    A[1]  ...  0    U[1] |  | x[1] |     | b[1] |
       | ...                    ... |  | ...  |     |  ... |
       |                            |  |      |  =  |      |
       |                            |  |      |     |      |
       |  0    ...    0  A[n]  U[n] |  | x[n] |     | b[n] |
       | L[0]  L[1]  ... L[n]   D   |  |  y   |     |  c   |
 *  @verbatim
 *
 *  - Diagonal blocks <tt>A1,...,An</tt> are instances of template class \b MAT_A_
 *  - Block <tt>D</tt> is instance of template class \b MAT_D_
 *  - Upper rectangular blocks are instances of template class \b MAT_U_
 *  - Lower rectangular blocks are instances of template class \b MAT_L_
 *  - All vector blocks are instances of template class \b VEC_
 *
 * \tparam <T_> data type (real_t, float, ...)
 * \tparam <MAT_A_> Class for block diagonal matrices
 * \tparam <MAT_U_> Class for lower and upper block matrices
 * \tparam <MAT_D_> Class for the last block diagonal matrix (the one used for
 * Schur complement
 *
  -----------------------------------------------------------------------------*/
{
   size_t nr0, nc0, nr, nc;
   Vect<T_> **z;
   z = new Vect<T_> *[nb];

   for (size_t n=0; n<nb; n++) {
      nr0 =  nc0 = 0;
      nr = U[n]->getNbRows();
      nc = U[n]->getNbColumns();

      if (fact)
	 A[n]->Factor();

      z[n] = new Vect<T_>(*(b[n]));
      Vect<T_> y(*(b[n]));
      A[n]->Solve(*(b[n]));
      L[n]->Mult(*(b[n]),y);

      if (fact) {
         for (size_t j=1; j<=nc; j++) {
            Vect<T_> zz(nr), w(nc);
            for (size_t k=1; k<=ms[n]->getNbNodes(); k++) {
               if (ms[n]->getPtrNode(k)->getCode(1)) {
                  zz(2*k-1) = (*U[n])(2*k-1,j);
                  zz(2*k  ) = (*U[n])(2*k  ,j);
               }
            }
            A[n]->Solve(zz);
            L[n]->Mult(zz,w);
            for (size_t i=1; i<=nc; i++)
               D->Add(i+nr0,j+nc0,-w(i));
         }
      }

      *b[n] *= -1;
      L[n]->MultAdd(*b[n],*c);
      nr0 = nr; nc0 = nc;
   }

   if (fact)
      D->Factor();
   D->Solve(*c);

   *c *= -1;
   for (size_t m=0; m<nb; m++) {
      U[m]->MultAdd(*c,*z[m]);
      A[m]->Solve(*z[m]);
      *b[m] = *z[m];
   }
   *c *= -1;
   delete z;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


#ifndef DOXYGEN_SHOULD_SKIP_THIS
template <class T_, class MAT_A_, class MAT_U_>
inline void Schur(int           nb,
                  Mesh**        ms,
                  MAT_A_**      A,
                  MAT_U_**      U,
                  MAT_U_**      L,
                  SpMatrix<T_>* D,
                  Vect<T_>**    b,
                  Vect<T_>*     c,
                  int           fact)
/*-----------------------------------------------------------------------------
 *           Solve a linear arrow block system by block factorization
 *
 *     The system is of the form:
 *
 *  @verbatim
       | A[0]   0    ...  0    U[0] |  | x[0] |     | b[0] |
       |  0    A[1]  ...  0    U[1] |  | x[1] |     | b[1] |
       | ...                    ... |  | ...  |     |  ... |
       |                            |  |      |  =  |      |
       |                            |  |      |     |      |
       |  0    ...    0  A[n]  U[n] |  | x[n] |     | b[n] |
       | L[0]  L[1]  ... L[n]   D   |  |  y   |     |  c   |
    @endverbatim
 *
 *  - Diagonal blocks <tt>A0,...,An</tt> are instances of template class \b MAT_A_
 *  - Block <tt>D</tt> is instance of template class \b MAT_D_
 *  - Upper rectangular blocks are instances of template class \b MAT_U_
 *  - Lower rectangular blocks are instances of template class \b MAT_L_
 *  - All vector blocks are instances of template class \b VEC_
 *
  -----------------------------------------------------------------------------*/
{
   real_t toler = 1.e-12;
   Vect<T_> **z;
   z = new Vect<T_> *[nb];

   for (size_t n=0; n<nb; n++) {
      size_t nr0=0, nc0=0;
      size_t nr = U[n]->getNbRows(),
             nc = U[n]->getNbColumns();

      if (fact)
         A[n]->Factor();

      z[n] = new Vect<T_>(*(b[n]));
      Vect<T_> y(*(b[n]));
      A[n]->Solve(*(b[n]));
      L[n]->Mult(*(b[n]),y);

      if (fact) {
         for (size_t j=1; j<=nc; j++) {
            Vect<T_> zz(nr), w(nc);
            for (size_t k=1; k<=ms[n]->getNbNodes(); k++) {
               if (ms[n]->getPtrNode(k)->getCode(1)) {
                  zz(2*k-1) = (*U[n])(2*k-1,j);
                  zz(2*k  ) = (*U[n])(2*k  ,j);
               }
            }
            A[n]->Solve(zz);
            L[n]->Mult(zz,w);
            for (size_t i=1; i<=nc; i++)
               D->Add(i+nr0,j+nc0,-w(i));
         }
      }

      *b[n] *= -1;
      L[n]->MultAdd(*b[n],*c);
      nr0 = nr; nc0 = nc;
   }

   IdentPrec<real_t,SpMatrix<real_t> > p(*D);
   Vect<T_> sol(c->size());
   int nb_it = BiCG(*D, p, *c, sol, 1000, toler, 1);
   *c = sol;

   *c *= -1;
   for (size_t m=0; m<nb; m++) {
      U[m]->MultAdd(*c,*z[m]);
      A[m]->Solve(*z[m]);
      *b[m] = *z[m];
   }
   *c *= -1;
   delete z;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */


/** \fn void Schur(MAT_A_ &A, MAT_U_ &U, MAT_U_ &L, MAT_D_ &D, Vect<T_> &b, Vect<T_> &c)
 *  \ingroup Solver
 *  \brief Solve a linear system of equations with a 2x2-block matrix.
 *  \details The linear system is of the form
 *
 *  @verbatim
        | A  U | |x|   |b|
        |      | | | = | |
        | L  D | |y|   |c|
    @endverbatim
 *
 * @param[in] A Instance of a matrix class for the first diagonal block. The matrix must
 * be invertible and factorizable (Do not use SpMatrix class)
 * where <tt>A</tt>, <tt>U</tt>, <tt>L</tt>, <tt>D</tt> are instances of matrix classes,
 * @param [in] U Instance of a matrix class for the upper triangle block. The matrix can
 * be rectangular
 * @param [in] L Instance of a matrix class for the lower triangle block. The matrix can
 * be rectangular
 * @param [in] D Instance of a matrix class for the second diagonal block. The matrix must
 * be factorizable (Do not use SpMatrix class)
 * @param [in,out] b Vector (Instance of class Vect) that contains the first block of
 * right-hand side on input and the first block of the solution on output. <tt>b</tt> must
 * have the same size as the dimension of <tt>A</tt>.
 * @param [in,out] c Vector (Instance of class Vect) that contains the second block of
 * right-hand side on output and the first block of the solution on output. <tt>c</tt> must
 * have the same size as the dimension of <tt>D</tt>.
 *
 * Template Arguments:
 *
 * \tparam <T_> data type (real_t, float, ...)
 * \tparam <MAT_A_> %Matrix class for <tt>A</tt>
 * \tparam <MAT_U_> %Matrix class for <tt>D</tt>
 * \tparam <MAT_D_> %Matrix class for <tt>U</tt> and <tt>L</tt>
 *
 */
template <class T_, class MAT_A_, class MAT_D_, class MAT_U_>
inline void Schur(MAT_A_ &A, MAT_U_ &U, MAT_U_ &L, MAT_D_ &D, Vect<T_> &b, Vect<T_> &c)
{
   size_t nr=U.getNbRows(), nc=U.getNbColumns();
   int ret = A.Factor();
   Vect<real_t> v(nr), w(nc), bb(b);
   A.Solve(b);
   L.Mult(b,w);
   c -= w;

   for (size_t j=1; j<=nc; j++) {
      for (size_t k=1; k<=nr; k++)
         v(k) = U(k,j);
      A.Solve(v);
      L.Mult(v,w);
      for (size_t i=1; i<=nc; i++)
         D.Add(i,j,-w(i));
   }

   ret = D.Factor();
   D.Solve(c);

   U.Mult(c,v);
   bb -= v;
   A.Solve(bb);
   b = bb;
}


/** \fn void Schur(Vect<MAT_A_> &A, Vect<MAT_U_> &U, Vect<MAT_U_> &L, MAT_D_ &D, const Vect<Vect<T_> > &b,
           const Vect<T_> &c, Vect<Vect<T_> > &x, Vect<T_> &y)
 *  \ingroup Solver
 *  \brief Solve a linear arrow block system by block factorization
 *
 *  \details The linear system is of the form
 *  @verbatim
       | A[0]   0    ...    0      U[0]   |  |  x[0]   |   |  b[0]   |
       |  0    A[1]  ...    0      U[1]   |  |  x[1]   |   |  b[1]   |
       | ...                       ...    |  |  ...    |   |   ...   |
       |                                  |  |         | = |         |
       |                                  |  |         |   |         |
       |  0    ...    0  A[nb-1]  U[nb-1] |  | x[nb-1] |   | b[nb-1] |
       | L[0]  L[1]  ... L[nb-1]     D    |  |  y      |   |   c     |
    @endverbatim
 *
 *  <ul>
 *    <li> Diagonal blocks <tt>A[0],...,A[nb-1]</tt> are instances of template class \b MAT_A_
 *    <li> Block <tt>D</tt> is a dense matrix (instance of template class \b MAT_D_)
 *    <li> Upper rectangular blocks are instances of template class \b MAT_U_
 *    <li> Lower rectangular blocks are instances of template class \b MAT_U_
 *  </ul>
 *  All vector blocks are instances of template class Vect<T_>
 *
 * @param [in] A Vect instance whose coefficients are diagonal blocks of the matrix 
 * (<tt>A[0], ...,A[nb-1]</tt>). Each block is an instance of class \b MAT_A_
 * @param [in] U Vect instance whose coefficients are upper last blocks of the matrix
 * (<tt>U[0], ..., U[nb-1]</tt>). Each block is an instance of class \b MAT_U_
 * @param [in] L Vect instance whose coefficients are lower last blocks of the matrix
 * (<tt>L[0], ...,L[nb-1]</tt>). Each block is an instance of class \b MAT_U_
 * @param [in] D Instance of class \b MAT_D_ (A matrix class) that contains last diagonal block of the matrix
 * @param [in] b Vect instance whose entries are right-hand side blocks (as instances of class Vect)
 * @param [in] c Vect instance that contains the last right-hand side block.
 * @param [out] x Vect instance whose entries are solution blocks (as instances of class Vect)
 * @param [out] y Vect instance that contains the last solution block.
 *
 * \tparam <T_> Data type (real_t, float, complex<real_t>, ...)
 * \tparam <MAT_A_> %Matrix class for <tt>A</tt>
 * \tparam <MAT_D_> %Matrix class for <tt>D</tt>
 * \tparam <MAT_U_> %Matrix class for <tt>U</tt> and <tt>L</tt> 
 *
 */
template<class T_, class MAT_A_, class MAT_D_, class MAT_U_>
void Schur(      Vect<MAT_A_>&    A,
                 Vect<MAT_U_>&    U,
                 Vect<MAT_U_>&    L,
                 MAT_D_&          D, 
           const Vect<Vect<T_> >& b,
           const Vect<T_>&        c,
                 Vect<Vect<T_> >& x,
                 Vect<T_>&        y)
{
   size_t n, i, j, k, nb=A.size();
   Vect<T_> cc(c);

// Loop on blocks
   for (n=0; n<nb; n++) {
      size_t nr=U[n].getNbRows(), nc=U[n].getNbColumns();

//    Factorize diagonal block
      A[n].Factor();

//    Calculate modified RHS
      Vect<T_> z(nc), zz(nr);
      A[n].Solve(b[n],zz);
      L[n].Mult(zz,z);
      cc -= z;

//    Calculate Schur complement matrix
      Vect<T_> t(nr), w(nc);
      for (j=1; j<=nc; j++) {
         for (k=1; k<=nr; k++)
            t(k) = U[n](k,j);
         A[n].Solve(t);
         L[n].Mult(t,w);
         for (i=1; i<=nc; i++)
            D(i,j) -= w(i);
      }
   }

// Solve the Schur system
   D.Factor();
   D.Solve(cc,y);

// Solve remaining diagonal block systems
   for (n=0; n<nb; n++) {
      Vect<T_> bb(b[n]), z(U[n].getNbRows());
      U[n].Mult(y,z);
      z -= b[n];
      A[n].Solve(z,x[n]);
      x[n] *= -1;
   }
}

} /* namespace OFELI */

#endif
