/*==============================================================================

                                    O  F  E  L  I

                            Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                      Class Laplace2DT3 : Laplace Equation
              using 3-Node Triangular finite element in two dimensions

  ==============================================================================*/


#include "equations/laplace/Laplace2DT3.h"

namespace OFELI {

Laplace2DT3::Laplace2DT3(Mesh&             ms,
                         SpMatrix<real_t>& A,
                         Vect<real_t>&     b)
            : Equation<real_t,3,3,2,2>(ms)
{
   _u = &b;
   _K.set(IDENTITY);
}


Laplace2DT3::Laplace2DT3(Mesh&         ms,
                         Vect<real_t>& b)
            : Equation<real_t,3,3,2,2>(ms)
{
   _u = &b;
}


Laplace2DT3::Laplace2DT3(Mesh& ms)
            : Equation<real_t,3,3,2,2>(ms)
{
   _K.set(IDENTITY);
}


Laplace2DT3::Laplace2DT3(Element* el)
{
   set(el);
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class Laplace2DT3 is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


Laplace2DT3::Laplace2DT3(Side* sd)
{
   set(sd);
#ifdef _OFELI_DEBUG
   std::clog << "An instance of class Laplace2DT3 is constructed.\n";
   std::clog << "File: " << __FILE__ << ", Line: " << __LINE__ << endl;
#endif
}


void Laplace2DT3::set(const Element* el)
{
   _nb_dof = 1;
   Init(el);
   Triang3 tr(_theElement);
   _area = tr.getArea();
   _dSh(1) = tr.DSh(1); _dSh(2) = tr.DSh(2); _dSh(3) = tr.DSh(3);
   eMat = 0;
   eRHS = 0;
}


void Laplace2DT3::set(const Side* sd)
{
   _nb_dof = 1;
   Init(sd);
   Line2 ln(_theSide);
   _length = ln.getLength();
   sMat = 0;
   sRHS = 0;
}


void Laplace2DT3::setDiffusivity(const LocalMatrix<real_t,2,2>& K)
{
   _K = K;
}


void Laplace2DT3::build()
{
   MESH_EL {
      set(theElement);
      LHS();
      BodyRHS(*_bf);
      _A->Assembly(theElement,eMat.get());
      ElementAssembly(*_u);
   }
   MESH_SD {
      set(theSide);
      BoundaryRHS(*_sf);
      SideAssembly(*_u);
   }
}


void Laplace2DT3::buildEigen(SkSMatrix<real_t>& K,
                             SkSMatrix<real_t>& M)
{
   MESH_EL {
      set(theElement);
      for (size_t i=1; i<=3; i++)
         for (size_t j=1; j<=3; j++)
            eMat(i,j) += _area*_dSh(i)*_dSh(j);
      ElementAssembly(K);
   }
   MESH_EL {
      set(theElement);
      real_t c = 0.5*OFELI_SIXTH*_area;
      eMat(1,1) += 2*c; eMat(2,2) += 2*c; eMat(3,3) += 2*c;
      eMat(1,2) +=   c; eMat(2,1) +=   c; eMat(1,3) +=   c;
      eMat(3,1) +=   c; eMat(2,3) +=   c; eMat(3,2) +=   c;
      ElementAssembly(M);
   }
}


void Laplace2DT3::buildEigen(SkSMatrix<real_t>& K,
                             Vect<real_t>&      M)
{
   MESH_EL {
      set(theElement);
      for (size_t i=1; i<=3; i++)
         for (size_t j=1; j<=3; j++)
            eMat(i,j) += _area*_dSh(i)*_dSh(j);
      eRHS(1) += OFELI_THIRD*_area;
      eRHS(2) += OFELI_THIRD*_area;
      eRHS(3) += OFELI_THIRD*_area;
      ElementAssembly(K);
      ElementAssembly(M);
   }
}


int Laplace2DT3::solve(Vect<real_t>& u)
{
   real_t toler = 1.e-8;
   Vect<real_t> x(_A->size());
   int nb_it = CG(*_A,Prec<real_t>(_A,ILU_PREC),*_u,x,1000,toler,1);
   u.insertBC(*_theMesh,x,*_bc);
   return nb_it;
}


void Laplace2DT3::LHS(real_t coef)
{
   for (size_t i=1; i<=3; i++)
      for (size_t j=1; j<=3; j++)
         eMat(i,j) += coef*_area*_K(i,j)*_dSh(i)*_dSh(j);
}


void Laplace2DT3::BodyRHS(const Vect<real_t>& f)
{
   eRHS(1) += f((*_theElement)(1)->n())*_area*OFELI_THIRD;
   eRHS(2) += f((*_theElement)(2)->n())*_area*OFELI_THIRD;
   eRHS(3) += f((*_theElement)(3)->n())*_area*OFELI_THIRD;
}


void Laplace2DT3::BoundaryRHS(const Vect<real_t>& h)
{
   if (_theSide->getCode(1)>0) {
      sRHS(1) += 0.5*_length*h((*_theSide)(1)->n());
      sRHS(2) += 0.5*_length*h((*_theSide)(2)->n());
   }
}


void Laplace2DT3::Axb(const Vect<real_t>& x,
                            Vect<real_t>& b)
{
   MESH_EL {
      set(theElement);
      LHS();
      BodyRHS(*_bf);
      ElementAssembly(*_u);
      AxbAssembly(TheElement,x,b);
   }
   MESH_SD {
      set(theSide);
      BoundaryRHS(*_sf);
      SideAssembly(*_u);
      AxbAssembly(TheElement,x,b);
   }
}

} /* namespace OFELI */
