/*==============================================================================

                                    O  F  E  L  I

                           Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                   Various functions to determine matrix graph

  ==============================================================================*/


#include "linear_algebra/GraphOfMatrix.h"
#include "linear_algebra/Vect.h"

namespace OFELI {

size_t SimpleSkyline(const Mesh&           m,
                           vector<size_t>& ch)
/*------------------------------------------------------------------------------
                      Set simple skyline structure of matrix
                              (for node based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t n;
   int l;
   size_t size = m.getNbNodes();
   ch.resize(size,0);

   mesh_elements(m) {
      int mini = size;
      for (n=1; n<=The_element.getNbNodes(); n++)
         mini = min(int(The_element(n)->n()),mini);
      for (n=1; n<=The_element.getNbNodes(); n++)
         if ((l=The_element(n)->n()) != 0)
            ch[l-1] = max(l-mini+1,int(ch[l-1]));
   }
   size_t length = 0;
   for (size_t i=0; i<size; i++)
      length += ch[i];
   return length;
}


size_t NodeSkyline(const Mesh&           m,
                         vector<size_t>& ch)
/*------------------------------------------------------------------------------
                           Set skyline structure of matrix
                              (for node based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t  i, k, l, n;
   int label_eq;
   size_t size = m.getNbEq();
   ch.resize(size,0);

   mesh_elements(m) {
      int mini = size;
      for (n=1; n<=The_element.getNbNodes(); n++) {
         the_node = The_element(n);
         for (k=1; k<=The_node.getNbDOF(); k++)
            if (The_node.getDOF(k) != 0)
               mini = min(int(The_node.getDOF(k)),mini);
      }
      for (n=1; n<=The_element.getNbNodes(); n++) {
         Node *nd = The_element(n);
         for (l=1; l<=nd->getNbDOF(); l++)
            if ((label_eq=nd->getDOF(l)) != 0)
               ch[label_eq-1] = max(label_eq-mini+1,int(ch[label_eq-1]));
      }
   }
   size_t length = 0;
   for (i=0; i<size; i++)
      length += ch[i];
   return length;
}


size_t NodeSkyline(const Mesh&           m,
                         vector<size_t>& ch,
                         size_t          dof1,
                         size_t          dof2)
/*------------------------------------------------------------------------------
                          Set skyline structure of matrix     
                    (for node based unknowns and selected DOF)
  ------------------------------------------------------------------------------*/
{
   size_t  k, n;
   int i;
   size_t size = m.getNbNodes()*(dof2-dof1+1);
   ch.resize(size,0);

   mesh_elements(m) {
      int mini = size;
      for (n=1; n<=The_element.getNbNodes(); n++) {
          the_node = The_element(n);
         for (k=dof1; k<=dof2; k++)
            if (The_node.getDOF(k) != 0)
               mini = min(int(The_node.getDOF(k)),mini);
      }
      for (n=1; n<=The_element.getNbNodes(); n++) {
         the_node = The_element(n);
         for (k=dof1; k<=dof2; k++)
            if ((i=The_node.getDOF(k)) != 0)
               ch[i-1] = max(i-mini+1,int(ch[i-1]));
      }
   }
   size_t length = 0;
   for (k=0; k<size; k++)
      length += ch[k];
   return length;
}


size_t NodeSkyline(const Mesh&           m,
                         vector<size_t>& ch,
                         size_t          dof)
/*------------------------------------------------------------------------------
                           Set skyline structure of matrix
                              (for node based unknowns)
  ------------------------------------------------------------------------------*/
{
   int    i;
   size_t size = m.getNbEq();
   ch.resize(size,0);
   mesh_elements(m) {
      int mini = size;
      for (size_t n=1; n<=The_element.getNbNodes(); n++) {
         the_node = The_element(n);
         if (The_node.getDOF(dof) != 0)
            mini = min(int(node_label),mini);
      }
      for (size_t n=1; n<=The_element.getNbNodes(); n++)
         if ((i=The_element(n)->n()) != 0)
             ch[i-1] = max(i-mini+1,int(ch[i-1]));
   }
   size_t length = 0;
   for (size_t k=0; k<size; k++)
      length += ch[k];
   return length;
}


size_t SideSkyline(const Mesh&           m,
                         vector<size_t>& ch,
                         size_t          dof1,
                         size_t          dof2)
/*------------------------------------------------------------------------------
                           Set skyline structure of matrix
                              (for side based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t  i, k, l, n;
   int   label_eq;
   size_t size = m.getNbEq();
   ch.resize(size,0);

   mesh_elements(m) {
      int mini = size;
      for (n=1; n<=The_element.getNbSides(); n++) {
         the_side = The_element.getPtrSide(n);
         for (k=dof1; k<=dof2; k++)
            if (The_side.getDOF(k) != 0)
               mini = min(int(The_side.getDOF(k)),mini);
      }
      for (n=1; n<=The_element.getNbSides(); n++) {
         the_side = The_element.getPtrSide(n);
         for (l=dof1; l<=dof2; l++)
            if ((label_eq=The_side.getDOF(l)) != 0)
               ch[label_eq-1] = max(int(ch[label_eq-1]),label_eq-mini+1);
      }
   }
   size_t length = 0;
   for (i=0; i<size; i++)
      length += ch[i];
   return length;
}


size_t SideSkyline(const Mesh&           m,
                         vector<size_t>& ch)
/*------------------------------------------------------------------------------
                           Set skyline structure of matrix
                              (for side based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t  i, k, l, n;
   int   label_eq;
   size_t size = m.getNbEq();
   ch.resize(size,0);

   mesh_elements(m) {
      int mini = size;
      for (n=1; n<=The_element.getNbSides(); n++) {
         the_side = The_element.getPtrSide(n);
         for (k=1; k<=The_side.getNbDOF(); k++)
            if (The_side.getDOF(k) != 0)
               mini = min(int(The_side.getDOF(k)),mini);
      }
      for (n=1; n<=The_element.getNbSides(); n++) {
         the_side = The_element.getPtrSide(n);
         for (l=1; l<=The_side.getNbDOF(); l++)
            if ((label_eq=The_side.getDOF(l)) != 0)
               ch[label_eq-1] = max(int(ch[label_eq-1]),label_eq-mini+1);
      }
   }
   size_t length = 0;
   for (i=0; i<size; i++)
      length += ch[i];
   return length;
}


size_t SideSkyline(const Mesh&           m,
                         vector<size_t>& ch,
                         size_t          dof)
/*------------------------------------------------------------------------------
                           Set skyline structure of matrix
                              (for side based unknowns)
  ------------------------------------------------------------------------------*/
{
   dof = 0;
   size_t i, n;
   int label_eq;
   size_t size = m.getNbEq();
   ch.resize(size,0);

   mesh_elements(m) {
      int mini = size;
      for (n=1; n<=The_element.getNbSides(); n++)
         mini = min(int(The_element.getSideLabel(n)),mini);
      for (n=1; n<=The_element.getNbSides(); n++) {
         the_side = The_element.getPtrSide(n);
         label_eq = side_label;
         ch[label_eq-1] = max(int(ch[label_eq-1]),label_eq-mini+1);
      }
   }
   size_t length = 0;
   for (i=0; i<size; i++)
      length += ch[i];
   return length;
}


size_t ElementSkyline(const Mesh&           m,
                            vector<size_t>& ch)
/*------------------------------------------------------------------------------
                           Set skyline structure of matrix
                             (for element based unknowns)
  ------------------------------------------------------------------------------*/
{
   const Element *el1, *el2;
   size_t size = m.getNbElements();
   int i, j;
   ch.resize(size,0);

   mesh_sides(m) {
      int mini = size;
      el1 = The_side.getNeighborElement(1);
      el2 = The_side.getNeighborElement(2);
      if (el1 && el2) {
         i = el1->n(); j = el2->n();
         mini = min(mini,min(i,j));
         ch[i-1] = max(int(ch[i-1]),i+1-mini);
         ch[j-1] = max(int(ch[j-1]),j+1-mini);
      }
   }
   size_t length = 0;
   for (size_t i=0; i<size; i++)
      length += ch[i];
   return length;
}


size_t ElementSkyline(const Mesh&           m,
                            vector<size_t>& ch,
                            size_t          dof)
/*------------------------------------------------------------------------------
                           Set skyline structure of matrix
                            (for element based unknowns)
  ------------------------------------------------------------------------------*/
{
   dof = 0;
   const Element *el1, *el2;
   size_t size = m.getNbEq();
   ch.resize(size,0);

   mesh_sides(m) {
      int mini = size;
      if ((el1=The_side.getNeighborElement(1)) &&
          (el2=The_side.getNeighborElement(2))) {
         mini = min(mini,min(int(el1->n()),int(el2->n())));
         int label_eq = el1->n();
         ch[label_eq-1] = max(int(ch[label_eq-1]),label_eq-mini+1);
         label_eq = el2->n();
         ch[label_eq-1] = max(int(ch[label_eq-1]),label_eq-mini+1);
      }
   }
   size_t length = 0;
   for (size_t i=0; i<size; i++)
      length += ch[i];
   return length;
}


size_t SimpleGraph(const Mesh&           m,
                         vector<long>&   xadj,
                         vector<size_t>& adjncy)
/*------------------------------------------------------------------------------
                            Set a simple graph of matrix
                             (for node based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t SimpleSkyline(const Mesh &m, vector<size_t> &ch);
   size_t k, l, length=0, size=m.getNbNodes();
   vector<size_t> ch(size,0);
   SimpleSkyline(m,ch);

   size_t ip = 0;
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbNodes(); in++) {
         if ((k=The_element(in)->n())!=0)
            for (size_t jn=1; jn<=The_element.getNbNodes(); jn++) {
               if ((l=The_element(jn)->n())!=0) {
                  ip++;
                  if (k!=l)
                     ip++;
               }
            }
      }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbNodes(); in++) {
         if ((k=The_element(in)->n())!=0)
            for (size_t jn=1; jn<=The_element.getNbNodes(); jn++) {
               if ((l=The_element(jn)->n())!=0) {
                  *(it++) = RC(k-1,l-1);
                  if (k!=l)
                     *(it++) = RC(l-1,k-1);
               }
            }
      }
   }
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   length = pp.size();
   xadj.resize(size+1);
   adjncy.resize(length);
   for (size_t i=0; i<size; i++)
      xadj[i] = 0;
   k = 0;
   for (size_t j=0; j<length; j++) {
      if (pp[j].first==pp[j].second)
         k++;
      xadj[k-1]++;
      adjncy[j] = pp[j].second + 1;
   }
   for (k=1; k<size; k++)
      xadj[k] += xadj[k-1];
   for (int ii=size; ii>0; ii--)
      xadj[ii] = xadj[ii-1] + 1;
   xadj[0] = 1;
   return length;
}


size_t NodeGraph(const Mesh&           m,
                       vector<size_t>& row_ptr,
                       vector<size_t>& col_ind)
/*------------------------------------------------------------------------------
                                Set graph of matrix
                             (for node based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t   l=0;
   size_t size = m.getNbEq();
   vector<size_t> ch(size);
   NodeSkyline(m,ch);

   size_t ip = 0;
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbNodes(); in++) {
         Node *nd1 = The_element(in);
         for (size_t k=1; k<=nd1->getNbDOF(); k++) {
            if (nd1->getDOF(k)!=0) {
               for (size_t jn=1; jn<=The_element.getNbNodes(); jn++) {
                  Node *nd2 = The_element(jn);
                  for (l=1; l<=nd2->getNbDOF(); l++) {
                     if (nd2->getDOF(l)!=0) {
                        ip++;
                        if (nd1->getDOF(k)!=nd2->getDOF(l))
                           ip++;
                     }
                  }
               }
            }
         }
      }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbNodes(); in++) {
         Node *nd1 = The_element(in);
         for (size_t k=1; k<=nd1->getNbDOF(); k++) {
            if (nd1->getDOF(k)!=0) {
               for (size_t jn=1; jn<=The_element.getNbNodes(); jn++) {
                  Node *nd2 = The_element(jn);
                  for (l=1; l<=nd2->getNbDOF(); l++) {
                     if (nd2->getDOF(l)!=0) {
                        *(it++) = RC(nd1->getDOF(k)-1,nd2->getDOF(l)-1);
                        if (nd1->getDOF(k)!=nd2->getDOF(l))
                           *(it++) = RC(nd2->getDOF(l)-1,nd1->getDOF(k)-1);
                     }
                  }
               }
            }
         }
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(size,pp,row_ptr,col_ind);
   return pp.size();
}


size_t NodeGraph(const Mesh&           m,
                       size_t          dof1,
                       size_t          dof2,
                       vector<size_t>& row_ptr,
                       vector<size_t>& col_ind)
/*------------------------------------------------------------------------------
                                Set graph of matrix
                             (for node based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t l=0, size=m.getNbEq();
   vector<size_t> ch(size);
   NodeSkyline(m,ch);

   size_t ip = 0;
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbNodes(); in++) {
         Node *nd1 = The_element(in);
         for (size_t k=dof1; k<=dof2; k++)
            if (nd1->getDOF(k)!=0)
               for (size_t jn=1; jn<=The_element.getNbNodes(); jn++) {
                  Node *nd2 = The_element(jn);
                  for (l=dof1; l<=dof2; l++) {
                     if (nd2->getDOF(l)!=0) {
                        ip++;
                        if (nd1->getDOF(k)!=nd2->getDOF(l))
                           ip++;
                     }
                  }
               }
      }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbNodes(); in++) {
         Node *nd1 = The_element(in);
         for (size_t k=dof1; k<=dof2; k++)
            if (nd1->getDOF(k)!=0)
               for (size_t jn=1; jn<=The_element.getNbNodes(); jn++) {
                  Node *nd2 = The_element(jn);
                  for (l=dof1; l<=dof2; l++)
                     if (nd2->getDOF(l)!=0) {
                        *(it++) = RC(nd1->getDOF(k)-1,nd2->getDOF(l)-1);
                        if (nd1->getDOF(k)!=nd2->getDOF(l))
                           *(it++) = RC(nd2->getDOF(l)-1,nd1->getDOF(k)-1);
                     }
               }
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(size,pp,row_ptr,col_ind);
   return pp.size();
}


size_t SideGraph(const Mesh&           m,
                       vector<size_t>& row_ptr,
                       vector<size_t>& col_ind)
/*------------------------------------------------------------------------------
                                 Set graph of matrix
                              (for side based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t l=0;
   size_t size = m.getNbEq();
   vector<size_t> ch(size);
   SideSkyline(m,ch);
   size_t ip = 0;
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbSides(); in++) {
         Side *sd1 = The_element.getPtrSide(in);
         for (size_t k=1; k<=sd1->getNbDOF(); k++) {
            if (sd1->getDOF(k)!=0) {
               for (size_t jn=1; jn<=The_element.getNbSides(); jn++) {
                  Side *sd2 = The_element.getPtrSide(jn);
                  for (l=1; l<=sd2->getNbDOF(); l++) {
                     if (sd2->getDOF(l)!=0) {
                        ip++;
                        if (sd1->getDOF(k)!=sd2->getDOF(l))
                           ip++;
                     }
                  }
               }
            }
         }
      }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbSides(); in++) {
         Side *sd1 = The_element.getPtrSide(in);
         for (size_t k=1; k<=sd1->getNbDOF(); k++) {
            if (sd1->getDOF(k)!=0)
               for (size_t jn=1; jn<=The_element.getNbSides(); jn++) {
                  Side *sd2 = The_element.getPtrSide(jn);
                  for (l=1; l<=sd2->getNbDOF(); l++) {
                     if (sd2->getDOF(l)!=0) {
                        *(it++) = RC(sd1->getDOF(k)-1,sd2->getDOF(l)-1);
                        if (sd1->getDOF(k)!=sd2->getDOF(l))
                           *(it++) = RC(sd2->getDOF(l)-1,sd1->getDOF(k)-1);
                     }
                  }
               }
         }
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(size,pp,row_ptr,col_ind);
   return pp.size();
}


size_t SideGraph(const Mesh&           m,
                       size_t          dof1,
                       size_t          dof2,
                       vector<size_t>& row_ptr,
                       vector<size_t>& col_ind)
/*------------------------------------------------------------------------------
                                 Set graph of matrix
                              (for side based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t l=0;
   size_t size = m.getNbEq();
   vector<size_t> ch(size);
   SideSkyline(m,ch);
   size_t ip = 0;
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbSides(); in++) {
         Side *sd1 = The_element.getPtrSide(in);
         for (size_t k=dof1; k<=dof2; k++) {
            if (sd1->getDOF(k)!=0) {
               for (size_t jn=1; jn<=The_element.getNbSides(); jn++) {
                  Side *sd2 = The_element.getPtrSide(jn);
                  for (l=dof1; l<=dof2; l++) {
                     if (sd2->getDOF(l)!=0) {
                        ip++;
                        if (sd1->getDOF(k)!=sd2->getDOF(l))
                           ip++;
                     }
                  }
               }
            }
         }
      }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_elements(m) {
      for (size_t in=1; in<=The_element.getNbSides(); in++) {
         Side *sd1 = The_element.getPtrSide(in);
         for (size_t k=dof1; k<=dof2; k++) {
            if (sd1->getDOF(k)!=0) {
               for (size_t jn=1; jn<=The_element.getNbSides(); jn++) {
                  Side *sd2 = The_element.getPtrSide(jn);
                  for (l=dof1; l<=dof2; l++) {
                     if (sd2->getDOF(l)!=0) {
                        *(it++) = RC(sd1->getDOF(k)-1,sd2->getDOF(l)-1);
                        if (sd1->getDOF(k)!=sd2->getDOF(l))
                           *(it++) = RC(sd2->getDOF(l)-1,sd1->getDOF(k)-1);
                     }
                  }
               }
            }
         }
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(size,pp,row_ptr,col_ind);
   return pp.size();
}


size_t ElementGraph(const Mesh&           m,
                          vector<size_t>& row_ptr,
                          vector<size_t>& col_ind)
/*------------------------------------------------------------------------------
                                 Set graph of matrix
                              (for element based unknowns)
  ------------------------------------------------------------------------------*/
{
   size_t size = m.getNbEq();
   vector<size_t> ch(size);
   ElementSkyline(m,ch);
   size_t ip = 0;

   Element *el1, *el2;
   mesh_sides(m)
      if ((el1=The_side.getNeighborElement(1)) && (el2=The_side.getNeighborElement(2)))
         ip += 4;

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_sides(m) {
      if ((el1=The_side.getNeighborElement(1)) && (el2=The_side.getNeighborElement(2))) {
         *(it++) = RC(el1->n()-1,el1->n()-1);
         *(it++) = RC(el2->n()-1,el2->n()-1);
         *(it++) = RC(el1->n()-1,el2->n()-1);
         *(it++) = RC(el2->n()-1,el1->n()-1);
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(size,pp,row_ptr,col_ind);
   return pp.size();
}


size_t SideNodeGraph(const Mesh&           m,
                           vector<size_t>& row_ptr,
                           vector<size_t>& col_ind)
/*------------------------------------------------------------------------------
                                Set graph of matrix
                             (for side/node coupling)
  ------------------------------------------------------------------------------*/
{
   size_t k=0;
   size_t size = m.getNbEq();
   vector<size_t> ch(size);
   NodeSkyline(m,ch);

   size_t ip = 0;
   mesh_sides(m) {
      for (k=1; k<=The_side.getNbDOF(); k++) {
         if (The_side.getCode(k)!=0) {
            for (size_t in=1; in<=The_side.getNbNodes(); in++) {
               the_node = The_side(in);
               for (size_t l=1; l<=The_node.getNbDOF(); l++)
                  if (The_node.getDOF(l)!=0)
                     ip++;
            }
         }
      }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_sides(m) {
      for (k=1; k<=The_side.getNbDOF(); k++) {
         if (The_side.getCode(k)!=0) {
            for (size_t in=1; in<=The_side.getNbNodes(); in++) {
               the_node = The_side(in);
               for (size_t l=1; l<=The_node.getNbDOF(); l++)
                  if (The_node.getDOF(l)!=0)
                     *(it++) = RC(The_node.getDOF(k)-1,The_node.getDOF(l)-1);
            }
         }
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(size,pp,row_ptr,col_ind);
   return pp.size();
}


size_t NodeSideGraph(const Mesh&           m,
                           vector<size_t>& row_ptr, 
                           vector<size_t>& col_ind)
/*------------------------------------------------------------------------------
                                Set graph of matrix
                             (for side/node coupling)
  ------------------------------------------------------------------------------*/
{
   size_t k=0, size=m.getNbEq();
   vector<size_t> ch(size);
   NodeSkyline(m,ch);

   size_t ip = 0;
   mesh_sides(m) {
      for (k=1; k<=The_side.getNbDOF(); k++) {
         if (The_side.getCode(k)!=0) {
            for (size_t in=1; in<=The_side.getNbNodes(); in++) {
               the_node = The_side(in);
               for (size_t l=1; l<=The_node.getNbDOF(); l++)
                  if (The_node.getDOF(l)!=0)
                     ip++;
            }
         }
      }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_sides(m) {
      for (k=1; k<=The_side.getNbDOF(); k++) {
         if (theSide->getCode(k)!=0) {
            for (size_t in=1; in<=The_side.getNbNodes(); in++) {
               the_node = The_side(in);
               for (size_t l=1; l<=The_node.getNbDOF(); l++)
                  if (The_node.getDOF(l)!=0)
                     *(it++) = RC(The_node.getDOF(l)-1,The_node.getDOF(k)-1);
            }
         }
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(size,pp,row_ptr,col_ind);
   return pp.size();
}


size_t NodeGraphScal(const Mesh&           m,
                           vector<size_t>& row_ptr,
                           vector<size_t>& col_ind)
{
   size_t k, i, in, jn, size = m.getNbNodes();
   vector<size_t> ch(size,0);

   mesh_elements(m) {
      int mini = size;
      for (i=1; i<=The_element.getNbNodes(); i++)
         mini = min(int(The_element(i)->n()),mini);
      for (i=1; i<=The_element.getNbNodes(); i++) {
         size_t label_eq = The_element.getNodeLabel(i);
         k = label_eq - mini + 1;
         if (k > ch[label_eq-1])
            ch[label_eq-1] = k;
         label_eq++;
      }
   }

   int ip = 0;
   mesh_elements(m) {
      for (in=1; in<=The_element.getNbNodes(); in++) {
         for (jn=1; jn<=The_element.getNbNodes(); jn++) {
            ip++;
            if (The_element(in)->n() != The_element(jn)->n())
               ip++;
         }
      }
   }

   size_t ls = 0;
   for (i=0; i<size; i++)
      ls += ch[i];
   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_elements(m) {
      for (in=1; in<=The_element.getNbNodes(); in++) {
         for (jn=1; jn<=The_element.getNbNodes(); jn++) {
            *(it++) = RC(The_element(in)->n()-1,The_element(jn)->n()-1);
            if (The_element(in)->n() != The_element(jn)->n())
               *(it++) = RC(The_element(jn)->n()-1,The_element(in)->n()-1);
         }
      }
   }
   pp.erase(it,pp.end());
   sort(pp.begin(),it);
   std::vector<RC>::iterator new_end = std::unique(pp.begin(), pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(size,pp,row_ptr,col_ind);
   return pp.size();
}


size_t NodeGraphScal(const Mesh&           m,
                           size_t          dof,
                           size_t          nb_eq,
                           vector<size_t>& row_ptr,
                           vector<size_t>& col_ind)
{
   size_t k, i, n, in, jn;
   vector<size_t> ch(nb_eq,0);
   mesh_elements(m) {
      int mini = nb_eq;
      for (n=1; n<=The_element.getNbNodes(); n++) {
         Node *nd = The_element(n);
         if (nd->getDOF(dof))
            mini = min(int(nd->getDOF(dof)),mini);
      }

      for (n=1; n<=The_element.getNbNodes(); n++) {
         Node *nd = The_element(n);
         size_t leq = nd->getDOF(dof);
         if (leq) {
            k = leq - mini + 1;
            if (k > ch[leq-1])
               ch[leq-1] = k;
         }
      }
   }

   size_t ls = 0;
   for (i=0; i<nb_eq; i++)
      ls += ch[i];

   size_t ip = 0;
   mesh_elements(m) {
      for (in=1; in<=The_element.getNbNodes(); in++)
         if (The_element(in)->getDOF(dof))
           for (jn=1; jn<=The_element.getNbNodes(); jn++)
              if (The_element(jn)->getDOF(dof)) {
                 ip++;
                 if (The_element(in)->getDOF(dof) != The_element(jn)->getDOF(dof))
                    ip++;
              }
   }

   std::vector<RC> pp(ip);
   std::vector<RC>::iterator it = pp.begin();
   mesh_elements(m) {
      for (in=1; in<=The_element.getNbNodes(); in++) {
         size_t ii = The_element(in)->getDOF(dof);
         if (ii)
            for (jn=1; jn<=The_element.getNbNodes(); jn++) {
                size_t jj = The_element(jn)->getDOF(dof);
                if (jj) {
                   *(it++) = RC(ii-1,jj-1);
                   if (ii != jj)
                      *(it++) = RC(jj-1,ii-1);
                }
            }
      }
   }
   pp.resize(it-pp.begin());
   sort(pp.begin(),pp.end());
   std::vector<RC>::iterator new_end = std::unique(pp.begin(),pp.end());
   pp.erase(new_end, pp.end());
   StoreGraph(nb_eq,pp,row_ptr,col_ind);
   return pp.size();
}


size_t XGraph(const Mesh&           m,
                    vector<size_t>& row_ptr,
                    vector<size_t>& col_ind)
{
   size_t k, p, i, j, in, jn, size = m.getNbNodes();
   vector<size_t> ch(size,0);
   mesh_elements(m) {
      int mini = size;
      for (i=1; i<=The_element.getNbNodes(); i++)
         mini = min(int(The_element(i)->n()),mini);
      for (i=1; i<=The_element.getNbNodes(); i++) {
         size_t label_eq = The_element(i)->n();
         k = label_eq - mini + 1;
         if (k > ch[label_eq-1])
            ch[label_eq-1] = k;
         label_eq++;
      }
   }
   size_t ls = 0;
   for (i=0; i<size; i++)
      ls += ch[i];

// Estimate array length
   size_t ip = 0;
   mesh_elements(m) {
      for (in=1; in<=The_element.getNbNodes(); in++)
         for (jn=1; jn<=The_element.getNbNodes(); jn++) {
            ip++;
            if (The_element(in)->n() != The_element(jn)->n())
               ip++;
         }
   }

   std::vector<RC> PP(ip);
   std::vector<RC>::iterator it = PP.begin();
   mesh_elements(m) {
      for (in=1; in<=The_element.getNbNodes(); in++)
         for (jn=1; jn<=The_element.getNbNodes(); jn++) {
            *(it++) = RC(The_element(in)->n()-1,The_element(jn)->n()-1);
            if (The_element(in)->n() != The_element(jn)->n())
               *(it++) = RC(The_element(jn)->n()-1,The_element(in)->n()-1);
         }
   }
   PP.resize(it-PP.begin());
   sort(PP.begin(),PP.end());
   std::vector<RC>::iterator new_end = std::unique(PP.begin(), PP.end());
   PP.erase(new_end, PP.end());
   StoreGraph(size,PP,row_ptr,col_ind);
   PP.clear();

// Estimate allocation amount
   ip = 0;
   for (i=0; i<size; i++) {
      ip++;
      for (j=row_ptr[i]-1; j<row_ptr[i+1]-1; j++) {
         if ((k=col_ind[j]) != i+1) {
            ip++;
           if (k-1 != i)
              ip++;
           for (size_t m=row_ptr[k-1]-1; m<row_ptr[k]-1; m++)
              if ((p=col_ind[m]) != i+1) {
                 ip++;
                 if (p-1 != i)
                    ip++;
              }
         }
      }
   }

// Extended graph
   std::vector<RC> p2(ip);
   std::vector<RC>::iterator itt = p2.begin();
   for (i=0; i<size; i++) {
      *(itt++) = RC(i,i);
      for (j=row_ptr[i]-1; j<row_ptr[i+1]-1; j++) {
         if ((k=col_ind[j]) != i+1) {
            *(itt++) = RC(k-1,i);
            if (k-1 != i)
               *(itt++) = RC(i,k-1);
            for (size_t m=row_ptr[k-1]-1; m<row_ptr[k]-1; m++) {
               if ((p=col_ind[m]) != i+1 && p != k) {
                  *(itt++) = RC(i,p-1);
                  if (p-1 != i)
                     *(itt++) = RC(p-1,i);
               }
            }
         }
      }
   }

   p2.resize(itt-p2.begin());
   sort(p2.begin(),p2.end());
   new_end = std::unique(p2.begin(), p2.end());
   p2.erase(new_end, p2.end());
   StoreGraph(size,p2,row_ptr,col_ind);
   return p2.size();
}


void StoreGraph(size_t                 size,
                const std::vector<RC>& p,
                vector<size_t>&        row_ptr,
                vector<size_t>&        col_ind)
{
   size_t length = p.size();
   row_ptr.resize(size+1);
   col_ind.resize(length);
   for (size_t i=0; i<size; i++)
      row_ptr[i] = 0;
   for (size_t j=0; j<length; j++) {
      row_ptr[p[j].first]++;
      col_ind[j] = p[j].second + 1;
   }
   for (size_t k=1; k<size; k++)
      row_ptr[k] += row_ptr[k-1];
   for (int ii=size; ii>0; ii--)
      row_ptr[ii] = row_ptr[ii-1] + 1;
   row_ptr[0] = 1;
}

   
size_t DG0Graph(const Mesh&                            m,
                      Vect<std::pair<size_t,size_t> >& I)
/*------------------------------------------------------------------------------
                                Set graph of matrix
                                (for DG P0/P0 matrix)
  ------------------------------------------------------------------------------*/
{
   Element *el1, *el2, *el;   
   std::vector<RC> pp;
   mesh_elements(m) {
      for (size_t i=1; i<=The_element.getNbSides(); i++) {
         if ((el=The_element.getNeighborElement(i))) {
            pp.push_back(RC(el->n(),element_label));
            for (size_t j=1; j<=el->getNbSides(); j++) {
               el1 = el->getPtrSide(j)->getNeighborElement(1);
               el2 = el->getPtrSide(j)->getNeighborElement(2);
               if (el1 && el1!=el)
                  pp.push_back(RC(el1->n(),element_label));
               if (el2 && el2!=el)
                  pp.push_back(RC(el2->n(),element_label));
            }
         }
      }
   }
   I.resize(pp.size());
   size_t i=1;
   for (vector<RC>::iterator it=pp.begin(); it!=pp.end(); it++)
      I(i++) = *it;
   return pp.size();
}


size_t DGGraph(const Mesh&                            m,
                     Vect<std::pair<size_t,size_t> >& I)
/*------------------------------------------------------------------------------
                         Set graph of matrix for DG matrix
  ------------------------------------------------------------------------------*/
{
   size_t n = 0;
   mesh_elements(m)
      n += The_element.getNbDOF()*The_element.getNbDOF();
   mesh_sides(m) {
      Element *el1=The_side.getNeighborElement(1), *el2=The_side.getNeighborElement(2);
      if (el1 && el2) {
         for (size_t i=1; i<=el1->getNbDOF(); i++)
            for (size_t j=1; j<=el2->getNbDOF(); j++)
               n += 2;
      }
   }
   I.setSize(n);

   size_t fd=1, k=1;
   mesh_elements(m) {
      The_element.setFirstDOF(fd);
      for (size_t i=1; i<=The_element.getNbDOF(); i++)
        for (size_t j=1; j<=The_element.getNbDOF(); j++)
           I(k++) = RC(fd+i-1,fd+j-1);
      fd += The_element.getNbDOF();
   }

   mesh_sides(m) {
      Element *el1=The_side.getNeighborElement(1), *el2=The_side.getNeighborElement(2);
      if (el1 && el2) {
         for (size_t i=1; i<=el1->getNbDOF(); i++) {
            for (size_t j=1; j<=el2->getNbDOF(); j++) {
               I(k++) = RC(el1->getFirstDOF()+i-1,el2->getFirstDOF()+j-1);
               I(k++) = RC(el2->getFirstDOF()+j-1,el1->getFirstDOF()+i-1);
            }
         }
      }
   }
   return I.size();
}

} /* namespace OFELI */
