/*==============================================================================

                                    O  F  E  L  I

                            Object  Finite  Element  Library

  ==============================================================================

   Copyright (C) 1998 - 2014 Rachid Touzani

   This file is part of OFELI.

   OFELI is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   OFELI is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OFELI. If not, see <http://www.gnu.org/licenses/>.

  ==============================================================================

                         Implementation of class 'Partition'

  ==============================================================================*/


#include "mesh/Partition.h"
#include <stdlib.h>
#include <math.h>
#include <fstream>
#include <iostream>
#include <string>

using std::cout;
using std::ifstream;
using std::ostream;
using std::endl;

#include <iomanip>
using std::setw;

#include <algorithm>
#include <vector>
#include "util/util.h"

namespace OFELI {


Partition::Partition(Mesh& mesh,
                     char* file)
{
   _theMesh = &mesh;
   ifstream ipf(file);
   int ne = _theMesh->getNbElements();
   int nn = _theMesh->getNbNodes();
   int *epart = new int [ne];
   int *npart = new int [nn];
   delete [] npart;
   Init(epart);
   delete [] epart;
}


Partition::Partition(Mesh&  mesh,
                     size_t n)
{
   int i, j, esize=0;
   _verbose = 1;
   _theMesh = &mesh;
   _nb_submesh = n;
   try {
      if (n > MAX_NB_SUBMESHES || n < 2)
         THROW_RT("Partition(Mesh,size_t): Number of submeshes " + itos(n) +
                  " exceeds its maximal value: " + itos(MAX_NB_SUBMESHES));
   }
   CATCH("Partition");
   int ne = _theMesh->getNbElements();
   int nn = _theMesh->getNbNodes();
   vector<size_t> type(ne);

// Prepare data for Metis

   cout << "Preparing data for Metis ..." << endl;
   i = 0;
   mesh_elements(*_theMesh) {
      type[i] = 0;
      if (the_element->getShape()==TRIANGLE)
         type[i] = 1;
      else if (the_element->getShape()==TETRAHEDRON)
         type[i] = 2;
      else if (the_element->getShape()==HEXAHEDRON)
         type[i] = 3;
      else if (the_element->getShape()==QUADRILATERAL)
         type[i] = 4;
      if (i==0) {
        switch (type[0]) {

            case 0:
               cout << "Unknown element type" << endl;
               exit(0);
               break;

            case 1:
               esize = 3;
               break;

            case 2:
               esize = 4;
               break;

            case 3:
               esize = 8;
               break;

            case 4:
               esize = 4;
               break;
        }
        if (int(the_element->getNbNodes())!=esize) {
           cout << "Error: Illegal number of element nodes" << endl;
           exit(0);
        }
      }

      else {
         if (type[i]!=type[0]) {
            cout << "Only one element shape is allowed." << endl;
            exit(0);
         }
      }
      i++;
   }

   int *elmnts = new int [esize*ne];
   mesh_elements(*_theMesh)
      for (j=0; j<esize; j++)
	elmnts[(element_label-1)*esize+j] = The_element(j+1)->n() - 1;

// Mesh partitioning by Metis

   cout << "Mesh partitioning by Metis ..." << endl;
   int *epart = new int [ne];
   int *npart = new int [nn];

#ifdef WITH_METIS
   int etype = type[0];
   int numflag=0, edgecut;
   METIS_PartMeshDual(&ne,&nn,elmnts,&etype,&numflag,&_nb_submesh,&edgecut,epart,npart);
#else
   cerr << "Error in class Partition: The library Metis is not installed on this computer.\n";
   cerr << "No mesh partitioning is available." << endl;
   cerr << "The program will continue without partitioning." << endl;
#endif
   delete [] npart;
   Init(epart);
   delete [] epart;
}


Partition::~Partition()
{
   if (_nb_submesh>0) {
     for (int i=0; i<_nb_submesh; i++) {
        delete _theSubMesh[i];
        delete [] _sm2m_node[i];
        delete [] _m2sm_node[i];
        delete [] _sm2m_element[i];
        delete [] _m2sm_element[i];
        delete _interface_side[i];
     }
     delete [] _sm2m_element;
     delete [] _m2sm_element;
     delete [] _sm2m_node;
     delete [] _m2sm_node;
   }
}


void Partition::Init(int* epart)
{
   int i, n;
   size_t j, k;
   size_t ne = _theMesh->getNbElements();
   size_t nn = _theMesh->getNbNodes();

   cout << "Creating submesh data ..." << endl;
   _sm2m_element = new size_t *[_nb_submesh];
   _m2sm_element = new size_t *[_nb_submesh];
   _sm2m_node = new size_t *[_nb_submesh];
   _m2sm_node = new size_t *[_nb_submesh];
   _nb_interface_sides.resize(_nb_submesh);
   for (i=0; i<_nb_submesh; i++) {
      _nb_interface_sides[i] = 0;
      _m2sm_element[i] = new size_t [ne];
      _sm2m_element[i] = new size_t [ne];
      for (j=0; j<ne; j++)
         _m2sm_element[i][j] = _sm2m_element[i][j] = 0;
      _m2sm_node[i] = new size_t [nn];
      _sm2m_node[i] = new size_t [nn];
      for (j=0; j<nn; j++)
         _m2sm_node[i][j] = _sm2m_node[i][j] = 0;
   }

// Number locally nodes and elements in submeshes

   Element *element;
   Node *node;
   for (int sd=0; sd<_nb_submesh; sd++) {
      size_t el_label=1, nd_label=1;
      _theSubMesh[sd] = new Mesh;
      _theSubMesh[sd]->setDim(_theMesh->getDim());
      mesh_elements(*_theMesh) {
         n = element_label;
         element = new Element(The_element);
         if (epart[n-1] == int(sd)) {
            _m2sm_element[sd][n-1] = el_label;
            _sm2m_element[sd][el_label-1] = n;
            element->setLabel(el_label++);
            _theSubMesh[sd]->Add(element);
            for (i=1; i<=int(the_element->getNbNodes()); i++) {
                Node *nd = The_element(i);
                node = new Node (*nd);
                size_t m = nd->n();
                if (_m2sm_node[sd][m-1]==0) {
                   _m2sm_node[sd][m-1] = nd_label;
                   _sm2m_node[sd][nd_label-1] = m;
                   node->setLabel(nd_label++);
                   _theSubMesh[sd]->Add(node);
                }
                node->setLabel(_m2sm_node[sd][m-1]);
                element->Replace(i,node);
            }
         }
      }
      if (_theMesh->NodesAreDOF()==true)
         _theSubMesh[sd]->setNodesForDOF();
      else if (_theMesh->SidesAreDOF()==true)
         _theSubMesh[sd]->setSidesForDOF();
      else if (_theMesh->ElementsAreDOF()==true)
         _theSubMesh[sd]->setElementsForDOF();
      _theSubMesh[sd]->NumberEquations();
      _theSubMesh[sd]->getBoundarySides();
   }

// Create Interface Information

   cout << "Creating interface information ..." << endl;
   _interface_side = new Interface *[_nb_submesh];
   for (i=0; i<_nb_submesh; i++)
      _interface_side[i] = new Interface [_theSubMesh[i]->getNbSides()];
   Side *ss1, *ss2;
   for (int sd1=0; sd1<_nb_submesh; sd1++) {
      for (int sd2=sd1+1; sd2<_nb_submesh; sd2++) {
         for (size_t i=1; i<=_theSubMesh[sd1]->getNbSides(); i++) {
            ss1 = _theSubMesh[sd1]->getPtrSide(i);
            for (j=1; j<=_theSubMesh[sd2]->getNbSides(); j++) {
               ss2 = _theSubMesh[sd2]->getPtrSide(j);
               if (EqualSides(sd1,ss1,sd2,ss2)) {
                  k = _nb_interface_sides[sd1];
                  _interface_side[sd1][k].side1 = ss1->n();
                  _interface_side[sd1][k].sd = sd2;
                  _interface_side[sd1][k].side = ss2->n();
                  _nb_interface_sides[sd1]++;
                  k = _nb_interface_sides[sd2];
                  _interface_side[sd2][k].side1 = ss2->n();
                  _interface_side[sd2][k].sd = sd1;
                  _interface_side[sd2][k].side = ss1->n();
                  _nb_interface_sides[sd2]++;
               }
            }
         }
      }
   }
}


bool Partition::EqualSides(      size_t sm1,
                           const Side*  sd1,
                                 size_t sm2,
                           const Side* sd2)
{
   size_t nn = sd1->getNbNodes();
   vector<size_t> s1(nn), s2(nn);
   for (size_t i=0; i<nn; i++) {
      s1[i] = _sm2m_node[sm1][sd1->getNodeLabel(i+1)-1];
      s2[i] = _sm2m_node[sm2][sd2->getNodeLabel(i+1)-1];
   }
   std::sort(s1.begin(),s1.end());
   std::sort(s2.begin(),s2.end());
   if (s1==s2)
      return true;
   else
      return false;
}


ostream& operator<<(      ostream&   s,
                    const Partition& p)
{
   s << "\nM E S H   P A R T I T I O N   I N F O R M A T I O N" << endl << endl;
   s << "Number of subdomains :   " << setw(6) << p._nb_submesh << endl << endl;
   for (int i=0; i<p._nb_submesh; i++) {
      s << endl << "Sub-domain :             " << setw(6) << i << endl;
      s << "Nb. of nodes :           " << setw(6) << p.getNbNodes(i) << endl;
      s << "Nb. of elements :        " << setw(6) << p.getNbElements(i) << endl;
      s << "Nb. of interface sides : " << setw(6) << p.getNbInterfaceSides(i) << endl;
      if (p._verbose > 1) {
        s << "Side label   Opposite submesh   Opposite Side" << endl;
        for (size_t j=1; j<=p._nb_interface_sides[i]; j++)
           s << setw(8)  << p._interface_side[i][j-1].side1
             << setw(14) << p._interface_side[i][j-1].sd
             << setw(18) << p._interface_side[i][j-1].side
             << endl;
      }
      s << endl;
   }
   return s;
}

}  /* namespace OFELI */
